@@ -1,5 +1,6 @@
 #include "system.h"
 
+#include <rpmio.h>
 #include "cpio.h"
 
 #define CPIO_NEWC_MAGIC	"070701"
@@ -10,6 +11,7 @@
    should both be the same, but really odd things are going to happen if
    that's not true! */
 
+/** */
 struct hardLink {
     struct hardLink * next;
     const char ** files;	/* nlink of these, used by install */
@@ -22,6 +24,7 @@ struct hardLink {
     struct stat sb;
 };
 
+/** */
 struct cpioCrcPhysicalHeader {
     char magic[6];
     char inode[8];
@@ -41,6 +44,7 @@ struct cpioCrcPhysicalHeader {
 
 #define	PHYS_HDR_SIZE	110		/* don't depend on sizeof(struct) */
 
+/** */
 struct cpioHeader {
     ino_t inode;
     mode_t mode;
@@ -53,41 +57,16 @@ struct cpioHeader {
     /*@owned@*/char * path;
 };
 
-static inline off_t saferead(CFD_t *cfd, /*@out@*/void * vbuf, size_t amount)
+/** */
+static inline off_t saferead(FD_t cfd, /*@out@*/void * vbuf, size_t amount)
 {
     off_t rc = 0;
     char * buf = vbuf;
 
     while (amount > 0) {
 	size_t nb;
-	switch (cfd->cpioIoType) {
-	default:
-#ifdef	PARANOID
-		fprintf(stderr, "\tsaferead(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
-		exit(1);
-		break;
-#endif
-	case cpioIoTypeDebug:
-		nb = amount;
-		fprintf(stderr, "\tsaferead(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
-		break;
-	case cpioIoTypeFd:
-		nb = fdRead(cfd->cpioFd, buf, amount);
-		break;
-	case cpioIoTypeFp:
-		nb = fread(buf, amount, 1, cfd->cpioFp);
-		if (nb)
-			nb *= amount;
-		break;
-	case cpioIoTypeGzFd:
-		nb = gzdRead(cfd->cpioGzFd, buf, amount);
-		break;
-#if ENABLE_BZIP2_PAYLOAD
-	case cpioIoTypeBzFd:
-		nb = bzdRead(cfd->cpioBzFd, buf, amount);
-		break;
-#endif
-	}
+
+	nb = Fread(buf, sizeof(buf[0]), amount, cfd);
 	if (nb <= 0)
 		return nb;
 	rc += nb;
@@ -99,58 +78,35 @@ static inline off_t saferead(CFD_t *cfd, /*@out@*/void * vbuf, size_t amount)
     return rc;
 }
 
-static inline off_t ourread(CFD_t * cfd, /*@out@*/void * buf, size_t size)
+/** */
+static inline off_t ourread(FD_t cfd, /*@out@*/void * buf, size_t size)
 {
     off_t i = saferead(cfd, buf, size);
     if (i > 0)
-	cfd->cpioPos += i;
+	fdSetCpioPos(cfd, fdGetCpioPos(cfd) + i);
     return i;
 }
 
-static inline void padinfd(CFD_t * cfd, int modulo)
+/** */
+static inline void padinfd(FD_t cfd, int modulo)
 {
     int buf[10];
     int amount;
     
-    amount = (modulo - cfd->cpioPos % modulo) % modulo;
+    amount = (modulo - fdGetCpioPos(cfd) % modulo) % modulo;
     (void)ourread(cfd, buf, amount);
 }
 
-static inline off_t safewrite(CFD_t *cfd, const void * vbuf, size_t amount)
+/** */
+static inline off_t safewrite(FD_t cfd, const void * vbuf, size_t amount)
 {
     off_t rc = 0;
     const char * buf = vbuf;
 
     while (amount > 0) {
 	size_t nb;
-	switch (cfd->cpioIoType) {
-	default:
-#ifdef	PARANOID
-		fprintf(stderr, "\tsafewrite(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
-		exit(1);
-		break;
-#endif
-	case cpioIoTypeDebug:
-		nb = amount;
-		fprintf(stderr, "\tsafewrite(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
-		break;
-	case cpioIoTypeFd:
-		nb = fdWrite(cfd->cpioFd, buf, amount);
-		break;
-	case cpioIoTypeFp:
-		nb = fwrite(buf, amount, 1, cfd->cpioFp);
-		if (nb)
-			nb *= amount;
-		break;
-	case cpioIoTypeGzFd:
-		nb = gzdWrite(cfd->cpioGzFd, buf, amount);
-		break;
-#if ENABLE_BZIP2_PAYLOAD
-	case cpioIoTypeBzFd:
-		nb = bzdWrite(cfd->cpioBzFd, buf, amount);
-		break;
-#endif
-	}
+
+	nb = Fwrite(buf, sizeof(buf[0]), amount, cfd);
 	if (nb <= 0)
 		return nb;
 	rc += nb;
@@ -163,7 +119,8 @@ static inline off_t safewrite(CFD_t *cfd, const void * vbuf, size_t amount)
     return rc; 
 }
 
-static inline int padoutfd(CFD_t * cfd, size_t * where, int modulo)
+/** */
+static inline int padoutfd(FD_t cfd, size_t * where, int modulo)
 {
     static int buf[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
     int amount;
@@ -176,6 +133,7 @@ static inline int padoutfd(CFD_t * cfd, size_t * where, int modulo)
     return 0;
 }
 
+/** */
 static int strntoul(const char *str, /*@out@*/char **endptr, int base, int num)
 {
     char * buf, * end;
@@ -201,7 +159,8 @@ static int strntoul(const char *str, /*@out@*/char **endptr, int base, int num)
 	sprintf(space, "%8.8lx", (unsigned long) (val)); \
 	memcpy(phys, space, 8);
 
-static int getNextHeader(CFD_t * cfd, struct cpioHeader * chPtr)
+/** */
+static int getNextHeader(FD_t cfd, /*@out@*/ struct cpioHeader * chPtr)
 {
     struct cpioCrcPhysicalHeader physHeader;
     int nameSize;
@@ -225,15 +184,11 @@ static int getNextHeader(CFD_t * cfd, struct cpioHeader * chPtr)
 
     GET_NUM_FIELD(physHeader.devMajor, major);
     GET_NUM_FIELD(physHeader.devMinor, minor);
-#ifndef __LCLINT__
-    chPtr->dev = makedev(major, minor);
-#endif
+    chPtr->dev = /*@-shiftsigned@*/ makedev(major, minor) /*@=shiftsigned@*/ ;
 
     GET_NUM_FIELD(physHeader.rdevMajor, major);
     GET_NUM_FIELD(physHeader.rdevMinor, minor);
-#ifndef __LCLINT__
-    chPtr->rdev = makedev(major, minor);
-#endif
+    chPtr->rdev = /*@-shiftsigned@*/ makedev(major, minor) /*@=shiftsigned@*/ ;
 
     GET_NUM_FIELD(physHeader.namesize, nameSize);
 
@@ -251,6 +206,7 @@ static int getNextHeader(CFD_t * cfd, struct cpioHeader * chPtr)
     return 0;
 }
 
+/** */
 int cpioFileMapCmp(const void * a, const void * b)
 {
     const struct cpioFileMapping * first = a;
@@ -260,6 +216,7 @@ int cpioFileMapCmp(const void * a, const void * b)
 }
 
 /* This could trash files in the path! I'm not sure that's a good thing */
+/** */
 static int createDirectory(char * path, mode_t perms)
 {
     struct stat sb;
@@ -296,6 +253,7 @@ static int createDirectory(char * path, mode_t perms)
     return 0;
 }
 
+/** */
 static int setInfo(struct cpioHeader * hdr)
 {
     int rc = 0;
@@ -321,9 +279,10 @@ static int setInfo(struct cpioHeader * hdr)
     return rc;
 }
 
-static int checkDirectory(char * filename)
+/** */
+static int checkDirectory(const char * filename)
 {
-    static char * lastDir = NULL;	/* XXX memory leak */
+    /*@only@*/ static char * lastDir = NULL;	/* XXX memory leak */
     static int lastDirLength = 0;
     static int lastDirAlloced = 0;
     int length = strlen(filename);
@@ -366,15 +325,16 @@ static int checkDirectory(char * filename)
     return rc;
 }
 
-static int expandRegular(CFD_t * cfd, struct cpioHeader * hdr,
+/** */
+static int expandRegular(FD_t cfd, struct cpioHeader * hdr,
 			 cpioCallback cb, void * cbData)
 {
-    FD_t out;
-    char buf[8192];
+    FD_t ofd;
+    char buf[BUFSIZ];
     int bytesRead;
     int left = hdr->size;
     int rc = 0;
-    struct cpioCallbackInfo cbInfo;
+    struct cpioCallbackInfo cbInfo = { NULL, 0, 0, 0 };
     struct stat sb;
 
     /* Rename the old file before attempting unlink to avoid EBUSY errors */
@@ -396,8 +356,8 @@ static int expandRegular(CFD_t * cfd, struct cpioHeader * hdr,
 	}
     }
 
-    out = fdOpen(hdr->path, O_CREAT | O_WRONLY, 0);
-    if (fdFileno(out) < 0) 
+    ofd = Fopen(hdr->path, "w.ufdio");
+    if (ofd == NULL || Ferror(ofd))
 	return CPIOERR_OPEN_FAILED;
 
     cbInfo.file = hdr->path;
@@ -410,7 +370,7 @@ static int expandRegular(CFD_t * cfd, struct cpioHeader * hdr,
 	    break;
 	}
 
-	if (fdWrite(out, buf, bytesRead) != bytesRead) {
+	if (Fwrite(buf, sizeof(buf[0]), bytesRead, ofd) != bytesRead) {
 	    rc = CPIOERR_COPY_FAILED;
 	    break;
 	}
@@ -420,17 +380,18 @@ static int expandRegular(CFD_t * cfd, struct cpioHeader * hdr,
 	/* don't call this with fileSize == fileComplete */
 	if (!rc && cb && left) {
 	    cbInfo.fileComplete = hdr->size - left;
-	    cbInfo.bytesProcessed = cfd->cpioPos;
+	    cbInfo.bytesProcessed = fdGetCpioPos(cfd);
 	    cb(&cbInfo, cbData);
 	}
     }
 
-    fdClose(out);
+    Fclose(ofd);
     
     return rc;
 }
 
-static int expandSymlink(CFD_t * cfd, struct cpioHeader * hdr)
+/** */
+static int expandSymlink(FD_t cfd, struct cpioHeader * hdr)
 {
     char buf[2048], buf2[2048];
     struct stat sb;
@@ -463,7 +424,8 @@ static int expandSymlink(CFD_t * cfd, struct cpioHeader * hdr)
     return 0;
 }
 
-static int expandFifo( /*@unused@*/ CFD_t * cfd, struct cpioHeader * hdr)
+/** */
+static int expandFifo( /*@unused@*/ FD_t cfd, struct cpioHeader * hdr)
 {
     struct stat sb;
 
@@ -480,7 +442,8 @@ static int expandFifo( /*@unused@*/ CFD_t * cfd, struct cpioHeader * hdr)
     return 0; 
 }
 
-static int expandDevice( /*@unused@*/ CFD_t * cfd, struct cpioHeader * hdr)
+/** */
+static int expandDevice( /*@unused@*/ FD_t cfd, struct cpioHeader * hdr)
 {
     struct stat sb;
 
@@ -498,18 +461,20 @@ static int expandDevice( /*@unused@*/ CFD_t * cfd, struct cpioHeader * hdr)
     return 0;
 }
 
+/** */
 static void freeLink( /*@only@*/ struct hardLink * li)
 {
     int i;
 
     for (i = 0; i < li->nlink; i++) {
 	if (li->files[i] == NULL) continue;
-	free((void *)li->files[i]);
+	/*@-unqualifiedtrans@*/ free((void *)li->files[i]) /*@=unqualifiedtrans@*/ ;
 	li->files[i] = NULL;
     }
     free(li->files);
 }
 
+/** */
 static int createLinks(struct hardLink * li, /*@out@*/const char ** failedFile)
 {
     int i;
@@ -533,7 +498,7 @@ static int createLinks(struct hardLink * li, /*@out@*/const char ** failedFile)
 	    return CPIOERR_LINK_FAILED;
 	}
 
-	free((void *)li->files[i]);
+	/*@-unqualifiedtrans@*/ free((void *)li->files[i]) /*@=unqualifiedtrans@*/ ;
 	li->files[i] = NULL;
 	li->linksLeft--;
     }
@@ -541,7 +506,8 @@ static int createLinks(struct hardLink * li, /*@out@*/const char ** failedFile)
     return 0;
 }
 
-static int eatBytes(CFD_t * cfd, int amount)
+/** */
+static int eatBytes(FD_t cfd, int amount)
 {
     char buf[4096];
     int bite;
@@ -556,7 +522,8 @@ static int eatBytes(CFD_t * cfd, int amount)
     return 0;
 }
 
-int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings, 
+/** */
+int cpioInstallArchive(FD_t cfd, struct cpioFileMapping * mappings, 
 		       int numMappings, cpioCallback cb, void * cbData,
 		       const char ** failedFile)
 {
@@ -567,11 +534,11 @@ int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
     struct cpioFileMapping needle;
     mode_t cpioMode;
     int olderr;
-    struct cpioCallbackInfo cbInfo;
+    struct cpioCallbackInfo cbInfo = { NULL, 0, 0, 0 };
     struct hardLink * links = NULL;
     struct hardLink * li = NULL;
 
-    cfd->cpioPos = 0;
+    fdSetCpioPos(cfd, 0);
     if (failedFile)
 	*failedFile = NULL;
 
@@ -585,7 +552,7 @@ int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
 	}
 
 	if (!strcmp(ch.path, TRAILER)) {
-	    free(ch.path);
+	    if (ch.path) free(ch.path);
 	    break;
 	}
 
@@ -602,7 +569,7 @@ int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
 
 	    if (map) {
 		if (map->mapFlags & CPIO_MAP_PATH) {
-		    free(ch.path);
+		    if (ch.path) free(ch.path);
 		    ch.path = xstrdup(map->fsPath);
 		} 
 
@@ -699,11 +666,11 @@ int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
 	    cbInfo.file = ch.path;
 	    cbInfo.fileSize = ch.size;
 	    cbInfo.fileComplete = ch.size;
-	    cbInfo.bytesProcessed = cfd->cpioPos;
+	    cbInfo.bytesProcessed = fdGetCpioPos(cfd);
 	    cb(&cbInfo, cbData);
 	}
 
-	free(ch.path);
+	if (ch.path) free(ch.path);
 	ch.path = NULL;
     } while (1 && !rc);
 
@@ -736,7 +703,8 @@ int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
     return rc;
 }
 
-static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map, 
+/** */
+static int writeFile(FD_t cfd, struct stat sb, struct cpioFileMapping * map, 
 		     /*@out@*/size_t * sizep, int writeData)
 {
     struct cpioCrcPhysicalHeader hdr;
@@ -761,7 +729,7 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
 	/* While linux puts the size of a symlink in the st_size field,
 	   I don't think that's a specified standard */
 
-	amount = readlink(map->fsPath, symbuf, sizeof(symbuf));
+	amount = Readlink(map->fsPath, symbuf, sizeof(symbuf));
 	if (amount <= 0) {
 	    return CPIOERR_READLINK_FAILED;
 	}
@@ -778,10 +746,10 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
     SET_NUM_FIELD(hdr.mtime, sb.st_mtime, buf);
     SET_NUM_FIELD(hdr.filesize, sb.st_size, buf);
 
-    num = major(sb.st_dev); SET_NUM_FIELD(hdr.devMajor, num, buf);
-    num = minor(sb.st_dev); SET_NUM_FIELD(hdr.devMinor, num, buf);
-    num = major(sb.st_rdev); SET_NUM_FIELD(hdr.rdevMajor, num, buf);
-    num = minor(sb.st_rdev); SET_NUM_FIELD(hdr.rdevMinor, num, buf);
+    num = major((unsigned)sb.st_dev); SET_NUM_FIELD(hdr.devMajor, num, buf);
+    num = minor((unsigned)sb.st_dev); SET_NUM_FIELD(hdr.devMinor, num, buf);
+    num = major((unsigned)sb.st_rdev); SET_NUM_FIELD(hdr.rdevMajor, num, buf);
+    num = minor((unsigned)sb.st_rdev); SET_NUM_FIELD(hdr.rdevMinor, num, buf);
 
     num = strlen(map->archivePath) + 1; SET_NUM_FIELD(hdr.namesize, num, buf);
     memcpy(hdr.checksum, "00000000", 8);
@@ -801,12 +769,14 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
 	size_t nmapped;
 #endif
 
-	if (fdFileno(datafd = fdOpen(map->fsPath, O_RDONLY, 0)) < 0)
+	/* XXX unbuffered mmap generates *lots* of fdio debugging */
+	datafd = Fopen(map->fsPath, "r.ufdio");
+	if (datafd == NULL || Ferror(datafd))
 	    return CPIOERR_OPEN_FAILED;
 
 #if HAVE_MMAP
 	nmapped = 0;
-	mapped = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fdFileno(datafd), 0);
+	mapped = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, Fileno(datafd), 0);
 	if (mapped != (void *)-1) {
 	    b = (char *)mapped;
 	    nmapped = sb.st_size;
@@ -825,11 +795,12 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
 	  } else
 #endif
 	  {
-	    amount = fdRead(datafd, b, 
-		     sb.st_size > sizeof(buf) ? sizeof(buf) : sb.st_size);
+	    amount = Fread(b, sizeof(buf[0]),
+			(sb.st_size > sizeof(buf) ? sizeof(buf) : sb.st_size),
+			datafd);
 	    if (amount <= 0) {
 		olderrno = errno;
-		fdClose(datafd);
+		Fclose(datafd);
 		errno = olderrno;
 		return CPIOERR_READ_FAILED;
 	    }
@@ -837,7 +808,7 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
 
 	    if ((rc = safewrite(cfd, b, amount)) != amount) {
 		olderrno = errno;
-		fdClose(datafd);
+		Fclose(datafd);
 		errno = olderrno;
 		return rc;
 	    }
@@ -851,7 +822,7 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
 	}
 #endif
 
-	fdClose(datafd);
+	Fclose(datafd);
     } else if (writeData && S_ISLNK(sb.st_mode)) {
 	if ((rc = safewrite(cfd, symbuf, amount)) != amount)
 	    return rc;
@@ -868,7 +839,8 @@ static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map,
     return 0;
 }
 
-static int writeLinkedFile(CFD_t *cfd, struct hardLink * hlink, 
+/** */
+static int writeLinkedFile(FD_t cfd, struct hardLink * hlink, 
 			   struct cpioFileMapping * mappings,
 			   cpioCallback cb, void * cbData,
 			   /*@out@*/size_t * sizep,
@@ -876,7 +848,7 @@ static int writeLinkedFile(CFD_t *cfd, struct hardLink * hlink,
 {
     int i, rc;
     size_t size, total;
-    struct cpioCallbackInfo cbinfo;
+    struct cpioCallbackInfo cbInfo = { NULL, 0, 0, 0 };
 
     total = 0;
 
@@ -891,8 +863,8 @@ static int writeLinkedFile(CFD_t *cfd, struct hardLink * hlink,
 	total += size;
 
 	if (cb) {
-	    cbinfo.file = mappings[i].archivePath;
-	    cb(&cbinfo, cbData);
+	    cbInfo.file = mappings[i].archivePath;
+	    cb(&cbInfo, cbData);
 	}
     }
 
@@ -911,21 +883,22 @@ static int writeLinkedFile(CFD_t *cfd, struct hardLink * hlink,
 	*sizep = total;
 
     if (cb) {
-	cbinfo.file = mappings[i].archivePath;
-	cb(&cbinfo, cbData);
+	cbInfo.file = mappings[i].archivePath;
+	cb(&cbInfo, cbData);
     }
 
     return 0;
 }
 
-int cpioBuildArchive(CFD_t *cfd, struct cpioFileMapping * mappings, 
+/** */
+int cpioBuildArchive(FD_t cfd, struct cpioFileMapping * mappings, 
 		     int numMappings, cpioCallback cb, void * cbData,
 		     unsigned int * archiveSize, const char ** failedFile)
 {
     size_t size, totalsize = 0;
     int rc;
     int i;
-    struct cpioCallbackInfo cbinfo;
+    struct cpioCallbackInfo cbInfo = { NULL, 0, 0, 0 };
     struct cpioCrcPhysicalHeader hdr;
     struct stat sb;
 /*@-fullinitblock@*/
@@ -937,9 +910,9 @@ int cpioBuildArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
 
     for (i = 0; i < numMappings; i++) {
 	if (mappings[i].mapFlags & CPIO_FOLLOW_SYMLINKS)
-	    rc = stat(mappings[i].fsPath, &sb);
+	    rc = Stat(mappings[i].fsPath, &sb);
 	else
-	    rc = lstat(mappings[i].fsPath, &sb);
+	    rc = Lstat(mappings[i].fsPath, &sb);
 
 	if (rc) {
 	    if (failedFile)
@@ -989,8 +962,8 @@ int cpioBuildArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
 	    }
 
 	    if (cb) {
-		cbinfo.file = mappings[i].archivePath;
-		cb(&cbinfo, cbData);
+		cbInfo.file = mappings[i].archivePath;
+		cb(&cbInfo, cbData);
 	    }
 
 	    totalsize += size;
@@ -1031,6 +1004,7 @@ int cpioBuildArchive(CFD_t *cfd, struct cpioFileMapping * mappings,
     return 0;
 }
 
+/** */
 const char * cpioStrerror(int rc)
 {
     static char msg[256];
@@ -1074,7 +1048,7 @@ const char * cpioStrerror(int rc)
 	if (l > 0) strncat(msg, s, l);
 	l -= strlen(s);
     }
-    if (rc & CPIOERR_CHECK_ERRNO) {
+    if ((rc & CPIOERR_CHECK_ERRNO) && myerrno) {
 	s = _(" failed - ");
 	if (l > 0) strncat(msg, s, l);
 	l -= strlen(s);
