@@ -1,6 +1,7 @@
 #include "system.h"
 
 #include <rpmlib.h>
+#include <rpmurl.h>
 
 unsigned int dbiIndexSetCount(dbiIndexSet set) {
     return set.count;
@@ -22,14 +23,16 @@ unsigned int dbiIndexRecordFileNumber(dbiIndexSet set, int recno) {
     return set.recs[recno].fileNumber;
 }
 
-dbiIndex * dbiOpenIndex(const char * filename, int flags, int perms, DBTYPE type) {
+dbiIndex * dbiOpenIndex(const char * urlfn, int flags, int perms, DBTYPE type) {
     dbiIndex * dbi;
+    const char * filename;
         
+    (void) urlPath(urlfn, &filename);
     dbi = xmalloc(sizeof(*dbi));
-    dbi->db = dbopen(filename, flags, perms, type, NULL);
-    if (!dbi->db) {
+    if (*filename == '\0' ||
+	(dbi->db = dbopen(filename, flags, perms, type, NULL)) == NULL) {
 	free(dbi);
-	rpmError(RPMERR_DBOPEN, _("cannot open file %s: %s"), filename, 
+	rpmError(RPMERR_DBOPEN, _("cannot open file %s: %s"), urlfn, 
 			      strerror(errno));
 	return NULL;
     }
@@ -148,6 +151,7 @@ dbiIndexSet dbiCreateIndexRecord(void) {
     
 void dbiFreeIndexRecord(dbiIndexSet set) {
     if (set.recs) free(set.recs);
+    set.recs = NULL;
 }
 
 /* returns 1 on failure */
