@@ -91,10 +91,12 @@ static void alCreate( /*@out@*/ struct availableList * al)
 
     al->index.index = NULL;
     al->index.size = 0;
-    alFreeIndex(al);
+
+    al->numDirs = 0;
+    al->dirs = NULL;
 }
 
-static void alFree( /*@only@*/ struct availableList * al)
+static void alFree(struct availableList * al)
 {
     int i;
     rpmRelocation * r;
@@ -104,8 +106,8 @@ static void alFree( /*@only@*/ struct availableList * al)
 	    free(al->list[i].provides);
 	if (al->list[i].providesEVR)
 	    free(al->list[i].providesEVR);
-	if (al->list[i].files)
-	    free(al->list[i].files);
+	if (al->list[i].baseNames)
+	    free(al->list[i].baseNames);
 	headerFree(al->list[i].h);
 
 	if (al->list[i].relocs) {
@@ -115,28 +117,60 @@ static void alFree( /*@only@*/ struct availableList * al)
 	    }
 	    free(al->list[i].relocs);
 	}
+	if (al->list[i].fd)
+	    al->list[i].fd = fdFree(al->list[i].fd, "alAddPackage (alFree)");
     }
 
+    for (i = 0; i < al->numDirs; i++) {
+	free(al->dirs[i].dirName);
+	free(al->dirs[i].files);
+    }
+
+    if (al->numDirs)
+	free(al->dirs);
+    al->dirs = NULL;
+
     if (al->alloced && al->list)
 	free(al->list);
     al->list = NULL;
     alFreeIndex(al);
 }
 
+static int dirInfoCompare(const void * one, const void * two) {
+    const struct dirInfo * a = one;
+    const struct dirInfo * b = two;
+    int lenchk = a->dirNameLen - b->dirNameLen;
+
+    if (lenchk)
+	return lenchk;
+    /* XXX FIXME: this might do "backward" strcmp for speed */
+    return strcmp(a->dirName, b->dirName);
+}
+
 static /*@exposed@*/ struct availablePackage * alAddPackage(struct availableList * al,
-					      Header h, const void * key,
-			 		      FD_t fd, rpmRelocation * relocs)
+		Header h, /*@dependent@*/ const void * key,
+		FD_t fd, rpmRelocation * relocs)
 {
     struct availablePackage * p;
     rpmRelocation * r;
     int i;
+    int_32 * dirIndexes;
+    const char ** dirNames;
+    int numDirs, dirNum;
+    int * dirMapping;
+    struct dirInfo dirNeedle;
+    struct dirInfo * dirMatch;
+    int first, last, fileNum;
+    int origNumDirs;
+    int pkgNum;
 
     if (al->size == al->alloced) {
 	al->alloced += 5;
 	al->list = xrealloc(al->list, sizeof(*al->list) * al->alloced);
     }
 
-    p = al->list + al->size++;
+    pkgNum = al->size++;
+    p = al->list + pkgNum;
     p->h = headerLink(h);	/* XXX reference held by transaction set */
 
     headerNVR(p->h, &p->name, &p->version, &p->release);
@@ -159,22 +193,74 @@ static /*@exposed@*/ struct availablePackage * alAddPackage(struct availableList
 	    p->provideFlags = NULL;
     }
 
-    if (!headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &p->files,
-	&p->filesCount)) {
+    if (!headerGetEntryMinMemory(h, RPMTAG_BASENAMES, NULL, (void **) 
+				 &p->baseNames, &p->filesCount)) {
 	p->filesCount = 0;
-	p->files = NULL;
+	p->baseNames = NULL;
+    } else {
+        headerGetEntryMinMemory(h, RPMTAG_DIRNAMES, NULL, (void **) 
+				 &dirNames, &numDirs);
+        headerGetEntryMinMemory(h, RPMTAG_DIRINDEXES, NULL, (void **) 
+				 &dirIndexes, NULL);
+
+	/* XXX FIXME: We ought to relocate the directory list here */
+
+        dirMapping = alloca(sizeof(*dirMapping) * numDirs);
+
+	/* allocated enough space for all the directories we could possible
+	   need to add */
+	al->dirs = xrealloc(al->dirs, 
+			    sizeof(*al->dirs) * (al->numDirs + numDirs));
+	origNumDirs = al->numDirs;
+
+	for (dirNum = 0; dirNum < numDirs; dirNum++) {
+	    dirNeedle.dirName = (char *) dirNames[dirNum];
+	    dirNeedle.dirNameLen = strlen(dirNames[dirNum]);
+	    dirMatch = bsearch(&dirNeedle, al->dirs, origNumDirs,
+			       sizeof(dirNeedle), dirInfoCompare);
+	    if (dirMatch) {
+		dirMapping[dirNum] = dirMatch - al->dirs;
+	    } else {
+		dirMapping[dirNum] = al->numDirs;
+		al->dirs[al->numDirs].dirName = xstrdup(dirNames[dirNum]);
+		al->dirs[al->numDirs].dirNameLen = strlen(dirNames[dirNum]);
+		al->dirs[al->numDirs].files = NULL;
+		al->dirs[al->numDirs].numFiles = 0;
+		al->numDirs++;
+	    }
+	}
+
+	free(dirNames);
+
+	first = 0;
+	while (first < p->filesCount) {
+	    last = first;
+	    while ((last + 1) < p->filesCount) {
+		if (dirIndexes[first] != dirIndexes[last + 1]) break;
+		last++;
+	    }
+
+	    dirMatch = al->dirs + dirMapping[dirIndexes[first]];
+	    dirMatch->files = xrealloc(dirMatch->files,
+		sizeof(*dirMatch->files) * 
+		    (dirMatch->numFiles + last - first + 1));
+	    for (fileNum = first; fileNum <= last; fileNum++) {
+		dirMatch->files[dirMatch->numFiles].baseName =
+		    p->baseNames[fileNum];
+		dirMatch->files[dirMatch->numFiles].pkgNum = pkgNum;
+		dirMatch->numFiles++;
+	    }
+
+	    first = last + 1;
+	}
+
+	if (origNumDirs + al->numDirs)
+	    qsort(al->dirs, al->numDirs, sizeof(dirNeedle), dirInfoCompare);
+
     }
 
-    /* We don't use these entries (and rpm >= 2 never have) and they are
-       pretty misleading. Let's just get rid of them so they don't confuse
-       anyone. */
-    if (headerIsEntry(h, RPMTAG_FILEUSERNAME))
-	headerRemoveEntry(h, RPMTAG_FILEUIDS);
-    if (headerIsEntry(h, RPMTAG_FILEGROUPNAME))
-	headerRemoveEntry(h, RPMTAG_FILEGIDS);
-
     p->key = key;
-    p->fd = fd;
+    p->fd = (fd ? fdLink(fd, "alAddPackage") : NULL);
 
     if (relocs) {
 	for (i = 0, r = relocs; r->oldPath || r->newPath; i++, r++);
@@ -211,12 +297,8 @@ static void alMakeIndex(struct availableList * al)
     if (ai->size) return;
 
     ai->size = al->size;
-    for (i = 0; i < al->size; i++) {
+    for (i = 0; i < al->size; i++) 
 	ai->size += al->list[i].providesCount;
-    }
-    for (i = 0; i < al->size; i++) {
-	ai->size += al->list[i].filesCount;
-    }
 
     if (ai->size) {
 	ai->index = xcalloc(ai->size, sizeof(*ai->index));
@@ -234,13 +316,6 @@ static void alMakeIndex(struct availableList * al)
 		ai->index[k].type = IET_PROVIDES;
 		k++;
 	    }
-
-	    for (j = 0; j < al->list[i].filesCount; j++) {
-		ai->index[k].package = al->list + i;
-		ai->index[k].entry = al->list[i].files[j];
-		ai->index[k].type = IET_FILE;
-		k++;
-	    }
 	}
 
 	qsort(ai->index, ai->size, sizeof(*ai->index), indexcmp);
@@ -260,7 +335,10 @@ static int intcmp(const void * a, const void *b)
     return 1;
 }
 
-static void parseEVR(char *evr, /*@exposed@*/ /*@out@*/const char **ep, /*@exposed@*/ /*@out@*/const char **vp, /*@exposed@*/ /*@out@*/const char **rp) /*@modifies evr,*ep,*vp,*rp @*/
+static void parseEVR(char *evr,
+	/*@exposed@*/ /*@out@*/ const char **ep,
+	/*@exposed@*/ /*@out@*/ const char **vp,
+	/*@exposed@*/ /*@out@*/const char **rp) /*@modifies evr,*ep,*vp,*rp @*/
 {
     const char *epoch;
     const char *version;		/* assume only version is present */
@@ -502,7 +580,7 @@ rpmTransactionSet rpmtransCreateSet(rpmdb db, const char * root)
 
     rpmdep->orderAlloced = 5;
     rpmdep->orderCount = 0;
-    rpmdep->order = xcalloc(rpmdep->orderAlloced, sizeof(*rpmdep->order)); /* XXX memory leak */
+    rpmdep->order = xcalloc(rpmdep->orderAlloced, sizeof(*rpmdep->order));
 
     return rpmdep;
 }
@@ -628,7 +706,8 @@ int rpmtransAddPackage(rpmTransactionSet rpmdep, Header h, FD_t fd,
     return 0;
 }
 
-void rpmtransAvailablePackage(rpmTransactionSet rpmdep, Header h, void * key)
+void rpmtransAvailablePackage(rpmTransactionSet rpmdep, Header h,
+	const void * key)
 {
     struct availablePackage * al;
     al = alAddPackage(&rpmdep->availablePackages, h, key, NULL, NULL);
@@ -641,16 +720,22 @@ void rpmtransRemovePackage(rpmTransactionSet rpmdep, int dboffset)
 
 void rpmtransFree(rpmTransactionSet rpmdep)
 {
-    alFree(&rpmdep->addedPackages);
-    alFree(&rpmdep->availablePackages);
+    struct availableList * addedPackages = &rpmdep->addedPackages;
+    struct availableList * availablePackages = &rpmdep->availablePackages;
+
+    alFree(addedPackages);
+    alFree(availablePackages);
     free(rpmdep->removedPackages);
     xfree(rpmdep->root);
+    free(rpmdep->order);
+    if (rpmdep->scriptFd)
+	rpmdep->scriptFd = fdFree(rpmdep->scriptFd, "rpmtransSetScriptFd (rpmtransFree");
 
     free(rpmdep);
 }
 
-void rpmdepFreeConflicts(struct rpmDependencyConflict * conflicts, int
-			 numConflicts)
+void rpmdepFreeConflicts(struct rpmDependencyConflict * conflicts,
+			int numConflicts)
 {
     int i;
 
@@ -666,14 +751,57 @@ void rpmdepFreeConflicts(struct rpmDependencyConflict * conflicts, int
     free(conflicts);
 }
 
-static /*@exposed@*/ struct availablePackage * alSatisfiesDepend(struct availableList * al,
-	const char *keyType, const char *keyDepend,
+/*@dependent@*/ /*@null@*/ static struct availablePackage *
+alFileSatisfiesDepend(struct availableList * al,
+	const char * keyType, const char * fileName)
+{
+    int i;
+    const char * dirName;
+    const char * baseName;
+    struct dirInfo dirNeedle;
+    struct dirInfo * dirMatch;
+
+    {	char * chptr = xstrdup(fileName);
+	dirName = chptr;
+	chptr = strrchr(chptr, '/');
+	chptr++;
+	*chptr = '\0';
+    }
+
+    dirNeedle.dirName = (char *) dirName;
+    dirNeedle.dirNameLen = strlen(dirName);
+    dirMatch = bsearch(&dirNeedle, al->dirs, al->numDirs,
+		       sizeof(dirNeedle), dirInfoCompare);
+    xfree(dirName);
+    if (!dirMatch) return NULL;
+
+    baseName = strrchr(fileName, '/') + 1;
+
+    /* XXX FIXME: these file lists should be sorted and bsearched */
+    for (i = 0; i < dirMatch->numFiles; i++) {
+	if (!strcmp(dirMatch->files[i].baseName, baseName)) {
+	    if (keyType)
+		rpmMessage(RPMMESS_DEBUG, _("%s: %s satisfied by added file "
+			    "list.\n"), keyType, fileName);
+	    return al->list + dirMatch->files[i].pkgNum;
+	}
+    }
+
+    return NULL;
+}
+
+/*@dependent@*/ /*@null@*/ static struct availablePackage * alSatisfiesDepend(
+	struct availableList * al,
+	const char * keyType, const char * keyDepend,
 	const char * keyName, const char * keyEVR, int keyFlags)
 {
     struct availableIndexEntry needle, * match;
-    struct availablePackage *p;
+    struct availablePackage * p;
     int i, rc;
 
+    if (*keyName == '/')
+	return alFileSatisfiesDepend(al, keyType, keyName);
+
     if (!al->index.size) return NULL;
 
     needle.entry = keyName;
@@ -709,13 +837,9 @@ static /*@exposed@*/ struct availablePackage * alSatisfiesDepend(struct availabl
 	    if (rc) break;
 	}
 	if (keyType && keyDepend && rc)
-	    rpmMessage(RPMMESS_DEBUG, _("%s: %s satisfied by added provide.\n"), keyType, keyDepend);
+	    rpmMessage(RPMMESS_DEBUG, _("%s: %s satisfied by added "
+			"provide.\n"), keyType, keyDepend);
     	break;
-    case IET_FILE:
-	rc = 1;
-	if (keyType && keyDepend && rc)
-	    rpmMessage(RPMMESS_DEBUG, _("%s: %s satisfied by added file list.\n"), keyType, keyDepend);
-	break;
     }
 
     if (rc)
@@ -727,9 +851,9 @@ static /*@exposed@*/ struct availablePackage * alSatisfiesDepend(struct availabl
 /* 2 == error */
 /* 1 == dependency not satisfied */
 static int unsatisfiedDepend(rpmTransactionSet rpmdep,
-	const char *keyType, const char *keyDepend,
+	const char * keyType, const char * keyDepend,
 	const char * keyName, const char * keyEVR, int keyFlags,
-	/*@out@*/struct availablePackage ** suggestion)
+	/*@out@*/ struct availablePackage ** suggestion)
 {
     dbiIndexSet matches;
     int rc = 0;	/* assume dependency is satisfied */
@@ -847,7 +971,7 @@ exit:
 }
 
 static int checkPackageDeps(rpmTransactionSet rpmdep, struct problemsSet * psp,
-		Header h, const char *keyName)
+		Header h, const char * keyName)
 {
     const char * name, * version, * release;
     const char ** requires, ** requiresEVR = NULL;
@@ -990,7 +1114,7 @@ static int checkPackageDeps(rpmTransactionSet rpmdep, struct problemsSet * psp,
 /* Adding: check name/provides key against each conflict match. */
 /* Erasing: check name/provides/filename key against each requiredby match. */
 static int checkPackageSet(rpmTransactionSet rpmdep, struct problemsSet * psp,
-	const char *key, dbiIndexSet *matches)
+	const char * key, dbiIndexSet * matches)
 {
     Header h;
     int i;
@@ -1214,7 +1338,7 @@ int rpmdepOrder(rpmTransactionSet rpmdep)
     qsort(orderList, rpmdep->addedPackages.size, sizeof(*orderList),
 	  orderListIndexCmp);
 
-    newOrder = xmalloc(sizeof(*newOrder) * rpmdep->orderCount); /* XXX memory leak */
+    newOrder = xmalloc(sizeof(*newOrder) * rpmdep->orderCount);
     for (i = 0, newOrderCount = 0; i < orderingCount; i++) {
 	key.alIndex = ordering[i];
 	needle = bsearch(&key, orderList, rpmdep->addedPackages.size,
@@ -1254,8 +1378,6 @@ int rpmdepCheck(rpmTransactionSet rpmdep,
 {
     struct availablePackage * p;
     int i, j;
-    const char ** files;
-    int fileCount;
     int rc;
     Header h = NULL;
     struct problemsSet ps;
@@ -1301,9 +1423,6 @@ int rpmdepCheck(rpmTransactionSet rpmdep,
 
     /* now look at the removed packages and make sure they aren't critical */
     for (i = 0; i < rpmdep->numRemovedPackages; i++) {
-	const char *name;
-	const char ** provides;
-	int providesCount;
 
 	h = rpmdbGetRecord(rpmdep->db, rpmdep->removedPackages[i]);
 	if (h == NULL) {
@@ -1313,39 +1432,69 @@ int rpmdepCheck(rpmTransactionSet rpmdep,
 	    goto exit;
 	}
 
-	headerNVR(h, &name, NULL, NULL);
+	{   const char * name;
+	    headerNVR(h, &name, NULL, NULL);
 
-	/* Erasing: check name against requiredby matches. */
-	if (checkDependentPackages(rpmdep, &ps, name))
-	    goto exit;
-
-	if (headerGetEntry(h, RPMTAG_PROVIDENAME, NULL, (void **) &provides,
-		 &providesCount)) {
-	    rc = 0;
-	    for (j = 0; j < providesCount; j++) {
-		/* Erasing: check provides against requiredby matches. */
-		if (checkDependentPackages(rpmdep, &ps, provides[j])) {
-		    rc = 1;
-		    break;
-		}
-	    }
-	    free(provides);
-	    if (rc)	goto exit;
+	    /* Erasing: check name against requiredby matches. */
+	    if (checkDependentPackages(rpmdep, &ps, name))
+		goto exit;
 	}
 
-	if (headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &files,
-		 &fileCount)) {
-	    rc = 0;
-	    for (j = 0; j < fileCount; j++) {
-		/* Erasing: check filename against requiredby matches. */
-		if (checkDependentPackages(rpmdep, &ps, files[j])) {
+	{   const char ** provides;
+	    int providesCount;
+
+	    if (headerGetEntry(h, RPMTAG_PROVIDENAME, NULL, (void **) &provides,
+				&providesCount)) {
+		rc = 0;
+		for (j = 0; j < providesCount; j++) {
+		    /* Erasing: check provides against requiredby matches. */
+		    if (checkDependentPackages(rpmdep, &ps, provides[j])) {
 		    rc = 1;
 		    break;
+		    }
 		}
+		xfree(provides);
+		if (rc)
+		    goto exit;
 	    }
+	}
 
-	    free(files);
-	    if (rc) goto exit;
+	{   const char ** baseNames, ** dirNames;
+	    int_32 * dirIndexes;
+	    int fileCount;
+	    char * fileName = NULL;
+	    int fileAlloced = 0;
+	    int len;
+
+	    if (headerGetEntry(h, RPMTAG_BASENAMES, NULL, 
+			   (void **) &baseNames, &fileCount)) {
+		headerGetEntry(h, RPMTAG_DIRNAMES, NULL, 
+			    (void **) &dirNames, NULL);
+		headerGetEntry(h, RPMTAG_DIRINDEXES, NULL, 
+			    (void **) &dirIndexes, NULL);
+		rc = 0;
+		for (j = 0; j < fileCount; j++) {
+		    len = strlen(baseNames[j]) + 1 + 
+			  strlen(dirNames[dirIndexes[j]]);
+		    if (len > fileAlloced) {
+			fileAlloced = len * 2;
+			fileName = xrealloc(fileName, fileAlloced);
+		    }
+		    strcpy(fileName, dirNames[dirIndexes[j]]);
+		    strcat(fileName, baseNames[j]);
+		    /* Erasing: check filename against requiredby matches. */
+		    if (checkDependentPackages(rpmdep, &ps, fileName)) {
+			rc = 1;
+			break;
+		    }
+		}
+
+		free(fileName);
+		free(baseNames);
+		free(dirNames);
+		if (rc)
+		    goto exit;
+	    }
 	}
 
 	headerFree(h);	h = NULL;
@@ -1362,9 +1511,8 @@ int rpmdepCheck(rpmTransactionSet rpmdep,
     return 0;
 
 exit:
-    if (h) {
+    if (h) 
 	headerFree(h);
-    }
     if (ps.problems)	free(ps.problems);
     return 1;
 }
