@@ -34,92 +34,57 @@ struct faFooter {
     unsigned int isFree; 
 } ;
 
-FD_t faFileno(faFile fa) {
-    return fa->fd;
-}
-
-off_t faLseek(faFile fa, off_t off, int op) {
-    return fdLseek(faFileno(fa), off, op);
-}
-
-static inline ssize_t faRead(faFile fa, /*@out@*/void *buf, size_t count) {
-    return fdRead(faFileno(fa), buf, count);
-}
-
-static inline ssize_t faWrite(faFile fa, const void *buf, size_t count) {
-    return fdWrite(faFileno(fa), buf, count);
-}
-
-int faFcntl(faFile fa, int op, void *lip) {
-    return fcntl(fdFileno(faFileno(fa)), op, lip);
-}
-
 /* flags here is the same as for open(2) - NULL returned on error */
-faFile faOpen(const char * path, int flags, int perms)
+FD_t fadOpen(const char * path, int flags, int perms)
 {
     struct faFileHeader newHdr;
-    struct faFile_s fas;
-    faFile fa;
-    off_t end;
+    FD_t fd;
 
-    if (flags & O_WRONLY) {
+    if (flags & O_WRONLY)
 	return NULL;
-    }
-    if (flags & O_RDWR) {
-	fas.readOnly = 0;
-    } else {
-	fas.readOnly = 1;
-    }
 
-    fas.fd = fdOpen(path, flags, perms);
-    if (fdFileno(fas.fd) < 0) return NULL;
-    fas.firstFree = 0;
-    fas.fileSize = 0;
+    fd = ufdio->_open(path, flags, perms);
+    if (Ferror(fd))
+	/* XXX Fstrerror */
+	return NULL;
+
+    fdSetIo(fd, fadio);
+    fadSetFirstFree(fd, 0);
+    fadSetFileSize(fd, Fseek(fd, 0, SEEK_END));
 
     /* is this file brand new? */
-    end = faLseek(&fas, 0, SEEK_END);
-    if (!end) {
+    if (fadGetFileSize(fd) == 0) {
 	newHdr.magic = FA_MAGIC;
 	newHdr.firstFree = 0;
-	if (faWrite(&fas, &newHdr, sizeof(newHdr)) != sizeof(newHdr)) {
-	    close(fdFileno(fas.fd));
+	if (Fwrite(&newHdr, sizeof(char), sizeof(newHdr), fd) != sizeof(newHdr)) {
+	    Fclose(fd);
 	    return NULL;
 	}
-	fas.firstFree = 0;
-	fas.fileSize = sizeof(newHdr);
+	fadSetFirstFree(fd, 0);
+	fadSetFileSize(fd, sizeof(newHdr));
     } else {
-	(void)faLseek(&fas, 0, SEEK_SET);
-	if (faRead(&fas, &newHdr, sizeof(newHdr)) != sizeof(newHdr)) {
-	    fdClose(fas.fd);
+	if (Pread(fd, &newHdr, sizeof(newHdr), 0) != sizeof(newHdr)) {
+	    Fclose(fd);
 	    return NULL;
 	}
 	if (newHdr.magic != FA_MAGIC) {
-	    fdClose(fas.fd);
+	    Fclose(fd);
 	    return NULL;
 	}
-	fas.firstFree = newHdr.firstFree;
+	fadSetFirstFree(fd, newHdr.firstFree);
+	fadSetFileSize(fd, Fseek(fd, 0, SEEK_END));
 
-	if (faLseek(&fas, 0, SEEK_END) < 0) {
-	    fdClose(fas.fd);
+	if (fadGetFileSize(fd) < 0) {
+	    Fclose(fd);
 	    return NULL;
 	}
-	
-	fas.fileSize = faLseek(&fas, 0, SEEK_CUR);
     }
 
-    /* XXX xmalloc is inappopriate here */
-    if ((fa = malloc(sizeof(*fa))) != NULL) {
-	fa->fd = fas.fd;
-	fa->readOnly = fas.readOnly;
-	fa->firstFree = fas.firstFree;
-	fa->fileSize = fas.fileSize;
-    }
-
-    return fa;
+    /*@-refcounttrans@*/ return fd /*@=refcounttrans@*/ ;
 }
 
 /* returns 0 on failure */
-unsigned int faAlloc(faFile fa, unsigned int size)
+unsigned int fadAlloc(FD_t fd, unsigned int size)
 {
     unsigned int nextFreeBlock;
     unsigned int newBlockOffset;
@@ -148,12 +113,11 @@ unsigned int faAlloc(faFile fa, unsigned int size)
     /* find a block via first fit - see Knuth vol 1 for why */
     /* XXX this could be optimized a bit still */
 
-    nextFreeBlock =  fa->firstFree;
+    nextFreeBlock = fadGetFirstFree(fd);
     newBlockOffset = 0;
 
     while (nextFreeBlock && !newBlockOffset) {
-	if (faLseek(fa, nextFreeBlock, SEEK_SET) < 0) return 0;
-	if (faRead(fa, &header, sizeof(header)) != sizeof(header)) return 0;
+	if (Pread(fd, &header, sizeof(header), nextFreeBlock) != sizeof(header)) return 0;
 
 /* XXX W2DO? exit(EXIT_FAILURE) forces the user to discover rpm --rebuilddb */
 	if (!header.isFree) {
@@ -181,8 +145,7 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 
 	footerOffset = newBlockOffset + header.size - sizeof(footer);
 
-	if (faLseek(fa, footerOffset, SEEK_SET) < 0) return 0;
-	if (faRead(fa, &footer, sizeof(footer)) != sizeof(footer)) 
+	if (Pread(fd, &footer, sizeof(footer), footerOffset) != sizeof(footer)) 
 	    return 0;
 	origFooter = footer;
 
@@ -192,15 +155,14 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 	footer.isFree = header.isFree = 0;
 
 	/* remove it from the free list before */
-	if (newBlockOffset == fa->firstFree) {
+	if (newBlockOffset == fadGetFirstFree(fd)) {
 	    faHeader.magic = FA_MAGIC;
 	    faHeader.firstFree = header.freeNext;
-	    fa->firstFree = header.freeNext;
+	    fadSetFirstFree(fd, header.freeNext);
 	    updateHeader = 1;
 	} else {
-	    if (faLseek(fa, header.freePrev, SEEK_SET) < 0) return 0;
-	    if (faRead(fa, &prevFreeHeader, sizeof(prevFreeHeader)) !=
-			 sizeof(prevFreeHeader)) 
+	    if (Pread(fd, &prevFreeHeader, sizeof(prevFreeHeader),
+			header.freePrev) != sizeof(prevFreeHeader)) 
 		return 0;
 	    origPrevFreeHeader = prevFreeHeader;
 
@@ -209,9 +171,8 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 
 	/* and after */
 	if (header.freeNext) {
-	    if (faLseek(fa, header.freeNext, SEEK_SET) < 0) return 0;
-	    if (faRead(fa, &nextFreeHeader, sizeof(nextFreeHeader)) !=
-			 sizeof(nextFreeHeader)) 
+	    if (Pread(fd, &nextFreeHeader, sizeof(nextFreeHeader),
+			header.freeNext) != sizeof(nextFreeHeader)) 
 		return 0;
 	    origNextFreeHeader = nextFreeHeader;
 
@@ -220,34 +181,26 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 
 	/* if any of these fail, try and restore everything before leaving */
 	if (updateHeader) {
-	    if (faLseek(fa, 0, SEEK_SET) < 0) 
-		return 0;
-	    else if (faWrite(fa, &faHeader, sizeof(faHeader)) !=
+	    if (Pwrite(fd, &faHeader, sizeof(faHeader), 0) !=
 			     sizeof(faHeader)) 
 		return 0;
 	} else {
-	    if (faLseek(fa, header.freePrev, SEEK_SET) < 0)
-		return 0;
-	    else if (faWrite(fa, &prevFreeHeader, sizeof(prevFreeHeader)) !=
-			 sizeof(prevFreeHeader))
+	    if (Pwrite(fd, &prevFreeHeader, sizeof(prevFreeHeader),
+			header.freePrev) != sizeof(prevFreeHeader))
 		return 0;
 	    restorePrevHeader = &origPrevFreeHeader;
 	}
 
 	if (header.freeNext) {
-	    if (faLseek(fa, header.freeNext, SEEK_SET) < 0)
-		return 0;
-	    else if (faWrite(fa, &nextFreeHeader, sizeof(nextFreeHeader)) !=
-			 sizeof(nextFreeHeader))
+	    if (Pwrite(fd, &nextFreeHeader, sizeof(nextFreeHeader),
+			header.freeNext) != sizeof(nextFreeHeader))
 		return 0;
 
 	    restoreNextHeader = &origNextFreeHeader;
 	}
 
 	if (!failed) {
-	    if (faLseek(fa, newBlockOffset, SEEK_SET) < 0) 
-		failed = 1;
-	    else if (faWrite(fa, &header, sizeof(header)) !=
+	    if (Pwrite(fd, &header, sizeof(header), newBlockOffset) !=
 			 sizeof(header)) {
 		failed = 1;
 		restoreHeader = &origHeader;
@@ -255,10 +208,8 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 	}
 
 	if (!failed) {
-	    if (faLseek(fa, footerOffset, SEEK_SET) < 0) 
-		failed = 1;
-	    else if (faWrite(fa, &footer, sizeof(footer)) !=
-			 sizeof(footer)) {
+	    if (Pwrite(fd, &footer, sizeof(footer),
+			footerOffset) != sizeof(footer)) {
 		failed = 1;
 		restoreFooter = &origFooter;
 	    }
@@ -267,30 +218,25 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 	if (failed) {
 	    if (updateHeader) {
 		faHeader.firstFree = newBlockOffset;
-		fa->firstFree = newBlockOffset;
-	    	(void)faLseek(fa, 0, SEEK_SET);
-	        (void)faWrite(fa, &faHeader, sizeof(faHeader));
+		fadSetFirstFree(fd, newBlockOffset);
+	        (void)Pwrite(fd, &faHeader, sizeof(faHeader), 0);
 	    } 
 
-	    if (restorePrevHeader) {
-	    	(void)faLseek(fa, header.freePrev, SEEK_SET);
-	    	(void)faWrite(fa, restorePrevHeader, sizeof(*restorePrevHeader));
-	    }
+	    if (restorePrevHeader)
+	    	(void)Pwrite(fd, restorePrevHeader, sizeof(*restorePrevHeader),
+				header.freePrev);
 
-	    if (restoreNextHeader) {
-	    	(void)faLseek(fa, header.freeNext, SEEK_SET);
-	    	(void)faWrite(fa, restoreNextHeader, sizeof(*restoreNextHeader));
-	    }
+	    if (restoreNextHeader)
+	    	(void)Pwrite(fd, restoreNextHeader, sizeof(*restoreNextHeader),
+				header.freeNext);
 
-	    if (restoreHeader) {
-	    	(void)faLseek(fa, newBlockOffset, SEEK_SET);
-	    	(void)faWrite(fa, restoreHeader, sizeof(header));
-	    }
+	    if (restoreHeader)
+	    	(void)Pwrite(fd, restoreHeader, sizeof(header),
+				newBlockOffset);
 
-	    if (restoreFooter) {
-	    	(void)faLseek(fa, footerOffset, SEEK_SET);
-	    	(void)faWrite(fa, restoreFooter, sizeof(footer));
-	    }
+	    if (restoreFooter)
+	    	(void)Pwrite(fd, restoreFooter, sizeof(footer),
+				footerOffset);
 
 	    return 0;
 	}
@@ -298,39 +244,35 @@ unsigned int faAlloc(faFile fa, unsigned int size)
 	char * space;
 
 	/* make a new block */
-	newBlockOffset = fa->fileSize;
+	newBlockOffset = fadGetFileSize(fd);
 	footerOffset = newBlockOffset + size - sizeof(footer);
 
 	space = alloca(size);
-	if (!space) return 0;
+	if (space == NULL) return 0;
+	memset(space, 0, size);
 
 	footer.isFree = header.isFree = 0;
 	footer.size = header.size = size;
 	header.freePrev = header.freeNext = 0;
 
 	/* reserve all space up front */
-        (void)faLseek(fa, newBlockOffset, SEEK_SET);
-	if (faWrite(fa, space, size) != size) {
+	/* XXX TODO: check max. no. of bytes to write */
+	if (Pwrite(fd, space, size, newBlockOffset) != size)
 	    return 0;
-	}
 
-        (void)faLseek(fa, newBlockOffset, SEEK_SET);
-	if (faWrite(fa, &header, sizeof(header)) != sizeof(header)) {
+	if (Pwrite(fd, &header, sizeof(header), newBlockOffset) != sizeof(header))
 	    return 0;
-	}
 
-        (void)faLseek(fa, footerOffset, SEEK_SET);
-	if (faWrite(fa, &footer, sizeof(footer)) != sizeof(footer)) {
+	if (Pwrite(fd, &footer, sizeof(footer), footerOffset) != sizeof(footer))
 	    return 0;
-	}
 
-	fa->fileSize += size;
+	fadSetFileSize(fd, fadGetFileSize(fd) + size);
     }
     
     return newBlockOffset + sizeof(header); 
 }
 
-void faFree(faFile fa, unsigned int offset)
+void fadFree(FD_t fd, unsigned int offset)
 {
     struct faHeader header;
     struct faFooter footer;
@@ -345,42 +287,38 @@ void faFree(faFile fa, unsigned int offset)
     offset -= sizeof(header);
 
     /* find out where in the (sorted) free list to put this */
-    prevFreeOffset = fa->firstFree;
+    prevFreeOffset = fadGetFirstFree(fd);
 
     if (!prevFreeOffset || (prevFreeOffset > offset)) {
-	nextFreeOffset = fa->firstFree;
+	nextFreeOffset = fadGetFirstFree(fd);
 	prevFreeOffset = 0;
     } else {
-	if (faLseek(fa, prevFreeOffset, SEEK_SET) < 0) return;
-	if (faRead(fa, &prevFreeHeader, sizeof(prevFreeHeader)) != 
-		sizeof(prevFreeHeader)) return;
+	if (Pread(fd, &prevFreeHeader, sizeof(prevFreeHeader),
+			prevFreeOffset) != sizeof(prevFreeHeader))
+	    return;
 
 	while (prevFreeHeader.freeNext && prevFreeHeader.freeNext < offset) {
 	    prevFreeOffset = prevFreeHeader.freeNext;
-	    if (faLseek(fa, prevFreeOffset, SEEK_SET) < 0) return;
-	    if (faRead(fa, &prevFreeHeader, sizeof(prevFreeHeader)) != 
-		    sizeof(prevFreeHeader)) return;
+	    if (Pread(fd, &prevFreeHeader, sizeof(prevFreeHeader),
+			prevFreeOffset) != sizeof(prevFreeHeader))
+		return;
 	} 
 
 	nextFreeOffset = prevFreeHeader.freeNext;
     }
 
     if (nextFreeOffset) {
-	if (faLseek(fa, nextFreeOffset, SEEK_SET) < 0) return;
-	if (faRead(fa, &nextFreeHeader, sizeof(nextFreeHeader)) != 
-		sizeof(nextFreeHeader)) return;
+	if (Pread(fd, &nextFreeHeader, sizeof(nextFreeHeader),
+			nextFreeOffset) != sizeof(nextFreeHeader))
+	    return;
     }
 
-    if (faLseek(fa, offset, SEEK_SET) < 0)
-	return;
-    if (faRead(fa, &header, sizeof(header)) != sizeof(header))
+    if (Pread(fd, &header, sizeof(header), offset) != sizeof(header))
 	return;
 
     footerOffset = offset + header.size - sizeof(footer);
 
-    if (faLseek(fa, footerOffset, SEEK_SET) < 0)
-	return;
-    if (faRead(fa, &footer, sizeof(footer)) != sizeof(footer))
+    if (Pread(fd, &footer, sizeof(footer), footerOffset) != sizeof(footer))
 	return;
 
     header.isFree = 1;
@@ -388,74 +326,68 @@ void faFree(faFile fa, unsigned int offset)
     header.freePrev = prevFreeOffset;
     footer.isFree = 1;
 
-    (void)faLseek(fa, offset, SEEK_SET);
-    (void)faWrite(fa, &header, sizeof(header));
+    /* XXX TODO: set max. no. of bytes to write */
+    (void)Pwrite(fd, &header, sizeof(header), offset);
 
-    (void)faLseek(fa, footerOffset, SEEK_SET);
-    (void)faWrite(fa, &footer, sizeof(footer));
+    (void)Pwrite(fd, &footer, sizeof(footer), footerOffset);
 
     if (nextFreeOffset) {
 	nextFreeHeader.freePrev = offset;
-	if (faLseek(fa, nextFreeOffset, SEEK_SET) < 0) return;
-	if (faWrite(fa, &nextFreeHeader, sizeof(nextFreeHeader)) != 
-		sizeof(nextFreeHeader)) return;
+	if (Pwrite(fd, &nextFreeHeader, sizeof(nextFreeHeader),
+			nextFreeOffset) != sizeof(nextFreeHeader))
+	    return;
     }
 
     if (prevFreeOffset) {
 	prevFreeHeader.freeNext = offset;
-	if (faLseek(fa, prevFreeOffset, SEEK_SET) < 0) return;
-	(void)faWrite(fa, &prevFreeHeader, sizeof(prevFreeHeader));
+	if (Pwrite(fd, &prevFreeHeader, sizeof(prevFreeHeader),
+			prevFreeOffset) != sizeof(prevFreeHeader))
+	    return;
     } else {
-	fa->firstFree = offset;
+	fadSetFirstFree(fd, offset);
 
 	faHeader.magic = FA_MAGIC;
-	faHeader.firstFree = fa->firstFree;
+	faHeader.firstFree = fadGetFirstFree(fd);
 
-	if (faLseek(fa, 0, SEEK_SET) < 0) return;
-	(void)faWrite(fa, &faHeader, sizeof(faHeader));
+	/* XXX TODO: set max. no. of bytes to write */
+	if (Pwrite(fd, &faHeader, sizeof(faHeader), 0) != sizeof(faHeader))
+	    return;
     }
 }
 
-void faClose(faFile fa) {
-    fdClose(fa->fd);
-    free(fa);
+int fadFirstOffset(FD_t fd)
+{
+    return fadNextOffset(fd, 0);
 }
 
-int faFirstOffset(faFile fa) {
-    return faNextOffset(fa, 0);
-}
-
-int faNextOffset(faFile fa, unsigned int lastOffset)
+int fadNextOffset(FD_t fd, unsigned int lastOffset)
 {
     struct faHeader header;
     int offset;
 
-    if (lastOffset) {
-	offset = lastOffset - sizeof(header);
-    } else {
-	offset = sizeof(struct faFileHeader);
-    }
+    offset = (lastOffset)
+	? (lastOffset - sizeof(header))
+	: sizeof(struct faFileHeader);
 
-    if (offset >= fa->fileSize) return 0;
-
-    (void)faLseek(fa, offset, SEEK_SET);
-    if (faRead(fa, &header, sizeof(header)) != sizeof(header)) {
+    if (offset >= fadGetFileSize(fd))
 	return 0;
-    }
-    if (!lastOffset && !header.isFree) return (offset + sizeof(header));
+
+    if (Pread(fd, &header, sizeof(header), offset) != sizeof(header))
+	return 0;
+
+    if (!lastOffset && !header.isFree)
+	return (offset + sizeof(header));
 
     do {
 	offset += header.size;
 
-	(void)faLseek(fa, offset, SEEK_SET);
-	if (faRead(fa, &header, sizeof(header)) != sizeof(header)) {
+	if (Pread(fd, &header, sizeof(header), offset) != sizeof(header))
 	    return 0;
-	}
 
 	if (!header.isFree) break;
-    } while (offset < fa->fileSize && header.isFree);
+    } while (offset < fadGetFileSize(fd) && header.isFree);
 
-    if (offset < fa->fileSize) {
+    if (offset < fadGetFileSize(fd)) {
 	/* Sanity check this to make sure we're not going in loops */
 	offset += sizeof(header);
 
