@@ -1,6 +1,8 @@
 #include "system.h"
 
 #include <rpmlib.h>
+#include <rpmmacro.h>	/* XXX for %_i18ndomains */
+#include "misc.h"
 
 static char * permsString(int mode)
 {
@@ -10,6 +12,7 @@ static char * permsString(int mode)
    
     if (mode & S_ISVTX) perms[10] = 't';
 
+    /*@-unrecog@*/
     if (mode & S_IRUSR) perms[1] = 'r';
     if (mode & S_IWUSR) perms[2] = 'w';
     if (mode & S_IXUSR) perms[3] = 'x';
@@ -21,6 +24,7 @@ static char * permsString(int mode)
     if (mode & S_IROTH) perms[7] = 'r';
     if (mode & S_IWOTH) perms[8] = 'w';
     if (mode & S_IXOTH) perms[9] = 'x';
+    /*@=unrecog@*/
 
     if (mode & S_ISUID) {
 	if (mode & S_IXUSR) 
@@ -55,15 +59,14 @@ static char * permsString(int mode)
 }
 
 static char * triggertypeFormat(int_32 type, const void * data, 
-	/*@unused@*/char * formatPrefix, /*@unused@*/int padding,
-	/*@unused@*/int element)
+	/*@unused@*/ char * formatPrefix, /*@unused@*/ int padding,
+	/*@unused@*/ int element)
 {
     const int_32 * item = data;
     char * val;
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else if (*item & RPMSENSE_TRIGGERIN) {
 	val = xstrdup("in");
     } else {
@@ -74,14 +77,13 @@ static char * triggertypeFormat(int_32 type, const void * data,
 }
 
 static char * permsFormat(int_32 type, const void * data, 
-		char * formatPrefix, int padding, /*@unused@*/int element)
+		char * formatPrefix, int padding, /*@unused@*/ int element)
 {
     char * val;
     char * buf;
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	val = xmalloc(15 + padding);
 	strcat(formatPrefix, "s");
@@ -94,15 +96,14 @@ static char * permsFormat(int_32 type, const void * data,
 }
 
 static char * fflagsFormat(int_32 type, const void * data, 
-		char * formatPrefix, int padding, /*@unused@*/int element)
+		char * formatPrefix, int padding, /*@unused@*/ int element)
 {
     char * val;
     char buf[15];
     int anint = *((int_32 *) data);
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	buf[0] = '\0';
 	if (anint & RPMFILE_DOC)
@@ -127,15 +128,14 @@ static char * fflagsFormat(int_32 type, const void * data,
 }
 
 static char * depflagsFormat(int_32 type, const void * data, 
-		char * formatPrefix, int padding, /*@unused@*/int element)
+		char * formatPrefix, int padding, /*@unused@*/ int element)
 {
     char * val;
     char buf[10];
     int anint = *((int_32 *) data);
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	buf[0] = '\0';
 
@@ -154,8 +154,9 @@ static char * depflagsFormat(int_32 type, const void * data,
     return val;
 }
 
-static int fsnamesTag( /*@unused@*/ Header h, int_32 * type, void ** data,
-	int_32 * count, int * freeData)
+static int fsnamesTag( /*@unused@*/ Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
 {
     const char ** list;
 
@@ -171,17 +172,18 @@ static int fsnamesTag( /*@unused@*/ Header h, int_32 * type, void ** data,
     return 0; 
 }
 
-static int instprefixTag(Header h, int_32 * type, void ** data, int_32 * count,
-		      int * freeData)
+static int instprefixTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
 {
     char ** array;
 
-    if (headerGetEntry(h, RPMTAG_INSTALLPREFIX, type, data, count)) {
+    if (headerGetEntry(h, RPMTAG_INSTALLPREFIX, type, (void **)data, count)) {
 	*freeData = 0;
 	return 0;
     } else if (headerGetEntry(h, RPMTAG_INSTPREFIXES, NULL, (void **) &array, 
 			      count)) {
-	*((char **) data) = xstrdup(array[0]);
+	*data = xstrdup(array[0]);
 	*freeData = 1;
 	*type = RPM_STRING_TYPE;
 	free(array);
@@ -191,21 +193,22 @@ static int instprefixTag(Header h, int_32 * type, void ** data, int_32 * count,
     return 1;
 }
 
-static int fssizesTag(Header h, int_32 * type, void ** data, int_32 * count,
-		      int * freeData)
+static int fssizesTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
 {
     const char ** filenames;
     int_32 * filesizes;
     uint_32 * usages;
     int numFiles;
 
-    if (!headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &filenames, NULL)) 
-	filenames = NULL;
-
     if (!headerGetEntry(h, RPMTAG_FILESIZES, NULL, (void **) &filesizes, 
 		       &numFiles)) {
 	filesizes = NULL;
 	numFiles = 0;
+	filenames = NULL;
+    } else {
+	rpmBuildFileList(h, &filenames, &numFiles);
     }
 
     if (rpmGetFilesystemList(NULL, count)) {
@@ -227,11 +230,14 @@ static int fssizesTag(Header h, int_32 * type, void ** data, int_32 * count,
 
     *data = usages;
 
+    if (filenames) free(filenames);
+
     return 0;
 }
 
-static int triggercondsTag(Header h, /*@out@*/int_32 * type, /*@out@*/void ** data, 
-			   /*@out@*/int_32 * count, /*@out@*/int * freeData)
+static int triggercondsTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
 {
     int_32 * indices, * flags;
     char ** names, ** versions;
@@ -290,8 +296,9 @@ static int triggercondsTag(Header h, /*@out@*/int_32 * type, /*@out@*/void ** da
     return 0;
 }
 
-static int triggertypeTag(Header h, int_32 * type, /*@out@*/void ** data, 
-			   int_32 * count, int * freeData)
+static int triggertypeTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
 {
     int_32 * indices, * flags;
     char ** conds, ** s;
@@ -330,7 +337,127 @@ static int triggertypeTag(Header h, int_32 * type, /*@out@*/void ** data,
     return 0;
 }
 
+static int filenamesTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
+{
+    *type = RPM_STRING_ARRAY_TYPE;
+
+    rpmBuildFileList(h, (const char ***) data, count);
+    *freeData = 1;
+
+    *freeData = 0;	/* XXX WTFO? */
+
+    return 0; 
+}
+
+/* I18N look aside diversions */
+
+int _nl_msg_cat_cntr;	/* XXX GNU gettext voodoo */
+static const char * language = "LANGUAGE";
+
+static char * _macro_i18ndomains = "%{?_i18ndomains:%{_i18ndomains}}";
+
+static int i18nTag(Header h, int_32 tag, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
+{
+    char * dstring = rpmExpand(_macro_i18ndomains, NULL);
+    int rc;
+
+    *type = RPM_STRING_TYPE;
+    *data = NULL;
+    *count = 0;
+    *freeData = 0;
+
+    if (dstring && *dstring) {
+	char *domain, *de;
+	const char * langval;
+	const char * msgkey;
+	const char * msgid;
+
+	{   const char * tn = tagName(tag);
+	    const char * n;
+	    char * mk;
+	    headerNVR(h, &n, NULL, NULL);
+	    mk = alloca(strlen(n) + strlen(tn) + sizeof("()"));
+	    sprintf(mk, "%s(%s)", n, tn);
+	    msgkey = mk;
+	}
+
+	/* change to en_US for msgkey -> msgid resolution */
+	langval = getenv(language);
+	setenv(language, "en_US", 1);
+	++_nl_msg_cat_cntr;
+
+	msgid = NULL;
+	for (domain = dstring; domain != NULL; domain = de) {
+	    de = strchr(domain, ':');
+	    if (de) *de++ = '\0';
+	    msgid = /*@-unrecog@*/ dgettext(domain, msgkey) /*@=unrecog@*/;
+	    if (msgid != msgkey) break;
+	}
+
+	/* restore previous environment for msgid -> msgstr resolution */
+	if (langval)
+	    setenv(language, langval, 1);
+	else
+	    unsetenv(language);
+	++_nl_msg_cat_cntr;
+
+	if (domain && msgid) {
+	    *data = xstrdup(/*@-unrecog@*/ dgettext(domain, msgid) /*@=unrecog@*/);
+	    *count = 1;
+	    *freeData = 1;
+	}
+	xfree(dstring); dstring = NULL;
+	if (*data) {
+	    return 0;
+	}
+    }
+
+    if (dstring) xfree(dstring);
+
+    rc = headerGetEntry(h, tag, type, (void **)data, count);
+
+    if (rc) {
+	*data = xstrdup(*data);
+	*freeData = 1;
+	return 0;
+    }
+
+    *freeData = 0;
+    *data = NULL;
+    *count = 0;
+    return 1;
+}
+
+static int summaryTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
+{
+    return i18nTag(h, RPMTAG_SUMMARY, type, data, count, freeData);
+}
+
+static int descriptionTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
+{
+    return i18nTag(h, RPMTAG_DESCRIPTION, type, data, count, freeData);
+}
+
+static int groupTag(Header h, /*@out@*/ int_32 * type,
+	/*@out@*/ const void ** data, /*@out@*/ int_32 * count,
+	/*@out@*/ int * freeData)
+{
+    return i18nTag(h, RPMTAG_GROUP, type, data, count, freeData);
+}
+
 const struct headerSprintfExtension rpmHeaderFormats[] = {
+    { HEADER_EXT_TAG, "RPMTAG_GROUP", { groupTag } },
+    { HEADER_EXT_TAG, "RPMTAG_DESCRIPTION", { descriptionTag } },
+    { HEADER_EXT_TAG, "RPMTAG_SUMMARY", { summaryTag } },
+    { HEADER_EXT_TAG, "RPMTAG_FILENAMES", { filenamesTag } },
     { HEADER_EXT_TAG, "RPMTAG_FSSIZES", { fssizesTag } },
     { HEADER_EXT_TAG, "RPMTAG_FSNAMES", { fsnamesTag } },
     { HEADER_EXT_TAG, "RPMTAG_INSTALLPREFIX", { instprefixTag } },
