@@ -14,6 +14,7 @@
 #define	ntohs(_x)	(_x)
 #define	htonl(_x)	(_x)
 #define	htons(_x)	(_x)
+/*@-observertrans@*/	/* XXX FIXME */
 #else
 #include <netinet/in.h>
 #endif	/* __LCLINT__ */
@@ -48,7 +49,7 @@ struct headerToken {
     int indexAlloced;
 
     int sorted;  
-    /*@refs@*/int usageCount;
+    /*@refs@*/ int usageCount;
 };
 
 struct entryInfo {
@@ -78,35 +79,85 @@ struct sprintfTag {
 };
 
 struct extensionCache {
-    int_32 type, count;
-    int avail, freeit;
-    void * data;
+    int_32 type;
+    int_32 count;
+    int avail;
+    int freeit;
+    const void * data;
 };
 
 struct sprintfToken {
     enum { PTOK_NONE = 0, PTOK_TAG, PTOK_ARRAY, PTOK_STRING, PTOK_COND } type;
     union {
 	struct {
-	    struct sprintfToken * format;
+	    /*@only@*/ struct sprintfToken * format;
 	    int numTokens;
 	} array;
 	struct sprintfTag tag;
 	struct {
-	    char * string;
+	    /*@dependent@*/ char * string;
 	    int len;
 	} string;
 	struct {
-	    struct sprintfToken * ifFormat;
+	    /*@only@*/ struct sprintfToken * ifFormat;
 	    int numIfTokens;
-	    struct sprintfToken * elseFormat;
+	    /*@only@*/ struct sprintfToken * elseFormat;
 	    int numElseTokens;
 	    struct sprintfTag tag;
 	} cond;
     } u;
 };
 
-static void copyEntry(struct indexEntry * entry, /*@out@*/int_32 *type,
-	/*@out@*/void **p, /*@out@*/int_32 *c, int minimizeMemory)
+#if HAVE_MCHECK_H
+static int probe_headers = 0;
+#define	HEADERPROBE(_h, _msg)	if (probe_headers) headerProbe((_h), (_msg))
+
+static void headerProbeAddr(Header h, const char * msg,
+	void * p, const char * imsg)
+{
+    const char * mchkstr = NULL;
+    switch (mprobe(p)) {
+    case MCHECK_DISABLED:
+    case MCHECK_OK:
+	return;
+	/*@notreached@*/ break;
+    case MCHECK_HEAD:
+	mchkstr = "HEAD";
+	break;
+    case MCHECK_TAIL:
+	mchkstr = "TAIL";
+	break;
+    case MCHECK_FREE:
+	mchkstr = "FREE";
+	break;
+    }
+    fprintf(stderr, "*** MCHECK_%s h %p", mchkstr, h);
+    if (imsg && p)
+	fprintf(stderr, " %s %p", imsg, p);
+    if (msg)
+	fprintf(stderr, " %s", msg);
+    fprintf(stderr, "\n");
+}
+
+static void headerProbe(Header h, const char *msg)
+{
+    char imsg[256];
+    int i;
+
+    headerProbeAddr(h, msg, h, "header");
+    sprintf(imsg, "index (used %d)", h->indexUsed);
+    headerProbeAddr(h, msg, h->index, imsg);
+    for (i = 0; i < h->indexUsed; i++) {
+	sprintf(imsg, "index[%d:%d].data", i, h->indexUsed);
+	headerProbeAddr(h, msg, h->index[i].data, imsg);
+    }
+}
+#else	/* HAVE_MCHECK_H */
+#define	HEADERPROBE(_h, _msg)
+#endif	/* HAVE_MCHECK_H */
+
+static void copyEntry(struct indexEntry * entry, /*@out@*/ int_32 * type,
+	/*@out@*/ void ** p, /*@out@*/ int_32 * c, int minimizeMemory)
 {
     int i, tableSize;
     char ** ptrEntry;
@@ -482,20 +533,20 @@ int headerWrite(FD_t fd, Header h, int magicp)
     p = doHeaderUnload(h, &length);
 
     if (magicp) {
-	nb = fdWrite(fd, header_magic, sizeof(header_magic));
+	nb = Fwrite(header_magic, sizeof(char), sizeof(header_magic), fd);
 	if (nb != sizeof(header_magic)) {
 	    free(p);
 	    return 1;
 	}
 	l = htonl(0);
-	nb = fdWrite(fd, &l, sizeof(l));
+	nb = Fwrite(&l, sizeof(char), sizeof(l), fd);
 	if (nb != sizeof(l)) {
 	    free(p);
 	    return 1;
 	}
     }
     
-    nb = fdWrite(fd, p, length);
+    nb = Fwrite(p, sizeof(char), length, fd);
     if (nb != length) {
 	free(p);
 	return 1;
@@ -521,7 +572,7 @@ Header headerRead(FD_t fd, int magicp)
     if (magicp == HEADER_MAGIC_YES)
 	i += 2;
 
-    if (timedRead(fd, block, i * sizeof(*block)) != (i * sizeof(*block)))
+    if (timedRead(fd, (char *)block, i * sizeof(*block)) != (i * sizeof(*block)))
 	return NULL;
     i = 0;
 
@@ -551,7 +602,7 @@ Header headerRead(FD_t fd, int magicp)
     *p++ = htonl(dl);
 
     totalSize -= sizeof(int_32) + sizeof(int_32);
-    if (timedRead(fd, p, totalSize) != totalSize)
+    if (timedRead(fd, (char *)p, totalSize) != totalSize)
 	return NULL;
     
     h = headerLoad(dataBlock);
@@ -571,20 +622,20 @@ int headerGzWrite(FD_t fd, Header h, int magicp)
     p = doHeaderUnload(h, &length);
 
     if (magicp) {
-	nb = gzdWrite(fd, header_magic, sizeof(header_magic));
+	nb = Fwrite(header_magic, sizeof(char), sizeof(header_magic), fd);
 	if (nb != sizeof(header_magic)) {
 	    free(p);
 	    return 1;
 	}
 	l = htonl(0);
-	nb = gzdWrite(fd, &l, sizeof(l));
+	nb = Fwrite(&l, sizeof(char), sizeof(l), fd);
 	if (nb != sizeof(l)) {
 	    free(p);
 	    return 1;
 	}
     }
     
-    nb = gzdWrite(fd, p, length);
+    nb = Fwrite(p, sizeof(char), length, fd);
     if (nb != length) {
 	free(p);
 	return 1;
@@ -605,24 +656,24 @@ Header headerGzRead(FD_t fd, int magicp)
     int totalSize;
 
     if (magicp == HEADER_MAGIC_YES) {
-	if (gzdRead(fd, &magic, sizeof(magic)) != sizeof(magic))
+	if (Fread(&magic, sizeof(char), sizeof(magic), fd) != sizeof(magic))
 	    return NULL;
 	if (memcmp(&magic, header_magic, sizeof(magic))) {
 	    return NULL;
 	}
 
-	if (gzdRead(fd, &reserved, sizeof(reserved)) != sizeof(reserved))
+	if (Fread(&reserved, sizeof(char), sizeof(reserved), fd) != sizeof(reserved))
 	    return NULL;
     }
     
     /* First read the index length (count of index entries) */
-    if (gzdRead(fd, &il, sizeof(il)) != sizeof(il)) 
+    if (Fread(&il, sizeof(char), sizeof(il), fd) != sizeof(il)) 
 	return NULL;
 
     il = ntohl(il);
 
     /* Then read the data length (number of bytes) */
-    if (gzdRead(fd, &dl, sizeof(dl)) != sizeof(dl)) 
+    if (Fread(&dl, sizeof(char), sizeof(dl), fd) != sizeof(dl)) 
 	return NULL;
 
     dl = ntohl(dl);
@@ -635,7 +686,7 @@ Header headerGzRead(FD_t fd, int magicp)
     *p++ = htonl(dl);
 
     totalSize -= sizeof(int_32) + sizeof(int_32);
-    if (gzdRead(fd, p, totalSize) != totalSize)
+    if (Fread(p, sizeof(char), totalSize, fd) != totalSize)
 	return NULL;
     
     h = headerLoad(block);
@@ -760,7 +811,7 @@ void headerDump(Header h, FILE *f, int flags,
 		}
 		break;
 	    default:
-		fprintf(stderr, _("Data type %d not supprted\n"), 
+		fprintf(stderr, _("Data type %d not supported\n"), 
 			(int) p->info.type);
 		exit(EXIT_FAILURE);
 		/*@notreached@*/ break;
@@ -907,7 +958,8 @@ static int headerMatchLocale(const char *td, const char *l, const char *le)
     return 0;
 }
 
-static char *headerFindI18NString(Header h, struct indexEntry *entry)
+/*@dependent@*/ static char *
+headerFindI18NString(Header h, struct indexEntry *entry)
 {
     const char *lang, *l, *le;
     struct indexEntry * table;
@@ -948,12 +1000,13 @@ static char *headerFindI18NString(Header h, struct indexEntry *entry)
     return entry->data;
 }
 
-static int intGetEntry(Header h, int_32 tag, /*@out@*/int_32 *type, /*@out@*/void **p, /*@out@*/int_32 *c,
-		       int minMem)
+static int intGetEntry(Header h, int_32 tag, /*@out@*/ int_32 *type,
+	/*@out@*/ void **p, /*@out@*/ int_32 *c, int minMem)
 {
     struct indexEntry * entry;
     char * chptr;
 
+    HEADERPROBE(h, "intGetEntry");
     /* First find the tag */
     entry = findEntry(h, tag, RPM_NULL_TYPE);
     if (!entry) {
@@ -1016,11 +1069,12 @@ void headerFree(Header h)
 	free(h->index[i].data);
 
     free(h->index);
-    free(h);
+    /*@-refcounttrans@*/ free(h); /*@=refcounttrans@*/
 }
 
 Header headerLink(Header h)
 {
+    HEADERPROBE(h, "headerLink");
     h->usageCount++;
     return h;
 }
@@ -1089,7 +1143,7 @@ static void copyData(int_32 type, /*@out@*/void * dstPtr, const void * srcPtr, i
     }
 }
 
-static void * grabData(int_32 type, /*@out@*/const void * p, int_32 c, int * lengthPtr)
+static void * grabData(int_32 type, const void * p, int_32 c, int * lengthPtr)
 {
     int length;
     void * data;
@@ -1162,12 +1216,11 @@ headerGetLangs(Header h)
     return table;
 }
 
-int headerAddI18NString(Header h, int_32 tag, char * string, char * lang)
+int headerAddI18NString(Header h, int_32 tag, const char * string, const char * lang)
 {
     struct indexEntry * table, * entry;
-    char * charArray[2];
     char * chptr;
-    char ** strArray;
+    const char ** strArray;
     int length;
     int ghosts;
     int i, langNum;
@@ -1181,16 +1234,17 @@ int headerAddI18NString(Header h, int_32 tag, char * string, char * lang)
     }
 
     if (!table && !entry) {
+	const char * charArray[2];
+	int count = 0;
 	if (!lang || (lang[0] == 'C' && lang[1] == '\0')) {
-	    charArray[0] = "C";
-	    if (!headerAddEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE, 
-				&charArray, 1)) return 0;
+	    charArray[count++] = "C";
 	} else {
-	    charArray[0] = "C";
-	    charArray[1] = lang;
-	    if (!headerAddEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE, 
-				&charArray, 2)) return 0;
+	    charArray[count++] = "C";
+	    charArray[count++] = lang;
 	}
+	if (!headerAddEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE, 
+			&charArray, count))
+	    return 0;
 	table = findEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE);
     }
 
@@ -1386,13 +1440,21 @@ static void freeFormat( /*@only@*/ struct sprintfToken * format, int num)
     int i;
 
     for (i = 0; i < num; i++) {
-	if (format[i].type == PTOK_ARRAY) 
+	switch (format[i].type) {
+	case PTOK_ARRAY:
 	    freeFormat(format[i].u.array.format, format[i].u.array.numTokens);
-	if (format[i].type == PTOK_COND) { 
+	    break;
+	case PTOK_COND:
 	    freeFormat(format[i].u.cond.ifFormat, 
 			format[i].u.cond.numIfTokens);
 	    freeFormat(format[i].u.cond.elseFormat, 
 			format[i].u.cond.numElseTokens);
+	    break;
+	case PTOK_NONE:
+	case PTOK_TAG:
+	case PTOK_STRING:
+	default:
+	    break;
 	}
     }
     free(format);
@@ -1406,7 +1468,6 @@ static void findTag(char * name, const struct headerTagTableEntry * tags,
     const struct headerTagTableEntry * entry;
     const struct headerSprintfExtension * ext;
     char * tagname;
-    int i;
 
     *tagMatch = NULL;
     *extMatch = NULL;
@@ -1419,29 +1480,30 @@ static void findTag(char * name, const struct headerTagTableEntry * tags,
 	tagname = name;
     }
 
+    /* Search extensions first to permit overriding header tags. */
+    ext = extensions;
+    while (ext->type != HEADER_EXT_LAST) {
+	if (ext->type == HEADER_EXT_TAG && !strcasecmp(ext->name, tagname))
+	    break;
+
+	if (ext->type == HEADER_EXT_MORE)
+	    ext = ext->u.more;
+	else
+	    ext++;
+    }
+
+    if (ext->type == HEADER_EXT_TAG) {
+	*extMatch = ext;
+	return;
+    }
+
+    /* Search header tags. */
     for (entry = tags; entry->name; entry++)
 	if (!strcasecmp(entry->name, tagname)) break;
 
     if (entry->name) {
 	*tagMatch = entry;
-    } else {
-	ext = extensions, i =0;
-	while (ext->type != HEADER_EXT_LAST) {
-	    if (ext->type == HEADER_EXT_TAG && 
-		!strcasecmp(ext->name, tagname)) {
-		break;
-	    }
-
-	    if (ext->type == HEADER_EXT_MORE)
-		ext = ext->u.more;
-	    else
-		ext++;
-	    i++;
-	}
-
-	if (ext->type == HEADER_EXT_TAG) {
-	    *extMatch = ext;
-	}
+	return;
     }
 }
 
@@ -1460,7 +1522,7 @@ static int parseFormat(char * str, const struct headerTagTableEntry * tags,
     struct sprintfToken * format;
     int numTokens;
     int currToken;
-    const struct headerTagTableEntry * entry;
+    const struct headerTagTableEntry * tag;
     const struct headerSprintfExtension * ext;
     int i;
     int done = 0;
@@ -1573,11 +1635,11 @@ static int parseFormat(char * str, const struct headerTagTableEntry * tags,
 		}
 
 		i = 0;
-		findTag(start, tags, extensions, &entry, &ext);
+		findTag(start, tags, extensions, &tag, &ext);
 
-		if (entry) {
+		if (tag) {
 		    format[currToken].u.tag.ext = NULL;
-		    format[currToken].u.tag.tag = entry->val;
+		    format[currToken].u.tag.tag = tag->val;
 		} else if (ext) {
 		    format[currToken].u.tag.ext = ext->u.tagFunction;
 		    format[currToken].u.tag.extNum = ext - extensions;
@@ -1669,12 +1731,13 @@ static int parseFormat(char * str, const struct headerTagTableEntry * tags,
 static int parseExpression(struct sprintfToken * token, char * str, 
 	const struct headerTagTableEntry * tags, 
 	const struct headerSprintfExtension * extensions,
-	/*@out@*/char ** endPtr, /*@out@*/const char ** error)
+	/*@out@*/ char ** endPtr, /*@out@*/ const char ** error)
 {
     char * chptr, * end;
     const struct headerTagTableEntry * tag;
     const struct headerSprintfExtension * ext;
 
+    *error = NULL;
     chptr = str;
     while (*chptr && *chptr != '?') chptr++;
 
@@ -1698,6 +1761,7 @@ static int parseExpression(struct sprintfToken * token, char * str,
     if (!*end) {
 	*error = _("} expected in expression");
 	freeFormat(token->u.cond.ifFormat, token->u.cond.numIfTokens);
+	token->u.cond.ifFormat = NULL;
 	return 1;
     }
 
@@ -1705,6 +1769,7 @@ static int parseExpression(struct sprintfToken * token, char * str,
     if (*chptr != ':' && *chptr != '|') {
 	*error = _(": expected following ? subexpression");
 	freeFormat(token->u.cond.ifFormat, token->u.cond.numIfTokens);
+	token->u.cond.ifFormat = NULL;
 	return 1;
     }
 
@@ -1718,6 +1783,7 @@ static int parseExpression(struct sprintfToken * token, char * str,
 	if (*chptr != '{') {
 	    *error = _("{ expected after : in expression");
 	    freeFormat(token->u.cond.ifFormat, token->u.cond.numIfTokens);
+	    token->u.cond.ifFormat = NULL;
 	    return 1;
 	}
 
@@ -1730,6 +1796,7 @@ static int parseExpression(struct sprintfToken * token, char * str,
 	if (!*end) {
 	    *error = _("} expected in expression");
 	    freeFormat(token->u.cond.ifFormat, token->u.cond.numIfTokens);
+	    token->u.cond.ifFormat = NULL;
 	    return 1;
 	}
 
@@ -1737,7 +1804,9 @@ static int parseExpression(struct sprintfToken * token, char * str,
 	if (*chptr != '|') {
 	    *error = _("| expected at end of expression");
 	    freeFormat(token->u.cond.ifFormat, token->u.cond.numIfTokens);
+	    token->u.cond.ifFormat = NULL;
 	    freeFormat(token->u.cond.elseFormat, token->u.cond.numElseTokens);
+	    token->u.cond.elseFormat = NULL;
 	    return 1;
 	}
     }
@@ -1764,9 +1833,9 @@ static int parseExpression(struct sprintfToken * token, char * str,
     return 0;
 }
 
-static int getExtension(Header h, headerTagTagFunction fn, /*@out@*/int_32 * typeptr,
-			/*@out@*/void ** data, /*@out@*/int_32 * countptr, 
-			struct extensionCache * ext)
+static int getExtension(Header h, headerTagTagFunction fn,
+	/*@out@*/ int_32 * typeptr, /*@out@*/ const void ** data,
+	/*@out@*/ int_32 * countptr, struct extensionCache * ext)
 {
     if (!ext->avail) {
 	if (fn(h, &ext->type, &ext->data, &ext->count, &ext->freeit))
@@ -1788,10 +1857,10 @@ static char * formatValue(struct sprintfTag * tag, Header h,
     int len;
     char buf[20];
     int_32 count, type;
-    void * data;
+    const void * data;
     unsigned int intVal;
     char * val = NULL;
-    char ** strarray;
+    const char ** strarray;
     int mayfree = 0;
     int countBuf;
     headerTagFormatFunction tagtype = NULL;
@@ -1805,7 +1874,7 @@ static char * formatValue(struct sprintfTag * tag, Header h,
 	    data = "(none)";		/* XXX i18n? NO!, sez; gafton */
 	}
     } else {
-	if (!headerGetEntry(h, tag->tag, &type, &data, &count)){
+	if (!headerGetEntry(h, tag->tag, &type, (void **)&data, &count)){
 	    count = 1;
 	    type = RPM_STRING_TYPE;	
 	    data = "(none)";		/* XXX i18n? NO!, sez; gafton */
@@ -1815,7 +1884,7 @@ static char * formatValue(struct sprintfTag * tag, Header h,
     }
 
     if (tag->arrayCount) {
-	if (type == RPM_STRING_ARRAY_TYPE) free(data);
+	if (type == RPM_STRING_ARRAY_TYPE) free((void *)data);
 
 	countBuf = count;
 	data = &countBuf;
@@ -1844,7 +1913,7 @@ static char * formatValue(struct sprintfTag * tag, Header h,
     
     switch (type) {
       case RPM_STRING_ARRAY_TYPE:
-	strarray = data;
+	strarray = (const char **)data;
 
 	if (tagtype) {
 	    val = tagtype(RPM_STRING_TYPE, strarray[element], buf, tag->pad, 0);
@@ -1858,7 +1927,7 @@ static char * formatValue(struct sprintfTag * tag, Header h,
 	    sprintf(val, buf, strarray[element]);
 	}
 
-	if (mayfree) free(data);
+	if (mayfree) free((void *)data);
 
 	break;
 
@@ -1901,8 +1970,7 @@ static char * formatValue(struct sprintfTag * tag, Header h,
 	break;
 
       default:
-	val = xmalloc(20);
-	strcpy(val, _("(unknown type)"));
+	val = xstrdup(_("(unknown type)"));
 	break;
     }
 
@@ -1919,7 +1987,6 @@ static char * singleSprintf(Header h, struct sprintfToken * token,
     int i, j;
     int numElements;
     int type;
-    void * data;
     struct sprintfToken * condFormat;
     int condNumFormats;
 
@@ -1977,6 +2044,7 @@ static char * singleSprintf(Header h, struct sprintfToken * token,
 		token->u.array.format[i].u.tag.justOne) continue;
 
 	    if (token->u.array.format[i].u.tag.ext) {
+		const void * data;
 		if (getExtension(h, token->u.array.format[i].u.tag.ext,
 				 &type, &data, &numElements, 
 				 extCache + 
@@ -1992,8 +2060,7 @@ static char * singleSprintf(Header h, struct sprintfToken * token,
 	}
 
 	if (numElements == -1) {
-	    val = xmalloc(20);
-	    strcpy(val, "(none)");	/* XXX i18n? NO!, sez; gafton */
+	    val = xstrdup("(none)");	/* XXX i18n? NO!, sez; gafton */
 	} else {
 	    alloced = numElements * token->u.array.numTokens * 20;
 	    val = xmalloc(alloced);
@@ -2046,7 +2113,7 @@ static void freeExtensionCache(const struct headerSprintfExtension * extensions,
     int i = 0;
 
     while (ext->type != HEADER_EXT_LAST) {
-	if (cache[i].freeit) free(cache[i].data);
+	if (cache[i].freeit) free((void *)cache[i].data);
 
 	i++;
 	if (ext->type == HEADER_EXT_MORE)
@@ -2066,7 +2133,8 @@ char * headerSprintf(Header h, const char * origFmt,
     char * fmtString;
     struct sprintfToken * format;
     int numTokens;
-    char * answer, * piece;
+    char * answer;
+    const char * piece;
     int answerLength;
     int answerAlloced;
     int pieceLength;
@@ -2101,7 +2169,7 @@ char * headerSprintf(Header h, const char * origFmt,
 
 	    strcat(answer, piece);
 	    answerLength += pieceLength;
-	    free(piece);
+	    free((void *)piece);
 	}
     }
 
@@ -2118,8 +2186,7 @@ static char * octalFormat(int_32 type, const void * data,
     char * val;
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	val = xmalloc(20 + padding);
 	strcat(formatPrefix, "o");
@@ -2135,8 +2202,7 @@ static char * hexFormat(int_32 type, const void * data,
     char * val;
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	val = xmalloc(20 + padding);
 	strcat(formatPrefix, "x");
@@ -2155,8 +2221,7 @@ static char * realDateFormat(int_32 type, const void * data,
     char buf[50];
 
     if (type != RPM_INT32_TYPE) {
-	val = xmalloc(20);
-	strcpy(val, _("(not a number)"));
+	val = xstrdup(_("(not a number)"));
     } else {
 	val = xmalloc(50 + padding);
 	strcat(formatPrefix, "s");
