@@ -1,12 +1,13 @@
 #include "system.h"
 
 #include <rpmlib.h>
+#include <rpmmacro.h>
+#include <rpmurl.h>
 
 #include "cpio.h"
 #include "install.h"
 #include "misc.h"
 #include "rpmdb.h"
-#include <rpmmacro.h>
 
 struct callbackInfo {
     unsigned long archiveSize;
@@ -91,8 +92,8 @@ static void freeFileMemory( /*@only@*/ struct fileMemory *fileMem)
 }
 
 /* files should not be preallocated */
-static int assembleFileList(Header h, /*@out@*/ struct fileMemory ** memPtr, 
-	 /*@out@*/ int * fileCountPtr, /*@out@*/ struct fileInfo ** filesPtr, 
+static int assembleFileList(Header h, /*@out@*/ struct fileMemory ** memPtr,
+	 /*@out@*/ int * fileCountPtr, /*@out@*/ struct fileInfo ** filesPtr,
 	 int stripPrefixLength, enum fileActions * actions)
 {
     uint_32 * fileFlags;
@@ -106,20 +107,20 @@ static int assembleFileList(Header h, /*@out@*/ struct fileMemory ** memPtr,
 
     *memPtr = mem;
 
-    if (!headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &mem->names, 
-		        fileCountPtr))
-	return 0;
+    if (!headerIsEntry(h, RPMTAG_BASENAMES)) return 0;
 
-    if (!headerGetEntry(h, RPMTAG_ORIGFILENAMES, NULL, 
-			(void **) &mem->cpioNames, NULL))
-	headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &mem->cpioNames, 
-		           fileCountPtr);
-    headerRemoveEntry(h, RPMTAG_ORIGFILENAMES);
+    rpmBuildFileList(h, &mem->names, fileCountPtr);
+
+    if (headerIsEntry(h, RPMTAG_ORIGBASENAMES)) {
+	buildOrigFileList(h, &mem->cpioNames, fileCountPtr);
+    } else {
+	rpmBuildFileList(h, &mem->cpioNames, fileCountPtr);
+    }
 
     fileCount = *fileCountPtr;
 
     files = *filesPtr = mem->files = xcalloc(fileCount, sizeof(*mem->files));
-    
+
     headerGetEntry(h, RPMTAG_FILEFLAGS, NULL, (void **) &fileFlags, NULL);
     headerGetEntry(h, RPMTAG_FILEMODES, NULL, (void **) &fileModes, NULL);
     headerGetEntry(h, RPMTAG_FILESIZES, NULL, (void **) &fileSizes, NULL);
@@ -156,7 +157,7 @@ static void setFileOwners(Header h, struct fileInfo * files, int fileCount)
 
     for (i = 0; i < fileCount; i++) {
 	if (unameToUid(fileOwners[i], &files[i].uid)) {
-	    rpmError(RPMERR_NOUSER, _("user %s does not exist - using root"), 
+	    rpmError(RPMERR_NOUSER, _("user %s does not exist - using root"),
 			fileOwners[i]);
 	    files[i].uid = 0;
 	    /* turn off the suid bit */
@@ -164,7 +165,7 @@ static void setFileOwners(Header h, struct fileInfo * files, int fileCount)
 	}
 
 	if (gnameToGid(fileGroups[i], &files[i].gid)) {
-	    rpmError(RPMERR_NOGROUP, _("group %s does not exist - using root"), 
+	    rpmError(RPMERR_NOGROUP, _("group %s does not exist - using root"),
 			fileGroups[i]);
 	    files[i].gid = 0;
 	    /* turn off the sgid bit */
@@ -209,17 +210,17 @@ static void trimChangelog(Header h)
 	return;
     }
 
-    if (!headerGetEntry(h, RPMTAG_CHANGELOGTIME, NULL, (void **) &times, 
+    if (!headerGetEntry(h, RPMTAG_CHANGELOGTIME, NULL, (void **) &times,
 			&count) ||
 	count < numToKeep) return;
     headerGetEntry(h, RPMTAG_CHANGELOGNAME, NULL, (void **) &names, &count);
     headerGetEntry(h, RPMTAG_CHANGELOGTEXT, NULL, (void **) &texts, &count);
 
-    headerModifyEntry(h, RPMTAG_CHANGELOGTIME, RPM_INT32_TYPE, times, 
+    headerModifyEntry(h, RPMTAG_CHANGELOGTIME, RPM_INT32_TYPE, times,
 		      numToKeep);
-    headerModifyEntry(h, RPMTAG_CHANGELOGNAME, RPM_STRING_ARRAY_TYPE, names, 
+    headerModifyEntry(h, RPMTAG_CHANGELOGNAME, RPM_STRING_ARRAY_TYPE, names,
 		      numToKeep);
-    headerModifyEntry(h, RPMTAG_CHANGELOGTEXT, RPM_STRING_ARRAY_TYPE, texts, 
+    headerModifyEntry(h, RPMTAG_CHANGELOGTEXT, RPM_STRING_ARRAY_TYPE, texts,
 		      numToKeep);
 
     free(names);
@@ -252,13 +253,13 @@ static int markReplacedFiles(rpmdb db, struct sharedFileInfo * replList)
 		headerFree(sh);
 	    }
 
-	    headerGetEntry(secHeader, RPMTAG_FILESTATES, &type, 
+	    headerGetEntry(secHeader, RPMTAG_FILESTATES, &type,
 			   (void **) &secStates, &count);
 	}
 
 	/* by now, secHeader is the right header to modify, secStates is
 	   the right states list to modify  */
-	
+
 	secStates[fileInfo->otherFileNum] = RPMFILE_STATE_REPLACED;
     }
 
@@ -279,20 +280,20 @@ static void callback(struct cpioCallbackInfo * cpioInfo, void * data)
 
     if (ourInfo->notify)
 	(void)ourInfo->notify(ourInfo->h, RPMCALLBACK_INST_PROGRESS,
-			cpioInfo->bytesProcessed, 
-			ourInfo->archiveSize, ourInfo->pkgKey, 
+			cpioInfo->bytesProcessed,
+			ourInfo->archiveSize, ourInfo->pkgKey,
 			ourInfo->notifyData);
 
     if (ourInfo->specFilePtr) {
 	chptr = cpioInfo->file + strlen(cpioInfo->file) - 5;
-	if (!strcmp(chptr, ".spec")) 
+	if (!strcmp(chptr, ".spec"))
 	    *ourInfo->specFilePtr = xstrdup(cpioInfo->file);
     }
 }
 
 /* NULL files means install all files */
 static int installArchive(FD_t fd, struct fileInfo * files,
-			  int fileCount, rpmCallbackFunction notify, 
+			  int fileCount, rpmCallbackFunction notify,
 			  void * notifyData, const void * pkgKey, Header h,
 			  /*@out@*/const char ** specFile, int archiveSize)
 {
@@ -301,6 +302,8 @@ static int installArchive(FD_t fd, struct fileInfo * files,
     int mappedFiles = 0;
     const char * failedFile = NULL;
     struct callbackInfo info;
+    FD_t cfd;
+    int urltype;
 
     if (!files) {
 	/* install all files */
@@ -325,11 +328,15 @@ static int installArchive(FD_t fd, struct fileInfo * files,
 	    if (!files[i].install) continue;
 
 	    map[mappedFiles].archivePath = files[i].cpioPath;
+#ifdef DYING
 	    map[mappedFiles].fsPath = files[i].relativePath;
+#else
+	    urltype = urlPath(files[i].relativePath, &map[mappedFiles].fsPath);
+#endif
 	    map[mappedFiles].finalMode = files[i].mode;
 	    map[mappedFiles].finalUid = files[i].uid;
 	    map[mappedFiles].finalGid = files[i].gid;
-	    map[mappedFiles].mapFlags = CPIO_MAP_PATH | CPIO_MAP_MODE | 
+	    map[mappedFiles].mapFlags = CPIO_MAP_PATH | CPIO_MAP_MODE |
 					CPIO_MAP_UID | CPIO_MAP_GID;
 	    mappedFiles++;
 	}
@@ -338,24 +345,21 @@ static int installArchive(FD_t fd, struct fileInfo * files,
     }
 
     if (notify)
-	(void)notify(h, RPMCALLBACK_INST_PROGRESS, 0, archiveSize, pkgKey, 
+	(void)notify(h, RPMCALLBACK_INST_PROGRESS, 0, archiveSize, pkgKey,
 	       notifyData);
 
-  { CFD_t cfdbuf, *cfd = &cfdbuf;
-    cfd->cpioPos = 0;
-    cfd->cpioIoType = cpioIoTypeGzFd;
-    cfd->cpioGzFd = gzdFdopen(fdDup(fdFileno(fd)), "r");
-    rc = cpioInstallArchive(cfd, map, mappedFiles, 
-		    ((notify && archiveSize) || specFile) ? callback : NULL, 
+    (void) Fflush(fd);
+    cfd = Fdopen(fdDup(Fileno(fd)), "r.gzdio");
+    rc = cpioInstallArchive(cfd, map, mappedFiles,
+		    ((notify && archiveSize) || specFile) ? callback : NULL,
 		    &info, &failedFile);
-    gzdClose(cfd->cpioGzFd);
-  }
+    Fclose(cfd);
     headerFree(info.h);
 
     if (rc) {
 	/* this would probably be a good place to check if disk space
 	   was used up - if so, we should return a different error */
-	rpmError(RPMERR_CPIO, _("unpacking of archive failed%s%s: %s"), 
+	rpmError(RPMERR_CPIO, _("unpacking of archive failed%s%s: %s"),
 		(failedFile != NULL ? _(" on file ") : ""),
 		(failedFile != NULL ? failedFile : ""),
 		cpioStrerror(rc));
@@ -379,7 +383,7 @@ static int installArchive(FD_t fd, struct fileInfo * files,
 /* 0 success */
 /* 1 bad magic */
 /* 2 error */
-static int installSources(Header h, const char * rootdir, FD_t fd, 
+static int installSources(Header h, const char * rootdir, FD_t fd,
 			  const char ** specFilePtr, rpmCallbackFunction notify,
 			  void * notifyData)
 {
@@ -401,52 +405,67 @@ static int installSources(Header h, const char * rootdir, FD_t fd,
 
     rpmMessage(RPMMESS_DEBUG, _("installing a source package\n"));
 
-    realSourceDir = rpmGetPath(rootdir, "/%{_sourcedir}", NULL);
-    if (stat(realSourceDir, &st) < 0) {
-	switch (errno) {
-	case ENOENT:
-	    /* XXX this will only create last component of directory path */
-	    if (mkdir(realSourceDir, 0755) == 0)
+    realSourceDir = rpmGenPath(rootdir, "%{_sourcedir}", "");
+    if ((rc = Stat(realSourceDir, &st)) < 0) {
+	int ut = urlPath(realSourceDir, NULL);
+	switch (ut) {
+	case URL_IS_PATH:
+	case URL_IS_UNKNOWN:
+	    if (errno != ENOENT)
 		break;
 	    /*@fallthrough@*/
-	default:
-	    rpmError(RPMERR_CREATE, _("cannot create %s: %s"),
-		realSourceDir, strerror(errno));
+	case URL_IS_FTP:
+	case URL_IS_HTTP:
+	    /* XXX this will only create last component of directory path */
+	    rc = Mkdir(realSourceDir, 0755);
+	    break;
+	case URL_IS_DASH:
+	    break;
+	}
+	if (rc < 0) {
+	    rpmError(RPMERR_CREATE, _("cannot create sourcedir %s"), realSourceDir);
 	    rc = 2;
 	    goto exit;
-	    /*@notreached@*/ break;
 	}
     }
-    if (access(realSourceDir, W_OK)) {
+    if ((rc = Access(realSourceDir, W_OK))) {
 	rpmError(RPMERR_CREATE, _("cannot write to %s"), realSourceDir);
 	rc = 2;
 	goto exit;
     }
     rpmMessage(RPMMESS_DEBUG, _("sources in: %s\n"), realSourceDir);
 
-    realSpecDir = rpmGetPath(rootdir, "/%{_specdir}", NULL);
-    if (stat(realSpecDir, &st) < 0) {
-	switch (errno) {
-	case ENOENT:
-	    /* XXX this will only create last component of directory path */
-	    if (mkdir(realSpecDir, 0755) == 0)
+    realSpecDir = rpmGenPath(rootdir, "%{_specdir}", "");
+    if ((rc = Stat(realSpecDir, &st)) < 0) {
+	int ut = urlPath(realSpecDir, NULL);
+	switch (ut) {
+	case URL_IS_PATH:
+	case URL_IS_UNKNOWN:
+	    if (errno != ENOENT)
 		break;
 	    /*@fallthrough@*/
-	default:
-	    rpmError(RPMERR_CREATE, _("cannot create %s"), realSpecDir);
+	case URL_IS_FTP:
+	case URL_IS_HTTP:
+	    /* XXX this will only create last component of directory path */
+	    rc = Mkdir(realSpecDir, 0755);
+	    break;
+	case URL_IS_DASH:
+	    break;
+	}
+	if (rc < 0) {
+	    rpmError(RPMERR_CREATE, _("cannot create specdir %s"), realSpecDir);
 	    rc = 2;
 	    goto exit;
-	    /*@notreached@*/ break;
 	}
     }
-    if (access(realSpecDir, W_OK)) {
+    if ((rc = Access(realSpecDir, W_OK))) {
 	rpmError(RPMERR_CREATE, _("cannot write to %s"), realSpecDir);
 	rc = 2;
 	goto exit;
     }
     rpmMessage(RPMMESS_DEBUG, _("spec file in: %s\n"), realSpecDir);
 
-    if (h != NULL && headerIsEntry(h, RPMTAG_FILENAMES)) {
+    if (h != NULL && headerIsEntry(h, RPMTAG_BASENAMES)) {
 	/* we can't remap v1 packages */
 	assembleFileList(h, &fileMem, &fileCount, &files, 0, NULL);
 
@@ -470,7 +489,7 @@ static int installSources(Header h, const char * rootdir, FD_t fd,
 	}
 
 	if (i < fileCount) {
-	    char *t = alloca(strlen(realSpecDir) + 
+	    char *t = alloca(strlen(realSpecDir) +
 			 strlen(files[i].cpioPath) + 5);
 	    strcpy(t, realSpecDir);
 	    strcat(t, "/");
@@ -494,15 +513,15 @@ static int installSources(Header h, const char * rootdir, FD_t fd,
 			    (void **) &archiveSizePtr, NULL))
 	archiveSizePtr = NULL;
 
-    chdir(realSourceDir);
+    Chdir(realSourceDir);
     if (installArchive(fd, fileCount > 0 ? files : NULL,
 			  fileCount, notify, notifyData, NULL, h,
-			  specFileIndex >=0 ? NULL : &specFile, 
+			  specFileIndex >=0 ? NULL : &specFile,
 			  archiveSizePtr ? *archiveSizePtr : 0)) {
 	rc = 2;
 	goto exit;
     }
-    chdir(currDir);
+    Chdir(currDir);
 
     if (specFileIndex == -1) {
 	if (specFile == NULL) {
@@ -526,13 +545,15 @@ static int installSources(Header h, const char * rootdir, FD_t fd,
 
 	xfree(specFile);
 
-	rpmMessage(RPMMESS_DEBUG, 
+	if (strcmp(instSpecFile, correctSpecFile)) {
+	    rpmMessage(RPMMESS_DEBUG,
 		    _("renaming %s to %s\n"), instSpecFile, correctSpecFile);
-	if (rename(instSpecFile, correctSpecFile)) {
-	    rpmError(RPMERR_RENAME, _("rename of %s to %s failed: %s"),
-		     instSpecFile, correctSpecFile, strerror(errno));
-	    rc = 2;
-	    goto exit;
+	    if ((rc = Rename(instSpecFile, correctSpecFile))) {
+		rpmError(RPMERR_RENAME, _("rename of %s to %s failed: %s"),
+			instSpecFile, correctSpecFile, strerror(errno));
+		rc = 2;
+		goto exit;
+	    }
 	}
 
 	if (specFilePtr)
@@ -553,19 +574,19 @@ exit:
 
 int rpmVersionCompare(Header first, Header second)
 {
-    char * one, * two;
+    const char * one, * two;
     int_32 * epochOne, * epochTwo;
     int rc;
 
     if (!headerGetEntry(first, RPMTAG_EPOCH, NULL, (void **) &epochOne, NULL))
 	epochOne = NULL;
-    if (!headerGetEntry(second, RPMTAG_EPOCH, NULL, (void **) &epochTwo, 
+    if (!headerGetEntry(second, RPMTAG_EPOCH, NULL, (void **) &epochTwo,
 			NULL))
 	epochTwo = NULL;
 
-    if (epochOne && !epochTwo) 
+    if (epochOne && !epochTwo)
 	return 1;
-    else if (!epochOne && epochTwo) 
+    else if (!epochOne && epochTwo)
 	return -1;
     else if (epochOne && epochTwo) {
 	if (*epochOne < *epochTwo)
@@ -573,7 +594,7 @@ int rpmVersionCompare(Header first, Header second)
 	else if (*epochOne > *epochTwo)
 	    return 1;
     }
-	
+
     headerGetEntry(first, RPMTAG_VERSION, NULL, (void **) &one, NULL);
     headerGetEntry(second, RPMTAG_VERSION, NULL, (void **) &two, NULL);
 
@@ -583,11 +604,11 @@ int rpmVersionCompare(Header first, Header second)
 
     headerGetEntry(first, RPMTAG_RELEASE, NULL, (void **) &one, NULL);
     headerGetEntry(second, RPMTAG_RELEASE, NULL, (void **) &two, NULL);
-    
+
     return rpmvercmp(one, two);
 }
 
-const char * fileActionString(enum fileActions a)
+const char *const fileActionString(enum fileActions a)
 {
     switch (a) {
       case FA_UNKNOWN: return "unknown";
@@ -607,8 +628,8 @@ const char * fileActionString(enum fileActions a)
 /* 0 success */
 /* 1 bad magic */
 /* 2 error */
-int rpmInstallSourcePackage(const char * rootdir, FD_t fd, 
-			    const char ** specFile, rpmCallbackFunction notify, 
+int rpmInstallSourcePackage(const char * rootdir, FD_t fd,
+			    const char ** specFile, rpmCallbackFunction notify,
 			    void * notifyData, char ** cookie)
 {
     int rc, isSource;
@@ -639,12 +660,12 @@ int rpmInstallSourcePackage(const char * rootdir, FD_t fd,
     }
 
     rpmInstallLoadMacros(h);
-    
+
     rc = installSources(h, rootdir, fd, specFile, notify, notifyData);
     if (h != NULL) {
  	headerFree(h);
     }
- 
+
     return rc;
 }
 
@@ -652,9 +673,9 @@ int rpmInstallSourcePackage(const char * rootdir, FD_t fd,
 /* 1 bad magic */
 /* 2 error */
 int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
-		         int flags, rpmCallbackFunction notify, 
-			 void * notifyData, const void * pkgKey, 
-			 enum fileActions * actions, 
+		         int flags, rpmCallbackFunction notify,
+			 void * notifyData, const void * pkgKey,
+			 enum fileActions * actions,
 			 struct sharedFileInfo * sharedList, FD_t scriptFd)
 {
     int rc;
@@ -676,7 +697,7 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 
     headerNVR(h, &name, &version, &release);
 
-    rpmMessage(RPMMESS_DEBUG, _("package: %s-%s-%s files test = %d\n"), 
+    rpmMessage(RPMMESS_DEBUG, _("package: %s-%s-%s files test = %d\n"),
 		name, version, release, flags & RPMTRANS_FLAG_TEST);
 
     rc = rpmdbFindPackage(db, name, &matches);
@@ -715,10 +736,10 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 	/*@-unrecog@*/ chroot(rootdir); /*@=unrecog@*/
     }
 
-    if (!(flags & RPMTRANS_FLAG_JUSTDB) && headerIsEntry(h, RPMTAG_FILENAMES)) {
+    if (!(flags & RPMTRANS_FLAG_JUSTDB) && headerIsEntry(h, RPMTAG_BASENAMES)) {
 	const char * defaultPrefix;
 	/* old format relocateable packages need the entire default
-	   prefix stripped to form the cpio list, while all other packages 
+	   prefix stripped to form the cpio list, while all other packages
 	   need the leading / stripped */
 	if (headerGetEntry(h, RPMTAG_DEFAULTPREFIX, NULL, (void **)
 				  &defaultPrefix, NULL)) {
@@ -735,7 +756,7 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
     } else {
 	files = NULL;
     }
-    
+
     if (flags & RPMTRANS_FLAG_TEST) {
 	rpmMessage(RPMMESS_DEBUG, _("stopping install as we're running --test\n"));
 	rc = 0;
@@ -743,7 +764,7 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
     }
 
     rpmMessage(RPMMESS_DEBUG, _("running preinstall script (if any)\n"));
-    if (runInstScript("/", h, RPMTAG_PREIN, RPMTAG_PREINPROG, scriptArg, 
+    if (runInstScript("/", h, RPMTAG_PREIN, RPMTAG_PREINPROG, scriptArg,
 		      flags & RPMTRANS_FLAG_NOSCRIPTS, scriptFd)) {
 	rc = 2;
 	goto exit;
@@ -799,11 +820,11 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 		break;
 	    }
 
-	    if (ext) {
+	    if (ext && access(files[i].relativePath, F_OK) == 0) {
 		newpath = alloca(strlen(files[i].relativePath) + 20);
 		strcpy(newpath, files[i].relativePath);
 		strcat(newpath, ext);
-		rpmError(RPMMESS_BACKUP, _("warning: %s saved as %s"), 
+		rpmError(RPMMESS_BACKUP, _("warning: %s saved as %s"),
 			files[i].relativePath, newpath);
 
 		if (rename(files[i].relativePath, newpath)) {
@@ -817,7 +838,7 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 
 	{   uint_32 * archiveSizePtr;
 
-	    if (!headerGetEntry(h, RPMTAG_ARCHIVESIZE, &type, 
+	    if (!headerGetEntry(h, RPMTAG_ARCHIVESIZE, &type,
 				(void **) &archiveSizePtr, &count))
 	    archiveSizePtr = NULL;
 
@@ -838,17 +859,17 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 	for (i = 0; i < fileCount; i++)
 	    fileStates[i] = files[i].state;
 
-	headerAddEntry(h, RPMTAG_FILESTATES, RPM_CHAR_TYPE, fileStates, 
+	headerAddEntry(h, RPMTAG_FILESTATES, RPM_CHAR_TYPE, fileStates,
 			fileCount);
 
 	free(fileStates);
 	if (fileMem) freeFileMemory(fileMem);
 	fileMem = NULL;
     } else if (flags & RPMTRANS_FLAG_JUSTDB) {
-	if (headerGetEntry(h, RPMTAG_FILENAMES, NULL, NULL, &fileCount)) {
+	if (headerGetEntry(h, RPMTAG_BASENAMES, NULL, NULL, &fileCount)) {
 	    fileStates = xmalloc(sizeof(*fileStates) * fileCount);
 	    memset(fileStates, RPMFILE_STATE_NORMAL, fileCount);
-	    headerAddEntry(h, RPMTAG_FILESTATES, RPM_CHAR_TYPE, fileStates, 
+	    headerAddEntry(h, RPMTAG_FILESTATES, RPM_CHAR_TYPE, fileStates,
 			    fileCount);
 	    free(fileStates);
 	}
@@ -900,7 +921,7 @@ int installBinaryPackage(const char * rootdir, rpmdb db, FD_t fd, Header h,
 	}
     }
 
-    if (sharedList) 
+    if (sharedList)
 	markReplacedFiles(db, sharedList);
 
     rc = 0;
