@@ -1,9 +1,13 @@
 #include "system.h"
 
+static int _debug = 0;
+
 #include <assert.h>
 #include <stdarg.h>
 
+#if !defined(isblank)
 #define	isblank(_c)	((_c) == ' ' || (_c) == '\t')
+#endif
 #define	iseol(_c)	((_c) == '\n' || (_c) == '\r')
 
 #define	STREQ(_t, _f, _fn)	((_fn) == (sizeof(_t)-1) && !strncmp((_t), (_f), (_fn)))
@@ -21,19 +25,30 @@
 #define RPMERR_BADSPEC stderr
 #undef	_
 #define	_(x)	x
-#define	xfree(_p)	free((void *)_p)
-typedef	int FD_t;
-#define	fdFileno(_x)	(_x)
-#define	fdOpen		open
-#define	fdRead		read
-#define	fdClose		close
+
+#define	vmefail()		(exit(1), NULL)
+#define	xfree(_p)		free((void *)_p)
+#define	urlPath(_xr, _r)	*(_r) = (_xr)
+
+typedef	FILE * FD_t;
+#define Fopen(_path, _fmode)	fopen(_path, "r");
+#define	Ferror			ferror
+#define Fstrerror(_fd)		strerror(errno)
+#define	Fread			fread
+#define	Fclose			fclose
+
 #else
+
 #include <rpmlib.h>
+#include <rpmio.h>
+#include <rpmurl.h>
+
 #endif
 
 #include <rpmmacro.h>
 
-struct MacroContext globalMacroContext;
+struct MacroContext rpmGlobalMacroContext;
+struct MacroContext rpmCLIMacroContext;
 
 typedef struct MacroBuf {
 	const char *s;		/* text to expand */
@@ -71,15 +86,12 @@ compareMacroName(const void *ap, const void *bp)
 	MacroEntry *ame = *((MacroEntry **)ap);
 	MacroEntry *bme = *((MacroEntry **)bp);
 
-	if (ame == NULL && bme == NULL) {
+	if (ame == NULL && bme == NULL)
 		return 0;
-	}
-	if (ame == NULL) {
+	if (ame == NULL)
 		return 1;
-	}
-	if (bme == NULL) {
+	if (bme == NULL)
 		return -1;
-	}
 	return strcmp(ame->name, bme->name);
 }
 
@@ -118,18 +130,18 @@ sortMacroTable(MacroContext *mc)
 }
 
 void
-dumpMacroTable(MacroContext *mc, FILE *f)
+rpmDumpMacroTable(MacroContext * mc, FILE * fp)
 {
 	int i;
 	int nempty = 0;
 	int nactive = 0;
 
 	if (mc == NULL)
-		mc = &globalMacroContext;
-	if (f == NULL)
-		f = stderr;
+		mc = &rpmGlobalMacroContext;
+	if (fp == NULL)
+		fp = stderr;
     
-	fprintf(f, "========================\n");
+	fprintf(fp, "========================\n");
 	for (i = 0; i < mc->firstFree; i++) {
 		MacroEntry *me;
 		if ((me = mc->macroTable[i]) == NULL) {
@@ -137,16 +149,16 @@ dumpMacroTable(MacroContext *mc, FILE *f)
 			nempty++;
 			continue;
 		}
-		fprintf(f, "%3d%c %s", me->level,
+		fprintf(fp, "%3d%c %s", me->level,
 			(me->used > 0 ? '=' : ':'), me->name);
 		if (me->opts && *me->opts)
-			fprintf(f, "(%s)", me->opts);
+			fprintf(fp, "(%s)", me->opts);
 		if (me->body && *me->body)
-			fprintf(f, "\t%s", me->body);
-		fprintf(f, "\n");
+			fprintf(fp, "\t%s", me->body);
+		fprintf(fp, "\n");
 		nactive++;
 	}
-	fprintf(f, _("======================== active %d empty %d\n"),
+	fprintf(fp, _("======================== active %d empty %d\n"),
 		nactive, nempty);
 }
 
@@ -157,7 +169,7 @@ findEntry(MacroContext *mc, const char *name, size_t namelen)
 	char namebuf[1024];
 
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 	if (! mc->firstFree)
 		return NULL;
 
@@ -181,7 +193,7 @@ findEntry(MacroContext *mc, const char *name, size_t namelen)
 /* fgets analogue that reads \ continuations. Last newline always trimmed. */
 
 static char *
-rdcl(char *buf, size_t size, FILE *fp, int escapes)
+rdcl(char *buf, size_t size, FD_t fd, int escapes)
 {
 	char *q = buf;
 	size_t nb = 0;
@@ -189,7 +201,8 @@ rdcl(char *buf, size_t size, FILE *fp, int escapes)
 
 	*q = '\0';
 	do {
-		if (fgets(q, size, fp) == NULL)	/* read next line */
+		/* read next line */
+		if (fgets(q, size, (FILE *)fdGetFp(fd)) == NULL)
 			break;
 		nb = strlen(q);
 		nread += nb;
@@ -773,7 +786,7 @@ doOutput(MacroBuf *mb, int waserror, const char *msg, size_t msglen)
 }
 
 static void
-doFoo(MacroBuf *mb, const char *f, size_t fn, const char *g, size_t glen)
+doFoo(MacroBuf *mb, int negate, const char *f, size_t fn, const char *g, size_t glen)
 {
 	char buf[BUFSIZ], *b = NULL, *be;
 	int c;
@@ -799,6 +812,14 @@ doFoo(MacroBuf *mb, const char *f, size_t fn, const char *g, size_t glen)
 			b++;
 	} else if (STREQ("expand", f, fn)) {
 		b = buf;
+	} else if (STREQ("verbose", f, fn)) {
+		if (negate)
+		    b = (rpmIsVerbose() ? NULL : buf);
+		else
+		    b = (rpmIsVerbose() ? buf : NULL);
+	} else if (STREQ("url2path", f, fn) || STREQ("u2p", f, fn)) {
+		(void)urlPath(buf, (const char **)&b);
+		if (*b == '\0') b = "/";
 	} else if (STREQ("uncompress", f, fn)) {
 		int compressed = 1;
 		for (b = buf; (c = *b) && isblank(c);)
@@ -886,7 +907,7 @@ expandMacro(MacroBuf *mb)
 	default:
 		SAVECHAR(mb, c);
 		continue;
-		break;
+		/*@notreached@*/ break;
 	}
 
 	/* Expand next macro */
@@ -946,7 +967,7 @@ expandMacro(MacroBuf *mb)
 
 		s = se;
 		continue;
-		break;
+		/*@notreached@*/ break;
 	case '{':		/* %{...}/%{...:...} substitution */
 		if ((se = matchchar(s, c, '}')) == NULL) {
 			rpmError(RPMERR_BADSPEC, _("Unterminated %c: %s"), c, s);
@@ -1038,7 +1059,7 @@ expandMacro(MacroBuf *mb)
 	}
 
 	if (STREQ("dump", f, fn)) {
-		dumpMacroTable(mb->mc, NULL);
+		rpmDumpMacroTable(mb->mc, NULL);
 		while (iseol(*se))
 			se++;
 		s = se;
@@ -1049,11 +1070,14 @@ expandMacro(MacroBuf *mb)
 	if (STREQ("basename", f, fn) ||
 	    STREQ("suffix", f, fn) ||
 	    STREQ("expand", f, fn) ||
+	    STREQ("verbose", f, fn) ||
 	    STREQ("uncompress", f, fn) ||
+	    STREQ("url2path", f, fn) ||
+	    STREQ("u2p", f, fn) ||
 	    STREQ("S", f, fn) ||
 	    STREQ("P", f, fn) ||
 	    STREQ("F", f, fn)) {
-		doFoo(mb, f, fn, g, gn);
+		doFoo(mb, negate, f, fn, g, gn);
 		s = se;
 		continue;
 	}
@@ -1092,7 +1116,7 @@ expandMacro(MacroBuf *mb)
 		if (g && g < ge) {		/* Expand X in %{?f:X} */
 			rc = expandT(mb, g, gn);
 		} else
-		if (me->body && *me->body) {	/* Expand %{?f}/%{?f*} */
+		if (me && me->body && *me->body) { /* Expand %{?f}/%{?f*} */
 			rc = expandT(mb, me->body, strlen(me->body));
 		}
 		s = se;
@@ -1175,7 +1199,7 @@ expandMacros(void *spec, MacroContext *mc, char *s, size_t slen)
 	if (s == NULL || slen <= 0)
 		return 0;
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 
 	tbuf = alloca(slen + 1);
 	memset(tbuf, 0, (slen + 1));
@@ -1207,7 +1231,7 @@ addMacro(MacroContext *mc, const char *n, const char *o, const char *b, int leve
 	MacroEntry **mep;
 
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 
 	/* If new name, expand macro table */
 	if ((mep = findEntry(mc, n, 0)) == NULL) {
@@ -1230,7 +1254,7 @@ delMacro(MacroContext *mc, const char *n)
 	MacroEntry **mep;
 
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 	/* If name exists, pop entry */
 	if ((mep = findEntry(mc, n, 0)) != NULL) {
 		popMacro(mep);
@@ -1246,26 +1270,51 @@ rpmDefineMacro(MacroContext *mc, const char *macro, int level)
 	MacroBuf macrobuf, *mb = &macrobuf;
 
 	/* XXX just enough to get by */
-	mb->mc = (mc ? mc : &globalMacroContext);
+	mb->mc = (mc ? mc : &rpmGlobalMacroContext);
 	(void)doDefine(mb, macro, level, 0);
 	return 0;
 }
 
+/* Load a macro context into rpmGlobalMacroContext */
 void
-initMacros(MacroContext *mc, const char *macrofiles)
+rpmLoadMacros(MacroContext * mc, int level)
+{
+	int i;
+
+	if (mc == NULL || mc == &rpmGlobalMacroContext)
+		return;
+
+	for (i = 0; i < mc->firstFree; i++) {
+		MacroEntry **mep, *me;
+		mep = &mc->macroTable[i];
+		me = *mep;
+
+		if (me == NULL)		/* XXX this should never happen */
+			continue;
+		addMacro(NULL, me->name, me->opts, me->body, (level - 1));
+	}
+}
+
+void
+rpmInitMacros(MacroContext *mc, const char *macrofiles)
 {
 	char *m, *mfile, *me;
 
 	if (macrofiles == NULL)
 		return;
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 
 	for (mfile = m = xstrdup(macrofiles); *mfile; mfile = me) {
-		FILE *fp;
+		FD_t fd;
 		char buf[BUFSIZ];
 
-		if ((me = strchr(mfile, ':')) != NULL)
+		for (me = mfile; (me = strchr(me, ':')) != NULL; me++) {
+			if (!(me[1] == '/' && me[2] == '/'))
+				break;
+		}
+
+		if (me && *me == ':')
 			*me++ = '\0';
 		else
 			me = mfile + strlen(mfile);
@@ -1283,13 +1332,16 @@ initMacros(MacroContext *mc, const char *macrofiles)
 		strncat(buf, mfile, sizeof(buf) - strlen(buf));
 		buf[sizeof(buf)-1] = '\0';
 
-		if ((fp=fopen(buf, "r")) == NULL)
+		fd = Fopen(buf, "r.fpio");
+		if (fd == NULL || Ferror(fd)) {
+			if (fd) Fclose(fd);
 			continue;
+		}
 
 		/* XXX Assume new fangled macro expansion */
 		max_macro_depth = 16;
 
-		while(rdcl(buf, sizeof(buf), fp, 1) != NULL) {
+		while(rdcl(buf, sizeof(buf), fd, 1) != NULL) {
 			char c, *n;
 
 			n = buf;
@@ -1300,19 +1352,22 @@ initMacros(MacroContext *mc, const char *macrofiles)
 			n++;	/* skip % */
 			(void)rpmDefineMacro(NULL, n, RMIL_MACROFILES);
 		}
-		fclose(fp);
+		Fclose(fd);
 	}
 	if (m)
 		free(m);
+
+	/* Reload cmdline macros */
+	rpmLoadMacros(&rpmCLIMacroContext, RMIL_CMDLINE);
 }
 
 void
-freeMacros(MacroContext *mc)
+rpmFreeMacros(MacroContext *mc)
 {
 	int i;
     
 	if (mc == NULL)
-		mc = &globalMacroContext;
+		mc = &rpmGlobalMacroContext;
 
 	for (i = 0; i < mc->firstFree; i++) {
 		MacroEntry *me;
@@ -1334,27 +1389,32 @@ int isCompressed(const char *file, int *compressed)
 {
     FD_t fd;
     ssize_t nb;
-    int rderrno;
+    int rc = -1;
     unsigned char magic[4];
 
     *compressed = COMPRESSED_NOT;
 
-    if (fdFileno(fd = fdOpen(file, O_RDONLY, 0)) < 0) {
-	rpmError(RPMERR_BADSPEC, _("File %s: %s"), file, strerror(errno));
+    fd = Fopen(file, "r.ufdio");
+    if (fd == NULL || Ferror(fd)) {
+	/* XXX Fstrerror */
+	rpmError(RPMERR_BADSPEC, _("File %s: %s"), file, Fstrerror(fd));
+	if (fd) Fclose(fd);
 	return 1;
     }
-    nb = fdRead(fd, magic, sizeof(magic));
-    rderrno = errno;
-    fdClose(fd);
-
+    nb = Fread(magic, sizeof(char), sizeof(magic), fd);
     if (nb < 0) {
-	rpmError(RPMERR_BADSPEC, _("File %s: %s"), file, strerror(rderrno));
-	return 1;
+	rpmError(RPMERR_BADSPEC, _("File %s: %s"), file, Fstrerror(fd));
+	rc = 1;
     } else if (nb < sizeof(magic)) {
 	rpmError(RPMERR_BADSPEC, _("File %s is smaller than %d bytes"),
 		file, sizeof(magic));
-	return 0;
+	rc = 0;
     }
+    Fclose(fd);
+    if (rc >= 0)
+	return rc;
+
+    rc = 0;
 
     if (((magic[0] == 0037) && (magic[1] == 0213)) ||  /* gzip */
 	((magic[0] == 0037) && (magic[1] == 0236)) ||  /* old gzip */
@@ -1369,7 +1429,7 @@ int isCompressed(const char *file, int *compressed)
 	*compressed = COMPRESSED_BZIP2;
     }
 
-    return 0;
+    return rc;
 }
 
 /* =============================================================== */
@@ -1397,7 +1457,7 @@ rpmExpand(const char *arg, ...)
     }
     va_end(ap);
     expandMacros(NULL, NULL, buf, sizeof(buf));
-    return xstrdup(buf);	/* XXX build memory leaks */
+    return xstrdup(buf);
 }
 
 int
@@ -1427,36 +1487,161 @@ rpmExpandNumeric(const char *arg)
     return rc;
 }
 
-/* Return concatenated and expanded path with multiple /'s removed */
+/* XXX FIXME: ../sbin/./../bin/ */
+char *rpmCleanPath(char * path)
+{
+    const char *s;
+    char *se, *t, *te;
+
+    s = t = te = path;
+    while (*s) {
+/*fprintf(stderr, "*** got \"%.*s\"\trest \"%s\"\n", (t-path), path, s); */
+	switch(*s) {
+	case ':':			/* handle url's */
+	    if (s[1] == '/' && s[2] == '/') {
+		*t++ = *s++;
+		*t++ = *s++;
+	    }
+	    break;
+	case '/':
+	    /* Move parent dir forward */
+	    for (se = te + 1; se < t && *se != '/'; se++)
+		;
+	    if (se < t && *se == '/') {
+		te = se;
+/*fprintf(stderr, "*** next pdir \"%.*s\"\n", (te-path), path); */
+	    }
+	    while (s[1] == '/')
+		s++;
+	    while (t > path && t[-1] == '/')
+		t--;
+	    break;
+	case '.':
+	    /* Leading .. is special */
+	    if (t == path && s[1] == '.') {
+		*t++ = *s++;
+		break;
+	    }
+	    /* Single . is special */
+	    if (t == path && s[1] == '\0') {
+		break;
+	    }
+	    /* Trim leading ./ , embedded ./ , trailing /. */
+	    if ((t == path || t[-1] == '/') && (s[1] == '/' || s[1] == '\0')) {
+/*fprintf(stderr, "*** Trim leading ./ , embedded ./ , trailing /.\n"); */
+		s++;
+		continue;
+	    }
+	    /* Trim embedded /../ and trailing /.. */
+	    if (t > path && t[-1] == '/' && s[1] == '.' && (s[2] == '/' || s[2] == '\0')) {
+		t = te;
+		/* Move parent dir forward */
+		if (te > path)
+		    for (--te; te > path && *te != '/'; te--)
+			;
+/*fprintf(stderr, "*** prev pdir \"%.*s\"\n", (te-path), path); */
+		s++;
+		s++;
+		continue;
+	    }
+	    break;
+	default:
+	    break;
+	}
+	*t++ = *s++;
+    }
+
+    /* Trim trailing / (but leave single / alone) */
+    if (t > &path[1] && t[-1] == '/')
+	t--;
+    *t = '\0';
+
+    return path;
+}
+
+/* Return concatenated and expanded canonical path. */
 const char *
 rpmGetPath(const char *path, ...)
 {
-    char buf[BUFSIZ], *p, *pe;
-    const char *s;
+    char buf[BUFSIZ];
+    const char * s;
+    char * t, * te;
     va_list ap;
 
     if (path == NULL)
 	return xstrdup("");
 
-    p = buf;
-    strcpy(p, path);
-    pe = p + strlen(p);
-    *pe = '\0';
+    t = buf;
+    te = stpcpy(t, path);
+    *te = '\0';
 
     va_start(ap, path);
     while ((s = va_arg(ap, const char *)) != NULL) {
-	/* XXX FIXME: this fixes only some of the "...//..." problems */
-	if (pe > p && pe[-1] == '/')
-	    while(*s && *s == '/')	s++;
-	if (*s != '\0') {
-	    strcpy(pe, s);
-	    pe += strlen(pe);
-	    *pe = '\0';
-	}
+	te = stpcpy(te, s);
+	*te = '\0';
     }
     va_end(ap);
     expandMacros(NULL, NULL, buf, sizeof(buf));
-    return xstrdup(buf);
+
+    return xstrdup( rpmCleanPath(buf) );
+}
+
+/* Merge 3 args into path, any or all of which may be a url. */
+
+const char * rpmGenPath(const char * urlroot, const char * urlmdir,
+		const char *urlfile)
+{
+    const char * xroot = rpmGetPath(urlroot, NULL), * root = xroot;
+    const char * xmdir = rpmGetPath(urlmdir, NULL), * mdir = xmdir;
+    const char * xfile = rpmGetPath(urlfile, NULL), * file = xfile;
+    const char * result;
+    const char * url = NULL;
+    int nurl = 0;
+    int ut;
+
+if (_debug)
+fprintf(stderr, "*** RGP xroot %s xmdir %s xfile %s\n", xroot, xmdir, xfile);
+    ut = urlPath(xroot, &root);
+    if (url == NULL && ut > URL_IS_DASH) {
+	url = xroot;
+	nurl = root - xroot;
+if (_debug)
+fprintf(stderr, "*** RGP ut %d root %s nurl %d\n", ut, root, nurl);
+    }
+    if (root == NULL || *root == '\0') root = "/";
+
+    ut = urlPath(xmdir, &mdir);
+    if (url == NULL && ut > URL_IS_DASH) {
+	url = xmdir;
+	nurl = mdir - xmdir;
+if (_debug)
+fprintf(stderr, "*** RGP ut %d mdir %s nurl %d\n", ut, mdir, nurl);
+    }
+    if (mdir == NULL || *mdir == '\0') mdir = "/";
+
+    ut = urlPath(xfile, &file);
+    if (url == NULL && ut > URL_IS_DASH) {
+	url = xfile;
+	nurl = file - xfile;
+if (_debug)
+fprintf(stderr, "*** RGP ut %d file %s nurl %d\n", ut, file, nurl);
+    }
+
+    if (url && nurl > 0) {
+	char *t = strncpy(alloca(nurl+1), url, nurl);
+	t[nurl] = '\0';
+	url = t;
+    } else
+	url = "";
+
+    result = rpmGetPath(url, root, "/", mdir, "/", file, NULL);
+
+    xfree(xroot);
+    xfree(xmdir);
+    xfree(xfile);
+if (_debug)
+fprintf(stderr, "*** RGP result %s\n", result);
+    return result;
 }
 
 /* =============================================================== */
@@ -1491,7 +1676,7 @@ main(int argc, char *argv[])
 	    exit(1);
 	}
 
-	initMacros(NULL, macrofiles);
+	rpmInitMacros(NULL, macrofiles);
 	for ( ; optind < argc; optind++) {
 	    const char *val;
 
@@ -1501,6 +1686,7 @@ main(int argc, char *argv[])
 		xfree(val);
 	    }
 	}
+	rpmFreeMacros(NULL);
 	return 0;
 }
 
@@ -1516,8 +1702,8 @@ main(int argc, char *argv[])
 	FILE *fp;
 	int x;
 
-	initMacros(NULL, macrofiles);
-	dumpMacroTable(NULL, NULL);
+	rpmInitMacros(NULL, macrofiles);
+	rpmDumpMacroTable(NULL, NULL);
 
 	if ((fp = fopen(testfile, "r")) != NULL) {
 		while(rdcl(buf, sizeof(buf), fp, 1)) {
@@ -1533,6 +1719,7 @@ main(int argc, char *argv[])
 		fprintf(stderr, "%d->%s\n <-\n", x, buf);
 		memset(buf, 0, sizeof(buf));
 	}
+	rpmFreeMacros(NULL);
 
 	return 0;
 }
