@@ -7,6 +7,8 @@
 #include "rpmbuild.h"
 #include <rpmurl.h>
 
+int specedit = 0;
+
 /* ======================================================================== */
 static char * permsString(int mode)
 {
@@ -131,11 +133,11 @@ static void printFileInfo(FILE *fp, const char * name,
 static int queryHeader(FILE *fp, Header h, const char * chptr)
 {
     char * str;
-    const char * error;
+    const char * errstr;
 
-    str = headerSprintf(h, chptr, rpmTagTable, rpmHeaderFormats, &error);
-    if (!str) {
-	fprintf(stderr, _("error in format: %s\n"), error);
+    str = headerSprintf(h, chptr, rpmTagTable, rpmHeaderFormats, &errstr);
+    if (str == NULL) {
+	fprintf(stderr, _("error in format: %s\n"), errstr);
 	return 1;
     }
 
@@ -153,7 +155,7 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
     const char * name, * version, * release;
     int_32 count, type;
     char * prefix = NULL;
-    const char ** fileList;
+    const char ** dirNames, ** baseNames;
     const char ** fileMD5List;
     const char * fileStatesList;
     const char ** fileOwnerList = NULL;
@@ -164,6 +166,7 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
     int_32 * fileGIDList = NULL;
     uint_16 * fileModeList;
     uint_16 * fileRdevList;
+    int_32 * dirIndexes;
     int i;
 
     headerNVR(h, &name, &version, &release);
@@ -175,8 +178,8 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
 	    queryHeader(fp, h, queryFormat);
 
 	if (queryFlags & QUERY_FOR_LIST) {
-	    if (!headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, 
-		 &count)) {
+	    if (!headerGetEntry(h, RPMTAG_BASENAMES, &type, 
+				(void **) &baseNames, &count)) {
 		fputs(_("(contains no files)"), fp);
 		fputs("\n", fp);
 	    } else {
@@ -184,6 +187,10 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
 			 (void **) &fileStatesList, &count)) {
 		    fileStatesList = NULL;
 		}
+		headerGetEntry(h, RPMTAG_DIRNAMES, NULL,
+			 (void **) &dirNames, NULL);
+		headerGetEntry(h, RPMTAG_DIRINDEXES, NULL, 
+			 (void **) &dirIndexes, NULL);
 		headerGetEntry(h, RPMTAG_FILEFLAGS, &type, 
 			 (void **) &fileFlagsList, &count);
 		headerGetEntry(h, RPMTAG_FILESIZES, &type, 
@@ -245,9 +252,10 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
 				fputs(    _("(no state)    "), fp);
 			    }
 			}
-			    
+
 			if (queryFlags & QUERY_FOR_DUMPFILES) {
-			    fprintf(fp, "%s %d %d %s 0%o ", fileList[i],
+			    fprintf(fp, "%s%s %d %d %s 0%o ", 
+				   dirNames[dirIndexes[i]], baseNames[i],
 				   fileSizeList[i], fileMTimeList[i],
 				   fileMD5List[i], fileModeList[i]);
 
@@ -273,28 +281,43 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
 				fprintf(fp, "X\n");
 
 			} else if (!rpmIsVerbose()) {
-			    fputs(fileList[i], fp);
+			    fputs(dirNames[dirIndexes[i]], fp);
+			    fputs(baseNames[i], fp);
 			    fputs("\n", fp);
-			} else if (fileOwnerList && fileGroupList) 
-			    printFileInfo(fp, fileList[i], fileSizeList[i],
-					  fileModeList[i], fileMTimeList[i],
-					  fileRdevList[i], fileOwnerList[i], 
-					  fileGroupList[i], -1, 
-					  -1, fileLinktoList[i]);
-			else if (fileUIDList && fileGIDList) {
-			    printFileInfo(fp, fileList[i], fileSizeList[i],
-					  fileModeList[i], fileMTimeList[i],
-					  fileRdevList[i], NULL, 
-					  NULL, fileUIDList[i], 
-					  fileGIDList[i], fileLinktoList[i]);
 			} else {
-			    rpmError(RPMERR_INTERNAL, _("package has "
-				    "neither file owner or id lists"));
+			    char * filespec;
+
+			    filespec = xmalloc(strlen(dirNames[dirIndexes[i]])
+					      + strlen(baseNames[i]) + 1);
+			    strcpy(filespec, dirNames[dirIndexes[i]]);
+			    strcat(filespec, baseNames[i]);
+					
+			    if (fileOwnerList && fileGroupList) {
+				printFileInfo(fp, filespec, fileSizeList[i],
+					      fileModeList[i], fileMTimeList[i],
+					      fileRdevList[i], 
+					      fileOwnerList[i], 
+					      fileGroupList[i], -1, 
+					      -1, fileLinktoList[i]);
+			    } else if (fileUIDList && fileGIDList) {
+				printFileInfo(fp, filespec, fileSizeList[i],
+					      fileModeList[i], fileMTimeList[i],
+					      fileRdevList[i], NULL, 
+					      NULL, fileUIDList[i], 
+					      fileGIDList[i], 
+					      fileLinktoList[i]);
+			    } else {
+				rpmError(RPMERR_INTERNAL, _("package has "
+					"neither file owner or id lists"));
+			    }
+
+			    free(filespec);
 			}
 		    }
 		}
 	    
-		free(fileList);
+		free(dirNames);
+		free(baseNames);
 		free(fileLinktoList);
 		free(fileMD5List);
 		if (fileOwnerList) free(fileOwnerList);
@@ -308,28 +331,29 @@ int showQueryPackage(QVA_t *qva, /*@unused@*/rpmdb db, Header h)
 static void
 printNewSpecfile(Spec spec)
 {
+    Header h = spec->packages->header;
     struct speclines *sl = spec->sl;
     struct spectags *st = spec->st;
-    char buf[8192];
+    const char * msgstr = NULL;
     int i, j;
 
     if (sl == NULL || st == NULL)
 	return;
 
     for (i = 0; i < st->st_ntags; i++) {
-	char *msgstr;
-	struct spectag *t;
-	t = st->st_t + i;
+	struct spectag * t = st->st_t + i;
+	const char * tn = tagName(t->t_tag);
+	const char * errstr;
+	char fmt[128];
 
-	/* XXX Summary tag often precedes name, so build msgid now. */
-	if (t->t_msgid == NULL) {
-	    char *n;
-	    headerGetEntry(spec->packages->header, RPMTAG_NAME, NULL,
-		(void **) &n, NULL);
-	    sprintf(buf, "%s(%s)", n, tagName(t->t_tag));
-	    t->t_msgid = xstrdup(buf);
+	fmt[0] = '\0';
+	(void) stpcpy( stpcpy( stpcpy( fmt, "%{"), tn), "}\n");
+	if (msgstr) xfree(msgstr);
+	msgstr = headerSprintf(h, fmt, rpmTagTable, rpmHeaderFormats, &errstr);
+	if (msgstr == NULL) {
+	    fprintf(stderr, _("can't query %s: %s\n"), tn, errstr);
+	    return;
 	}
-	msgstr = xstrdup(/*@-unrecog@*/ dgettext(specedit, t->t_msgid) /*@=unrecog@*/);
 
 	switch(t->t_tag) {
 	case RPMTAG_SUMMARY:
@@ -338,10 +362,10 @@ printNewSpecfile(Spec spec)
 	    sl->sl_lines[t->t_startx] = NULL;
 	    if (t->t_lang && strcmp(t->t_lang, RPMBUILD_DEFAULT_LANG))
 		continue;
-	    sprintf(buf, "%s: %s\n",
-		((t->t_tag == RPMTAG_GROUP) ? "Group" : "Summary"),
-		msgstr);
-	    sl->sl_lines[t->t_startx] = xstrdup(buf);
+	    {   char *buf = xmalloc(strlen(tn) + sizeof(": ") + strlen(msgstr));
+		(void) stpcpy( stpcpy( stpcpy(buf, tn), ": "), msgstr);
+		sl->sl_lines[t->t_startx] = buf;
+	    }
 	    break;
 	case RPMTAG_DESCRIPTION:
 	    for (j = 1; j < t->t_nlines; j++) {
@@ -359,6 +383,7 @@ printNewSpecfile(Spec spec)
 	    break;
 	}
     }
+    if (msgstr) xfree(msgstr);
 
     for (i = 0; i < sl->sl_nlines; i++) {
 	if (sl->sl_lines[i] == NULL)
@@ -367,7 +392,8 @@ printNewSpecfile(Spec spec)
     }
 }
 
-void rpmDisplayQueryTags(FILE * f) {
+void rpmDisplayQueryTags(FILE * f)
+{
     const struct headerTagTableEntry * t;
     int i;
     const struct headerSprintfExtension * ext = rpmHeaderFormats;
@@ -377,12 +403,18 @@ void rpmDisplayQueryTags(FILE * f) {
     }
 
     while (ext->name) {
-	if (ext->type == HEADER_EXT_TAG)
-	    fprintf(f, "%s\n", ext->name + 7), ext++;
-	else if (ext->type == HEADER_EXT_MORE)
+	if (ext->type == HEADER_EXT_MORE) {
 	    ext = ext->u.more;
-	else
-	    ext++;
+	    continue;
+	}
+	/* XXX don't print query tags twice. */
+	for (i = 0, t = rpmTagTable; i < rpmTagTableSize; i++, t++) {
+	    if (!strcmp(t->name, ext->name))
+	    	break;
+	}
+	if (i >= rpmTagTableSize && ext->type == HEADER_EXT_TAG)
+	    fprintf(f, "%s\n", ext->name + 7);
+	ext++;
     }
 }
 
@@ -415,11 +447,10 @@ int showMatches(QVA_t *qva, rpmdb db, dbiIndexSet matches, QVF_t showPackage)
 /*
  * XXX Eliminate linkage loop into librpmbuild.a
  */
-int	(*parseSpecVec) (Spec *specp, const char *specFile, const char *buildRoot,
-                int inBuildArch, const char *passPhrase, char *cookie, int anyarch,
-                int force) = NULL;
+int	(*parseSpecVec) (Spec *specp, const char *specFile, const char *rootdir,
+		const char *buildRoot, int inBuildArch, const char *passPhrase,
+		char *cookie, int anyarch, int force) = NULL;
 void	(*freeSpecVec) (Spec spec) = NULL;
-char	*specedit = NULL;
 
 int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	rpmdb db, QVF_t showPackage)
@@ -434,43 +465,63 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
     char *end = NULL;
 
     switch (source) {
-      case RPMQV_RPM:
-      {	FD_t fd;
+    case RPMQV_RPM:
+    {	int argc = 0;
+	const char ** argv = NULL;
+	int i;
 
-	fd = ufdOpen(arg, O_RDONLY, 0);
-	if (fdFileno(fd) < 0) {
-	    fprintf(stderr, _("open of %s failed: %s\n"), arg,urlStrerror(arg));
-	    ufdClose(fd);
-	    retcode = 1;
-	    break;
-	}
-
-	retcode = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);
-
-	ufdClose(fd);
-
-	switch (retcode) {
-	case 0:
-	    if (h == NULL) {
-		fprintf(stderr, _("old format source packages cannot "
-			"be queried\n"));
+	rc = rpmGlob(arg, &argc, &argv);
+	if (rc)
+	    return 1;
+	for (i = 0; i < argc; i++) {
+	    FD_t fd;
+	    fd = Fopen(argv[i], "r.ufdio");
+	    if (Ferror(fd)) {
+		/* XXX Fstrerror */
+		fprintf(stderr, _("open of %s failed: %s\n"), argv[i],
+#ifndef	NOTYET
+			urlStrerror(argv[i]));
+#else
+			Fstrerror(fd));
+#endif
+		if (fd) Fclose(fd);
 		retcode = 1;
 		break;
 	    }
-	    retcode = showPackage(qva, db, h);
-	    headerFree(h);
-	    break;
-	case 1:
-	    fprintf(stderr, _("%s does not appear to be a RPM package\n"), arg);
-	    /*@fallthrough@*/
-	case 2:
-	    fprintf(stderr, _("query of %s failed\n"), arg);
-	    retcode = 1;
-	    break;
-	}
-      } break;
 
-      case RPMQV_SPECFILE:
+	    retcode = rpmReadPackageHeader(fd, &h, &isSource, NULL, NULL);
+
+	    Fclose(fd);
+
+	    switch (retcode) {
+	    case 0:
+		if (h == NULL) {
+		    fprintf(stderr, _("old format source packages cannot "
+			"be queried\n"));
+		    retcode = 1;
+		    break;
+		}
+		retcode = showPackage(qva, db, h);
+		headerFree(h);
+		break;
+	    case 1:
+		fprintf(stderr, _("%s does not appear to be a RPM package\n"),
+				argv[i]);
+		/*@fallthrough@*/
+	    case 2:
+		fprintf(stderr, _("query of %s failed\n"), argv[i]);
+		retcode = 1;
+		break;
+	    }
+	}
+	if (argv) {
+	    for (i = 0; i < argc; i++)
+		xfree(argv[i]);
+	    xfree(argv);
+	}
+    }	break;
+
+    case RPMQV_SPECFILE:
 	if (showPackage != showQueryPackage)
 	    return 1;
 
@@ -487,7 +538,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	int anyarch = 1;
 	int force = 1;
 
-	rc = parseSpecVec(&spec, arg, buildRoot, inBuildArch, passPhrase,
+	rc = parseSpecVec(&spec, arg, "/", buildRoot, inBuildArch, passPhrase,
 		cookie, anyarch, force);
 	if (rc || spec == NULL) {
 	    
@@ -497,7 +548,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	    break;
 	}
 
-	if (specedit != NULL) {
+	if (specedit) {
 	    printNewSpecfile(spec);
 	    freeSpecVec(spec);
 	    retcode = 0;
@@ -510,7 +561,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	freeSpecVec(spec);
       }	break;
 
-      case RPMQV_ALL:
+    case RPMQV_ALL:
 	for (offset = rpmdbFirstRecNum(db);
 	     offset != 0;
 	     offset = rpmdbNextRecNum(db, offset)) {
@@ -525,7 +576,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_GROUP:
+    case RPMQV_GROUP:
 	if (rpmdbFindByGroup(db, arg, &matches)) {
 	    fprintf(stderr, _("group %s does not contain any packages\n"), arg);
 	    retcode = 1;
@@ -535,17 +586,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_WHATPROVIDES:
-	if (rpmdbFindByProvides(db, arg, &matches)) {
-	    fprintf(stderr, _("no package provides %s\n"), arg);
-	    retcode = 1;
-	} else {
-	    retcode = showMatches(qva, db, matches, showPackage);
-	    dbiFreeIndexRecord(matches);
-	}
-	break;
-
-      case RPMQV_TRIGGEREDBY:
+    case RPMQV_TRIGGEREDBY:
 	if (rpmdbFindByTriggeredBy(db, arg, &matches)) {
 	    fprintf(stderr, _("no package triggers %s\n"), arg);
 	    retcode = 1;
@@ -555,7 +596,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_WHATREQUIRES:
+    case RPMQV_WHATREQUIRES:
 	if (rpmdbFindByRequiredBy(db, arg, &matches)) {
 	    fprintf(stderr, _("no package requires %s\n"), arg);
 	    retcode = 1;
@@ -565,7 +606,19 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_PATH:
+    case RPMQV_WHATPROVIDES:
+	if (arg[0] != '/') {
+	    if (rpmdbFindByProvides(db, arg, &matches)) {
+		fprintf(stderr, _("no package provides %s\n"), arg);
+		retcode = 1;
+	    } else {
+		retcode = showMatches(qva, db, matches, showPackage);
+		dbiFreeIndexRecord(matches);
+	    }
+	    break;
+	}
+	/*@fallthrough@*/
+    case RPMQV_PATH:
 	if (rpmdbFindByFile(db, arg, &matches)) {
 	    int myerrno = 0;
 	    if (access(arg, F_OK) != 0)
@@ -585,7 +638,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_DBOFFSET:
+    case RPMQV_DBOFFSET:
 	recNumber = strtoul(arg, &end, 10);
 	if ((*end) || (end == arg) || (recNumber == ULONG_MAX)) {
 	    fprintf(stderr, _("invalid package number: %s\n"), arg);
@@ -602,7 +655,7 @@ int rpmQueryVerify(QVA_t *qva, enum rpmQVSources source, const char * arg,
 	}
 	break;
 
-      case RPMQV_PACKAGE:
+    case RPMQV_PACKAGE:
 	rc = rpmdbFindByLabel(db, arg, &matches);
 	if (rc == 1) {
 	    retcode = 1;
@@ -754,9 +807,8 @@ struct poptOption rpmQueryPoptTable[] = {
 		POPT_QUERYFORMAT, NULL, NULL },
 	{ "queryformat", '\0', POPT_ARG_STRING, 0, POPT_QUERYFORMAT,
 		N_("use the following query format"), "QUERYFORMAT" },
-	{ "specedit", '\0', POPT_ARG_STRING, &specedit, 0,
-		N_("substitute i18n sections from the following catalogue"),
-			"METACATALOGUE" },
+	{ "specedit", '\0', POPT_ARG_VAL, &specedit, -1,
+		N_("substitute i18n sections into spec file"), NULL },
 	{ "state", 's', 0, 0, 's',
 		N_("display the states of the listed files"), NULL },
 	{ "verbose", 'v', 0, 0, 'v',
