@@ -5,6 +5,7 @@
 #include <sys/signal.h>
 
 #include <rpmlib.h>
+#include <rpmurl.h>
 #include <rpmmacro.h>	/* XXX for rpmGetPath */
 
 #include "falloc.h"
@@ -15,6 +16,8 @@
 /*@access dbiIndexSet@*/
 /*@access dbiIndexRecord@*/
 
+extern int _noDirTokens;
+
 const char *rpmdb_filenames[] = {
     "packages.rpm",
     "nameindex.rpm",
@@ -39,7 +42,7 @@ const char *rpmdb_filenames[] = {
    right area w/o a linear search through the database. */
 
 struct rpmdb_s {
-    faFile pkgs;
+    FD_t pkgs;
     dbiIndex * nameIndex, * fileIndex, * groupIndex, * providesIndex;
     dbiIndex * requiredbyIndex, * conflictsIndex, * triggerIndex;
 };
@@ -59,59 +62,6 @@ static void unblockSignals(void)
     sigprocmask(SIG_SETMASK, &signalMask, NULL);
 }
 
-int rpmdbOpenForTraversal(const char * prefix, rpmdb * rpmdbp)
-{
-    const char * dbpath;
-    int rc = 0;
-
-    dbpath = rpmGetPath("%{_dbpath}", NULL);
-    if (dbpath == NULL || dbpath[0] == '%') {
-	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
-	return 1;
-    }
-
-    if (openDatabase(prefix, dbpath, rpmdbp, O_RDONLY, 0644, 
-		     RPMDB_FLAG_MINIMAL)) {
-	rc = 1;
-    }
-    xfree(dbpath);
-    return rc;
-}
-
-int rpmdbOpen (const char * prefix, rpmdb *rpmdbp, int mode, int perms)
-{
-    const char * dbpath;
-    int rc;
-
-    dbpath = rpmGetPath("%{_dbpath}", NULL);
-    if (dbpath == NULL || dbpath[0] == '%') {
-	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
-	return 1;
-    }
-
-    rc = openDatabase(prefix, dbpath, rpmdbp, mode, perms, 0);
-    xfree(dbpath);
-    return rc;
-}
-
-int rpmdbInit (const char * prefix, int perms)
-{
-    rpmdb db;
-    const char * dbpath;
-    int rc;
-
-    dbpath = rpmGetPath("%{_dbpath}", NULL);
-    if (dbpath == NULL || dbpath[0] == '%') {
-	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
-	return 1;
-    }
-
-    rc = openDatabase(prefix, dbpath, &db, O_CREAT | O_RDWR, perms, 
-			RPMDB_FLAG_JUSTCHECK);
-    xfree(dbpath);
-    return rc;
-}
-
 static int openDbFile(const char * prefix, const char * dbpath, const char * shortName, 
 	 int justCheck, int mode, int perms, dbiIndex ** db, DBTYPE type)
 {
@@ -119,7 +69,13 @@ static int openDbFile(const char * prefix, const char * dbpath, const char * sho
     char * filename = alloca(len);
 
     *filename = '\0';
-    if (prefix && *prefix) strcat(filename, prefix); 
+    switch (urlIsURL(dbpath)) {
+    case URL_IS_UNKNOWN:
+	if (prefix && *prefix) strcat(filename, prefix); 
+	break;
+    default:
+	break;
+    }
     strcat(filename, dbpath);
     strcat(filename, shortName);
 
@@ -177,20 +133,32 @@ int openDatabase(const char * prefix, const char * dbpath, rpmdb *rpmdbp, int mo
     }
     
     filename = alloca(strlen(prefix) + strlen(dbpath) + 40);
+    *filename = '\0';
 
-    strcpy(filename, prefix); 
+    switch (urlIsURL(dbpath)) {
+    case URL_IS_UNKNOWN:
+	strcat(filename, prefix); 
+	break;
+    default:
+	break;
+    }
     strcat(filename, dbpath);
+    rpmCleanPath(filename);
 
     rpmMessage(RPMMESS_DEBUG, _("opening database mode 0x%x in %s\n"),
 	mode, filename);
 
+    if (filename[strlen(filename)-1] != '/')
+	strcat(filename, "/");
     strcat(filename, "packages.rpm");
 
     db = newRpmdb();
 
     if (!justcheck || !rpmfileexists(filename)) {
-	if ((db->pkgs = faOpen(filename, mode, perms)) == NULL) {
-	    rpmError(RPMERR_DBOPEN, _("failed to open %s\n"), filename);
+	db->pkgs = fadOpen(filename, mode, perms);
+	if (Ferror(db->pkgs)) {
+	    rpmError(RPMERR_DBOPEN, _("failed to open %s: %s\n"), filename,
+		Fstrerror(db->pkgs));
 	    return 1;
 	}
 
@@ -202,7 +170,7 @@ int openDatabase(const char * prefix, const char * dbpath, rpmdb *rpmdbp, int mo
 	
 	if (mode & O_RDWR) {
 	    lockinfo.l_type = F_WRLCK;
-	    if (faFcntl(db->pkgs, F_SETLK, (void *) &lockinfo)) {
+	    if (Fcntl(db->pkgs, F_SETLK, (void *) &lockinfo)) {
 		rpmError(RPMERR_FLOCK, _("cannot get %s lock on database"), 
 			 _("exclusive"));
 		rpmdbClose(db);
@@ -210,7 +178,7 @@ int openDatabase(const char * prefix, const char * dbpath, rpmdb *rpmdbp, int mo
 	    } 
 	} else {
 	    lockinfo.l_type = F_RDLCK;
-	    if (faFcntl(db->pkgs, F_SETLK, (void *) &lockinfo)) {
+	    if (Fcntl(db->pkgs, F_SETLK, (void *) &lockinfo)) {
 		rpmError(RPMERR_FLOCK, _("cannot get %s lock on database"), 
 			 _("shared"));
 		rpmdbClose(db);
@@ -274,9 +242,40 @@ int openDatabase(const char * prefix, const char * dbpath, rpmdb *rpmdbp, int mo
      return rc;
 }
 
+static int doRpmdbOpen (const char * prefix, /*@out@*/ rpmdb * rpmdbp,
+			int mode, int perms, int flags)
+{
+    const char * dbpath = rpmGetPath("%{_dbpath}", NULL);
+    int rc;
+
+    if (!(dbpath && dbpath[0] != '%')) {
+	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
+	rc = 1;
+    } else
+    	rc = openDatabase(prefix, dbpath, rpmdbp, mode, perms, flags);
+    xfree(dbpath);
+    return rc;
+}
+
+int rpmdbOpenForTraversal(const char * prefix, rpmdb * rpmdbp)
+{
+    return doRpmdbOpen(prefix, rpmdbp, O_RDONLY, 0644, RPMDB_FLAG_MINIMAL);
+}
+
+int rpmdbOpen (const char * prefix, rpmdb *rpmdbp, int mode, int perms)
+{
+    return doRpmdbOpen(prefix, rpmdbp, mode, perms, 0);
+}
+
+int rpmdbInit (const char * prefix, int perms)
+{
+    rpmdb db;
+    return doRpmdbOpen(prefix, &db, (O_CREAT | O_RDWR), perms, RPMDB_FLAG_JUSTCHECK);
+}
+
 void rpmdbClose (rpmdb db)
 {
-    if (db->pkgs != NULL) faClose(db->pkgs);
+    if (db->pkgs != NULL) Fclose(db->pkgs);
     if (db->fileIndex) dbiCloseIndex(db->fileIndex);
     if (db->groupIndex) dbiCloseIndex(db->groupIndex);
     if (db->nameIndex) dbiCloseIndex(db->nameIndex);
@@ -288,56 +287,66 @@ void rpmdbClose (rpmdb db)
 }
 
 int rpmdbFirstRecNum(rpmdb db) {
-    return faFirstOffset(db->pkgs);
+    return fadFirstOffset(db->pkgs);
 }
 
 int rpmdbNextRecNum(rpmdb db, unsigned int lastOffset) {
     /* 0 at end */
-    return faNextOffset(db->pkgs, lastOffset);
+    return fadNextOffset(db->pkgs, lastOffset);
 }
 
-static Header doGetRecord(rpmdb db, unsigned int offset, int pristine) {
+static Header doGetRecord(rpmdb db, unsigned int offset, int pristine)
+{
     Header h;
-    char ** fileList;
-    char ** newList;
+    const char ** fileNames;
     int fileCount = 0;
     int i;
 
-    (void)faLseek(db->pkgs, offset, SEEK_SET);
+    (void)Fseek(db->pkgs, offset, SEEK_SET);
 
-    h = headerRead(faFileno(db->pkgs), HEADER_MAGIC_NO);
+    h = headerRead(db->pkgs, HEADER_MAGIC_NO);
 
-    if (pristine) return h;
+    if (pristine || h == NULL) return h;
 
-    /* the RPM used to buildmuch of RH 5.1 could produce packages whose
+    /* the RPM used to build much of RH 5.1 could produce packages whose
        file lists did not have leading /'s. Now is a good time to fix
        that */
 
-    if (!headerGetEntryMinMemory(h, RPMTAG_FILENAMES, NULL, 
-			   (void **) &fileList, &fileCount)) return h;
+    /* If this tag isn't present, either no files are in the package or
+       we're dealing with a package that has just the compressed file name
+       list */
+    if (!headerGetEntryMinMemory(h, RPMTAG_OLDFILENAMES, NULL, 
+			   (void **) &fileNames, &fileCount)) return h;
 
     for (i = 0; i < fileCount; i++) 
-	if (*fileList[i] != '/') break;
+	if (*fileNames[i] != '/') break;
 
     if (i == fileCount) {
-	free(fileList);
-	return h;
+	free(fileNames);
+    } else {	/* bad header -- let's clean it up */
+	const char ** newFileNames = alloca(sizeof(*newFileNames) * fileCount);
+	for (i = 0; i < fileCount; i++) {
+	    char * newFileName = alloca(strlen(fileNames[i]) + 2);
+	    if (*fileNames[i] != '/') {
+		newFileName[0] = '/';
+		newFileName[1] = '\0';
+	    } else
+		newFileName[0] = '\0';
+	    strcat(newFileName, fileNames[i]);
+	    newFileNames[i] = newFileName;
+	}
+
+	free(fileNames);
+
+	headerModifyEntry(h, RPMTAG_OLDFILENAMES, RPM_STRING_ARRAY_TYPE, 
+			  newFileNames, fileCount);
     }
 
-    /* bad header -- let's clean it up */
-    newList = alloca(sizeof(*newList) * fileCount);
-    for (i = 0; i < fileCount; i++) {
-	newList[i] = alloca(strlen(fileList[i]) + 2);
-	if (*fileList[i] == '/')
-	    strcpy(newList[i], fileList[i]);
-	else
-	    sprintf(newList[i], "/%s", fileList[i]);
-    }
-
-    free(fileList);
-
-    headerModifyEntry(h, RPMTAG_FILENAMES, RPM_STRING_ARRAY_TYPE, 
-		      newList, fileCount);
+    /* The file list was moved to a more compressed format which not
+       only saves memory (nice), but gives fingerprinting a nice, fat
+       speed boost (very nice). Go ahead and convert old headers to
+       the new style (this is a noop for new headers) */
+    compressFilelist(h);
 
     return h;
 }
@@ -349,38 +358,60 @@ Header rpmdbGetRecord(rpmdb db, unsigned int offset)
 
 int rpmdbFindByFile(rpmdb db, const char * filespec, dbiIndexSet * matches)
 {
-    const char * basename;
+    const char * dirName;
+    const char * baseName;
     fingerPrint fp1, fp2;
     dbiIndexSet allMatches;
     int i, rc;
-    Header h;
-    char ** fileList;
+    fingerPrintCache fpc;
 
-    fp1 = fpLookup(filespec, 0);
+    if ((baseName = strrchr(filespec, '/')) != NULL) {
+    	char * t;
+	size_t len;
 
-    basename = strrchr(filespec, '/');
-    if (!basename) 
-	basename = filespec;
-    else
-	basename++;
+    	len = baseName - filespec + 1;
+	t = strncpy(alloca(len + 1), filespec, len);
+	t[len] = '\0';
+	dirName = t;
+	baseName++;
+    } else {
+	dirName = "";
+	baseName = filespec;
+    }
 
-    rc = dbiSearchIndex(db->fileIndex, basename, &allMatches);
-    if (rc) return rc;
+    fpc = fpCacheCreate(20);
+    fp1 = fpLookup(fpc, dirName, baseName, 1);
+
+    rc = dbiSearchIndex(db->fileIndex, baseName, &allMatches);
+    if (rc) {
+	fpCacheFree(fpc);
+	return rc;
+    }
 
     *matches = dbiCreateIndexRecord();
     i = 0;
     while (i < allMatches.count) {
+	const char ** baseNames, ** dirNames;
+	int_32 * dirIndexes;
+	Header h;
+
 	if ((h = rpmdbGetRecord(db, allMatches.recs[i].recOffset)) == NULL) {
 	    i++;
 	    continue;
 	}
 
-	headerGetEntryMinMemory(h, RPMTAG_FILENAMES, NULL, 
-				(void **) &fileList, NULL);
+	headerGetEntryMinMemory(h, RPMTAG_BASENAMES, NULL, 
+				(void **) &baseNames, NULL);
+	headerGetEntryMinMemory(h, RPMTAG_DIRINDEXES, NULL, 
+				(void **) &dirIndexes, NULL);
+	headerGetEntryMinMemory(h, RPMTAG_DIRNAMES, NULL, 
+				(void **) &dirNames, NULL);
 
 	do {
-	    fp2 = fpLookup(fileList[allMatches.recs[i].fileNumber], 1);
-	    if (FP_EQUAL(fp1, fp2)) 
+	    int num = allMatches.recs[i].fileNumber;
+
+	    fp2 = fpLookup(fpc, dirNames[dirIndexes[num]], baseNames[num], 1);
+	    if (FP_EQUAL(fp1, fp2))
 		dbiAppendIndexRecord(matches, allMatches.recs[i]);
 
 	    i++;
@@ -388,12 +419,15 @@ int rpmdbFindByFile(rpmdb db, const char * filespec, dbiIndexSet * matches)
 			((i == 0) || (allMatches.recs[i].recOffset == 
 				allMatches.recs[i - 1].recOffset)));
 
-	free(fileList);
+	free(baseNames);
+	free(dirNames);
 	headerFree(h);
     }
 
     dbiFreeIndexRecord(allMatches);
 
+    fpCacheFree(fpc);
+
     if (!matches->count) {
 	dbiFreeIndexRecord(*matches);
 	return 1;
@@ -462,10 +496,9 @@ int rpmdbRemove(rpmdb db, unsigned int offset, int tolerant)
     int type;
     unsigned int count;
     dbiIndexRecord rec;
-    char ** fileList, ** providesList, ** requiredbyList;
+    char ** baseNames, ** providesList, ** requiredbyList;
     char ** conflictList, ** triggerList;
     int i;
-    char * basename;
 
     /* structure assignment */
     rec = dbiReturnIndexRecordInstance(offset, 0);
@@ -542,28 +575,22 @@ int rpmdbRemove(rpmdb db, unsigned int offset, int tolerant)
 	free(conflictList);
     }
 
-    if (headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, 
+    if (headerGetEntry(h, RPMTAG_BASENAMES, &type, (void **) &baseNames, 
 	 &count)) {
 	for (i = 0; i < count; i++) {
-	    basename = strrchr(fileList[i], '/');
-	    if (!basename) 
-		basename = fileList[i];
-	    else
-		basename++;
-
 	    rpmMessage(RPMMESS_DEBUG, _("removing file index for %s\n"), 
-			basename);
+			baseNames[i]);
 	    /* structure assignment */
 	    rec = dbiReturnIndexRecordInstance(offset, i);
-	    removeIndexEntry(db->fileIndex, basename, rec, tolerant, 
+	    removeIndexEntry(db->fileIndex, baseNames[i], rec, tolerant, 
 			     "file index");
 	}
-	free(fileList);
+	free(baseNames);
     } else {
 	rpmMessage(RPMMESS_DEBUG, _("package has no files\n"));
     }
 
-    faFree(db->pkgs, offset);
+    fadFree(db->pkgs, offset);
 
     dbiSyncIndex(db->nameIndex);
     dbiSyncIndex(db->groupIndex);
@@ -602,7 +629,7 @@ int rpmdbAdd(rpmdb db, Header dbentry)
 {
     unsigned int dboffset;
     unsigned int i, j;
-    const char ** fileList;
+    const char ** baseNames;
     const char ** providesList;
     const char ** requiredbyList;
     const char ** conflictList;
@@ -612,6 +639,7 @@ int rpmdbAdd(rpmdb db, Header dbentry)
     int count = 0, providesCount = 0, requiredbyCount = 0, conflictCount = 0;
     int triggerCount = 0;
     int type;
+    int newSize;
     int rc = 0;
 
     headerGetEntry(dbentry, RPMTAG_NAME, &type, (void **) &name, &count);
@@ -620,8 +648,27 @@ int rpmdbAdd(rpmdb db, Header dbentry)
     if (!group) group = "Unknown";
 
     count = 0;
-    headerGetEntry(dbentry, RPMTAG_FILENAMES, &type, (void **) &fileList, 
-	           &count);
+
+    headerGetEntry(dbentry, RPMTAG_BASENAMES, &type, (void **) 
+		    &baseNames, &count);
+
+    if (_noDirTokens) {
+	const char ** newBaseNames;
+	char * data;
+	int len;
+	len = count * sizeof(*baseNames);
+	for (i = 0; i < count; i++)
+	    len += strlen(baseNames[i]) + 1;
+	newBaseNames = xmalloc(len);
+	data = (char *) newBaseNames + count;
+	for (i = 0; i < count; i++) {
+	    newBaseNames[i] = data;
+	    data = stpcpy(data, baseNames[i]);
+	    *data++ = '\0';
+	}
+	expandFilelist(dbentry);
+    }
+
     headerGetEntry(dbentry, RPMTAG_PROVIDENAME, &type, (void **) &providesList, 
 	           &providesCount);
     headerGetEntry(dbentry, RPMTAG_REQUIRENAME, &type, 
@@ -633,12 +680,16 @@ int rpmdbAdd(rpmdb db, Header dbentry)
 
     blockSignals();
 
-    dboffset = faAlloc(db->pkgs, headerSizeof(dbentry, HEADER_MAGIC_NO));
+    newSize = headerSizeof(dbentry, HEADER_MAGIC_NO);
+    dboffset = fadAlloc(db->pkgs, newSize);
     if (!dboffset) {
 	rc = 1;
     } else {
-	(void)faLseek(db->pkgs, dboffset, SEEK_SET);
-	rc = headerWrite(faFileno(db->pkgs), dbentry, HEADER_MAGIC_NO);
+	/* XXX TODO: set max. no. of bytes to write */
+	(void)Fseek(db->pkgs, dboffset, SEEK_SET);
+	fdSetContentLength(db->pkgs, newSize);
+	rc = headerWrite(db->pkgs, dbentry, HEADER_MAGIC_NO);
+	fdSetContentLength(db->pkgs, -1);
     }
 
     if (rc) {
@@ -671,14 +722,7 @@ int rpmdbAdd(rpmdb db, Header dbentry)
 	rc += addIndexEntry(db->providesIndex, providesList[i], dboffset, 0);
 
     for (i = 0; i < count; i++) {
-	const char * basename;
-	basename = strrchr(fileList[i], '/');
-	if (!basename) 
-	    basename = fileList[i];
-	else
-	    basename++;
-	if (*basename)
-	    rc += addIndexEntry(db->fileIndex, basename, dboffset, i);
+	rc += addIndexEntry(db->fileIndex, baseNames[i], dboffset, i);
     }
 
     dbiSyncIndex(db->nameIndex);
@@ -695,7 +739,7 @@ exit:
     if (providesCount) free(providesList);
     if (conflictCount) free(conflictList);
     if (triggerCount) free(triggerList);
-    if (count) free(fileList);
+    if (count) free(baseNames);
 
     return rc;
 }
@@ -703,7 +747,7 @@ exit:
 int rpmdbUpdateRecord(rpmdb db, int offset, Header newHeader)
 {
     Header oldHeader;
-    int oldSize;
+    int oldSize, newSize;
     int rc = 0;
 
     oldHeader = doGetRecord(db, offset, 1);
@@ -716,7 +760,11 @@ int rpmdbUpdateRecord(rpmdb db, int offset, Header newHeader)
     oldSize = headerSizeof(oldHeader, HEADER_MAGIC_NO);
     headerFree(oldHeader);
 
-    if (oldSize != headerSizeof(newHeader, HEADER_MAGIC_NO)) {
+    if (_noDirTokens)
+	expandFilelist(newHeader);
+
+    newSize = headerSizeof(newHeader, HEADER_MAGIC_NO);
+    if (oldSize != newSize) {
 	rpmMessage(RPMMESS_DEBUG, _("header changed size!"));
 	if (rpmdbRemove(db, offset, 1))
 	    return 1;
@@ -726,9 +774,12 @@ int rpmdbUpdateRecord(rpmdb db, int offset, Header newHeader)
     } else {
 	blockSignals();
 
-	(void)faLseek(db->pkgs, offset, SEEK_SET);
+	/* XXX TODO: set max. no. of bytes to write */
+	(void)Fseek(db->pkgs, offset, SEEK_SET);
 
-	rc = headerWrite(faFileno(db->pkgs), newHeader, HEADER_MAGIC_NO);
+	fdSetContentLength(db->pkgs, newSize);
+	rc = headerWrite(db->pkgs, newHeader, HEADER_MAGIC_NO);
+	fdSetContentLength(db->pkgs, -1);
 
 	unblockSignals();
     }
@@ -794,7 +845,7 @@ int rpmdbMoveDatabase(const char * rootdir, const char * olddbpath, const char *
     for (rpmdbfnp = rpmdb_filenames; *rpmdbfnp; rpmdbfnp++) {
 	sprintf(ofilename, "%s/%s/%s", rootdir, olddbpath, *rpmdbfnp);
 	sprintf(nfilename, "%s/%s/%s", rootdir, newdbpath, *rpmdbfnp);
-	if (rename(ofilename, nfilename)) rc = 1;
+	if (Rename(ofilename, nfilename)) rc = 1;
     }
 
     return rc;
@@ -827,16 +878,21 @@ int rpmdbFindFpList(rpmdb db, fingerPrint * fpList, dbiIndexSet * matchList,
     int i, j;
     int start, end;
     int num;
+    int_32 fc;
+    const char ** dirNames, ** baseNames;
+    const char ** fullBaseNames;
+    int_32 * dirIndexes, * fullDirIndexes;
+    fingerPrintCache fpc;
 
-    /* this may be worth batching by basename, but probably not as
-       basenames are quite unique as it is */
+    /* this may be worth batching by baseName, but probably not as
+       baseNames are quite unique as it is */
 
     intMatches = xcalloc(intMatchesAlloced, sizeof(*intMatches));
 
     /* Gather all matches from the database */
     for (i = 0; i < numItems; i++) {
 	dbiIndexSet matches;
-	switch (dbiSearchIndex(db->fileIndex, fpList[i].basename, &matches)) {
+	switch (dbiSearchIndex(db->fileIndex, fpList[i].baseName, &matches)) {
 	default:
 	    break;
 	case 2:
@@ -868,6 +924,8 @@ int rpmdbFindFpList(rpmdb db, fingerPrint * fpList, dbiIndexSet * matchList,
     for (i = 0; i < numItems; i++)
 	matchList[i] = dbiCreateIndexRecord();
 
+    fpc = fpCacheCreate(numIntMatches);
+
     /* For each set of files matched in a package ... */
     for (start = 0; start < numIntMatches; start = end) {
 	struct intMatch * im;
@@ -890,34 +948,41 @@ int rpmdbFindFpList(rpmdb db, fingerPrint * fpList, dbiIndexSet * matchList,
 	    return 1;
 	}
 
-	{   const char ** fullfl, **fl;
-	    int_32 fc;
-	   
-	    headerGetEntryMinMemory(h, RPMTAG_FILENAMES, NULL, 
-				(void **) &fullfl, &fc);
+	headerGetEntryMinMemory(h, RPMTAG_DIRNAMES, NULL, 
+			    (void **) &dirNames, NULL);
+	headerGetEntryMinMemory(h, RPMTAG_BASENAMES, NULL, 
+			    (void **) &fullBaseNames, &fc);
+	headerGetEntryMinMemory(h, RPMTAG_DIRINDEXES, NULL, 
+			    (void **) &fullDirIndexes, NULL);
 
-	    fl = xcalloc(num, sizeof(*fl));
-	    for (i = 0; i < num; i++)
-		fl[i] = fullfl[im[i].rec.fileNumber];
-	    free(fullfl);
-	    fps = xcalloc(num, sizeof(*fps));
-	    fpLookupList(fl, fps, num, 1);
-	    free(fl);
+	baseNames = xcalloc(num, sizeof(*baseNames));
+	dirIndexes = xcalloc(num, sizeof(*dirIndexes));
+	for (i = 0; i < num; i++) {
+	    baseNames[i] = fullBaseNames[im[i].rec.fileNumber];
+	    dirIndexes[i] = fullDirIndexes[im[i].rec.fileNumber];
 	}
 
+	fps = xcalloc(num, sizeof(*fps));
+	fpLookupList(fpc, dirNames, baseNames, dirIndexes, num, fps);
+
+	free(dirNames);
+	free(fullBaseNames);
+	free(baseNames);
+	free(dirIndexes);
+
 	/* Add db (recnum,filenum) to list for fingerprint matches. */
 	for (i = 0; i < num; i++) {
 	    j = im[i].fpNum;
-	    if (FP_EQUAL(fps[i], fpList[j]))
+	    if (FP_EQUAL_DIFFERENT_CACHE(fps[i], fpList[j]))
 		dbiAppendIndexRecord(&matchList[j], im[i].rec);
 	}
 
 	headerFree(h);
 
 	free(fps);
-
     }
 
+    fpCacheFree(fpc);
     free(intMatches);
 
     return 0;
