@@ -4,15 +4,7 @@
 #include <rpmmacro.h>
 #include <rpmurl.h>
 
-#ifdef DYING
-#include "build/rpmbuild.h"
-
-static void printHash(const unsigned long amount, const unsigned long total);
-static void * showProgress(const Header h, const rpmCallbackType what, 
-			   const unsigned long amount, 
-			   const unsigned long total,
-			   const void * pkgKey, void * data);
-#endif
+#include "misc.h"
 
 static int hashesPrinted = 0;
 
@@ -46,11 +38,16 @@ static void * showProgress(const Header h, const rpmCallbackType what,
 
     switch (what) {
       case RPMCALLBACK_INST_OPEN_FILE:
-	fd = fdOpen(filename, O_RDONLY, 0);
+	fd = Fopen(filename, "r.ufdio");
+	fd = fdLink(fd, "persist (showProgress)");
 	return fd;
 
       case RPMCALLBACK_INST_CLOSE_FILE:
-	fdClose(fd);
+	fd = fdFree(fd, "persist (showProgress)");
+	if (fd) {
+	    Fclose(fd);
+	    fd = NULL;
+	}
 	break;
 
       case RPMCALLBACK_INST_START:
@@ -93,27 +90,28 @@ static void * showProgress(const Header h, const rpmCallbackType what,
     return rc;
 }	
 
-int rpmInstall(const char * rootdir, const char ** argv, int transFlags, 
+int rpmInstall(const char * rootdir, const char ** fileArgv, int transFlags, 
 	      int interfaceFlags, int probFilter, 
-	      rpmRelocation * relocations) {
+	      rpmRelocation * relocations)
+{
     rpmdb db = NULL;
     FD_t fd;
     int i;
     int mode, rc, major;
-    const char ** packages, ** tmpPackages;
-    const char ** filename;
-    int numPackages;
-    int numTmpPackages = 0, numBinaryPackages = 0, numSourcePackages = 0;
+    const char ** pkgURL = NULL;
+    const char ** tmppkgURL = NULL;
+    const char ** fileURL;
+    int numPkgs;
+    int numTmpPkgs = 0, numRPMS = 0, numSRPMS = 0;
     int numFailed = 0;
     Header h;
     int isSource;
     rpmTransactionSet rpmdep = NULL;
     int numConflicts;
     int stopInstall = 0;
-    size_t nb;
     int notifyFlags = interfaceFlags | (rpmIsVerbose() ? INSTALL_LABEL : 0 );
     int dbIsOpen = 0;
-    const char ** sourcePackages;
+    const char ** sourceURL;
     rpmRelocation * defaultReloc;
 
     if (transFlags & RPMTRANS_FLAG_TEST) 
@@ -126,71 +124,101 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
     if (defaultReloc && !defaultReloc->newPath) defaultReloc = NULL;
 
     rpmMessage(RPMMESS_DEBUG, _("counting packages to install\n"));
-    for (filename = argv, numPackages = 0; *filename; filename++, numPackages++)
+    for (fileURL = fileArgv, numPkgs = 0; *fileURL; fileURL++, numPkgs++)
 	;
 
-    rpmMessage(RPMMESS_DEBUG, _("found %d packages\n"), numPackages);
+    rpmMessage(RPMMESS_DEBUG, _("found %d packages\n"), numPkgs);
 
-    nb = (numPackages + 1) * sizeof(char *);
-    packages = alloca(nb);
-    memset(packages, 0, nb);
-    tmpPackages = alloca(nb);
-    memset(tmpPackages, 0, nb);
-    nb = (numPackages + 1) * sizeof(Header);
+    pkgURL = xcalloc( (numPkgs + 1), sizeof(*pkgURL) );
+    tmppkgURL = xcalloc( (numPkgs + 1), sizeof(*tmppkgURL) );
 
     rpmMessage(RPMMESS_DEBUG, _("looking for packages to download\n"));
-    for (filename = argv, i = 0; *filename; filename++) {
+    for (fileURL = fileArgv, i = 0; *fileURL; fileURL++) {
 
-	switch (urlIsURL(*filename)) {
+	switch (urlIsURL(*fileURL)) {
 	case URL_IS_FTP:
 	case URL_IS_HTTP:
-	case URL_IS_PATH:
 	{   int myrc;
+	    int j;
 	    const char *tfn;
+	    const char ** argv;
+	    int argc;
 
-	    if (rpmIsVerbose())
-		fprintf(stdout, _("Retrieving %s\n"), *filename);
-
-	    {	char tfnbuf[64];
-		strcpy(tfnbuf, "rpm-xfer.XXXXXX");
-		/* XXX watchout for rootdir = NULL */
-		tfn = rpmGetPath( (rootdir ? rootdir : ""),
-		    "%{_tmppath}/", mktemp(tfnbuf), NULL);
+	    myrc = rpmGlob(*fileURL, &argc, &argv);
+	    if (myrc) {
+		rpmMessage(RPMMESS_ERROR, 
+			_("skipping %s - rpmGlob failed(%d)\n"),
+			*fileURL, myrc);
+		numFailed++;
+		pkgURL[i] = NULL;
+		break;
+	    }
+	    if (argc > 1) {
+		numPkgs += argc - 1;
+		pkgURL = xrealloc(pkgURL, (numPkgs + 1) * sizeof(*pkgURL));
+		tmppkgURL = xrealloc(tmppkgURL, (numPkgs + 1) * sizeof(*tmppkgURL));
 	    }
 
-	    rpmMessage(RPMMESS_DEBUG, _(" ... as %s\n"), tfn);
-	    myrc = urlGetFile(*filename, tfn);
-	    if (myrc < 0) {
-		rpmMessage(RPMMESS_ERROR, 
+	    for (j = 0; j < argc; j++) {
+
+		if (rpmIsVerbose())
+		    fprintf(stdout, _("Retrieving %s\n"), argv[j]);
+
+		{   char tfnbuf[64];
+		    strcpy(tfnbuf, "rpm-xfer.XXXXXX");
+		    /*@-unrecog@*/ mktemp(tfnbuf) /*@=unrecog@*/;
+		    tfn = rpmGenPath(rootdir, "%{_tmppath}/", tfnbuf);
+		}
+
+		/* XXX undefined %{name}/%{version}/%{release} here */
+		/* XXX %{_tmpdir} does not exist */
+		rpmMessage(RPMMESS_DEBUG, _(" ... as %s\n"), tfn);
+		myrc = urlGetFile(argv[j], tfn);
+		if (myrc < 0) {
+		    rpmMessage(RPMMESS_ERROR, 
 			_("skipping %s - transfer failed - %s\n"), 
-			*filename, ftpStrerror(myrc));
-		numFailed++;
-		packages[i] = NULL;
-		xfree(tfn);
-	    } else {
-		tmpPackages[numTmpPackages++] = packages[i++] = tfn;
+			argv[j], ftpStrerror(myrc));
+		    numFailed++;
+		    pkgURL[i] = NULL;
+		    xfree(tfn);
+		} else {
+		    tmppkgURL[numTmpPkgs++] = pkgURL[i++] = tfn;
+		}
+	    }
+	    if (argv) {
+		for (j = 0; j < argc; j++)
+		    xfree(argv[j]);
+		xfree(argv);
+		argv = NULL;
 	    }
 	}   break;
+	case URL_IS_PATH:
 	default:
-	    packages[i++] = *filename;
+	    pkgURL[i++] = *fileURL;
 	    break;
 	}
     }
+    pkgURL[i] = NULL;
+    tmppkgURL[numTmpPkgs] = NULL;
 
-    sourcePackages = alloca(sizeof(*sourcePackages) * i);
+    sourceURL = alloca(sizeof(*sourceURL) * i);
 
-    rpmMessage(RPMMESS_DEBUG, _("retrieved %d packages\n"), numTmpPackages);
+    rpmMessage(RPMMESS_DEBUG, _("retrieved %d packages\n"), numTmpPkgs);
 
     /* Build up the transaction set. As a special case, v1 source packages
        are installed right here, only because they don't have headers and
        would create all sorts of confusion later. */
 
-    for (filename = packages; *filename; filename++) {
-	fd = fdOpen(*filename, O_RDONLY, 0);
-	if (fdFileno(fd) < 0) {
-	    rpmMessage(RPMMESS_ERROR, _("cannot open file %s\n"), *filename);
+    for (fileURL = pkgURL; *fileURL; fileURL++) {
+	const char * fileName;
+	(void) urlPath(*fileURL, &fileName);
+	fd = Fopen(*fileURL, "r.ufdio");
+	if (fd == NULL || Ferror(fd)) {
+	    rpmMessage(RPMMESS_ERROR, _("cannot open file %s: %s\n"), *fileURL,
+		Fstrerror(fd));
+	    if (fd) Fclose(fd);
 	    numFailed++;
-	    packages[i] = NULL;
+	    pkgURL[i] = NULL;
 	    continue;
 	}
 
@@ -198,20 +226,20 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 
 	switch (rc) {
 	case 1:
-	    fdClose(fd);
+	    Fclose(fd);
 	    rpmMessage(RPMMESS_ERROR, 
 			_("%s does not appear to be a RPM package\n"), 
-			*filename);
+			*fileURL);
 	    break;
 	default:
-	    rpmMessage(RPMMESS_ERROR, _("%s cannot be installed\n"), *filename);
+	    rpmMessage(RPMMESS_ERROR, _("%s cannot be installed\n"), *fileURL);
 	    numFailed++;
-	    packages[i] = NULL;
+	    pkgURL[i] = NULL;
 	    break;
 	case 0:
 	    if (isSource) {
-		sourcePackages[numSourcePackages++] = *filename;
-		fdClose(fd);
+		sourceURL[numSRPMS++] = fileName;
+		Fclose(fd);
 	    } else {
 		if (!dbIsOpen) {
 		    if (rpmdbOpen(rootdir, &db, mode, 0644)) {
@@ -228,44 +256,44 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 		}
 
 		if (defaultReloc) {
-		    char ** paths;
-		    char * name;
+		    const char ** paths;
 		    int c;
 
 		    if (headerGetEntry(h, RPMTAG_PREFIXES, NULL,
 				       (void **) &paths, &c) && (c == 1)) {
 			defaultReloc->oldPath = xstrdup(paths[0]);
-			free(paths);
+			xfree(paths);
 		    } else {
-			headerGetEntry(h, RPMTAG_NAME, NULL, (void **) &name,
-				       NULL);
+			const char * name;
+			headerNVR(h, &name, NULL, NULL);
 			rpmMessage(RPMMESS_ERROR, 
 			       _("package %s is not relocateable\n"), name);
 
-			return numPackages;
+			goto errxit;
+			/*@notreached@*/
 		    }
 		}
 
-		rc = rpmtransAddPackage(rpmdep, h, NULL, *filename,
+		rc = rpmtransAddPackage(rpmdep, h, NULL, fileName,
 			       (interfaceFlags & INSTALL_UPGRADE) != 0,
 			       relocations);
 
 		headerFree(h);	/* XXX reference held by transaction set */
-		fdClose(fd);
+		Fclose(fd);
 
 		switch(rc) {
 		case 0:
 			break;
 		case 1:
 			rpmMessage(RPMMESS_ERROR, 
-			    _("error reading from file %s\n"), *filename);
-			return numPackages;
+			    _("error reading from file %s\n"), *fileURL);
+			goto errxit;
 			/*@notreached@*/ break;
 		case 2:
 			rpmMessage(RPMMESS_ERROR, 
 			    _("file %s requires a newer version of RPM\n"),
-			    *filename);
-			return numPackages;
+			    *fileURL);
+			goto errxit;
 			/*@notreached@*/ break;
 		}
 
@@ -274,19 +302,19 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 		    defaultReloc->oldPath = NULL;
 		}
 
-		numBinaryPackages++;
+		numRPMS++;
 	    }
 	    break;
 	}
     }
 
     rpmMessage(RPMMESS_DEBUG, _("found %d source and %d binary packages\n"), 
-		numSourcePackages, numBinaryPackages);
+		numSRPMS, numRPMS);
 
-    if (numBinaryPackages && !(interfaceFlags & INSTALL_NODEPS)) {
+    if (numRPMS && !(interfaceFlags & INSTALL_NODEPS)) {
 	struct rpmDependencyConflict * conflicts;
 	if (rpmdepCheck(rpmdep, &conflicts, &numConflicts)) {
-	    numFailed = numPackages;
+	    numFailed = numPkgs;
 	    stopInstall = 1;
 	}
 
@@ -294,19 +322,19 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 	    rpmMessage(RPMMESS_ERROR, _("failed dependencies:\n"));
 	    printDepProblems(stderr, conflicts, numConflicts);
 	    rpmdepFreeConflicts(conflicts, numConflicts);
-	    numFailed = numPackages;
+	    numFailed = numPkgs;
 	    stopInstall = 1;
 	}
     }
 
-    if (numBinaryPackages && !(interfaceFlags & INSTALL_NOORDER)) {
+    if (numRPMS && !(interfaceFlags & INSTALL_NOORDER)) {
 	if (rpmdepOrder(rpmdep)) {
-	    numFailed = numPackages;
+	    numFailed = numPkgs;
 	    stopInstall = 1;
 	}
     }
 
-    if (numBinaryPackages && !stopInstall) {
+    if (numRPMS && !stopInstall) {
 	rpmProblemSet probs = NULL;
 ;
 	rpmMessage(RPMMESS_DEBUG, _("installing binary packages\n"));
@@ -314,8 +342,8 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 				    NULL, &probs, transFlags, probFilter);
 
 	if (rc < 0) {
-	    numFailed += numBinaryPackages;
-	} else if (rc) {
+	    numFailed += numRPMS;
+	} else if (rc > 0) {
 	    numFailed += rc;
 	    rpmProblemSetPrint(stderr, probs);
 	}
@@ -323,14 +351,15 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 	if (probs) rpmProblemSetFree(probs);
     }
 
-    if (numBinaryPackages) rpmtransFree(rpmdep);
+    if (numRPMS) rpmtransFree(rpmdep);
 
-    if (numSourcePackages && !stopInstall) {
-	for (i = 0; i < numSourcePackages; i++) {
-	    fd = fdOpen(sourcePackages[i], O_RDONLY, 0);
-	    if (fdFileno(fd) < 0) {
-		rpmMessage(RPMMESS_ERROR, _("cannot open file %s\n"), 
-			   sourcePackages[i]);
+    if (numSRPMS && !stopInstall) {
+	for (i = 0; i < numSRPMS; i++) {
+	    fd = Fopen(sourceURL[i], "r.ufdio");
+	    if (fd == NULL || Ferror(fd)) {
+		rpmMessage(RPMMESS_ERROR, _("cannot open file %s: %s\n"), 
+			   sourceURL[i], Fstrerror(fd));
+		if (fd) Fclose(fd);
 		continue;
 	    }
 
@@ -338,20 +367,33 @@ int rpmInstall(const char * rootdir, const char ** argv, int transFlags,
 		numFailed += rpmInstallSourcePackage(rootdir, fd, NULL,
 				showProgress, (void *) notifyFlags, NULL);
 
-	    fdClose(fd);
+	    Fclose(fd);
 	}
     }
 
-    for (i = 0; i < numTmpPackages; i++) {
-	unlink(tmpPackages[i]);
-	xfree(tmpPackages[i]);
+    for (i = 0; i < numTmpPkgs; i++) {
+	Unlink(tmppkgURL[i]);
+	xfree(tmppkgURL[i]);
     }
+    xfree(tmppkgURL);
+    xfree(pkgURL);
 
     /* FIXME how do we close our various fd's? */
 
     if (dbIsOpen) rpmdbClose(db);
 
     return numFailed;
+
+errxit:
+    if (tmppkgURL) {
+	for (i = 0; i < numTmpPkgs; i++)
+	    xfree(tmppkgURL[i]);
+	xfree(tmppkgURL);
+    }
+    if (pkgURL)
+	xfree(pkgURL);
+    if (dbIsOpen) rpmdbClose(db);
+    return numPkgs;
 }
 
 int rpmErase(const char * rootdir, const char ** argv, int transFlags,
@@ -454,9 +496,10 @@ int rpmInstallSource(const char * rootdir, const char * arg, const char ** specF
     FD_t fd;
     int rc;
 
-    fd = fdOpen(arg, O_RDONLY, 0);
-    if (fdFileno(fd) < 0) {
-	rpmMessage(RPMMESS_ERROR, _("cannot open %s\n"), arg);
+    fd = Fopen(arg, "r.ufdio");
+    if (fd == NULL || Ferror(fd)) {
+	rpmMessage(RPMMESS_ERROR, _("cannot open %s: %s\n"), arg, Fstrerror(fd));
+	if (fd) Fclose(fd);
 	return 1;
     }
 
@@ -471,7 +514,7 @@ int rpmInstallSource(const char * rootdir, const char * arg, const char ** specF
 	if (cookie && *cookie) free(*cookie);
     }
 
-    fdClose(fd);
+    Fclose(fd);
 
     return rc;
 }
