@@ -22,6 +22,17 @@ int rpmReadPackageHeader(FD_t fd, /*@out@*/ Header * hdr,
 int headerNVR(Header h, /*@out@*/ const char **np, /*@out@*/ const char **vp,
 	/*@out@*/ const char **rp);
 
+void	rpmBuildFileList(Header h, /*@out@*/ const char *** fileListPtr, 
+			/*@out@*/ int * fileCountPtr);
+
+/*
+ * XXX This is a "dressed" entry to headerGetEntry to do:
+ *	1) DIRNAME/BASENAME/DIRINDICES -> FILENAMES tag conversions.
+ *	2) i18n lookaside (if enabled).
+ */
+int rpmHeaderGetEntry(Header h, int_32 tag, /*@out@*/ int_32 *type,
+        /*@out@*/ void **p, /*@out@*/int_32 *c);
+
    /* 0 = success */
    /* 1 = bad magic */
    /* 2 = error */
@@ -64,7 +75,7 @@ extern const struct headerSprintfExtension rpmHeaderFormats[];
 #define	RPMTAG_POSTIN			1024
 #define	RPMTAG_PREUN			1025
 #define	RPMTAG_POSTUN			1026
-#define	RPMTAG_FILENAMES		1027
+#define	RPMTAG_OLDFILENAMES		1027 /* obsolete */
 #define	RPMTAG_FILESIZES		1028
 #define	RPMTAG_FILESTATES		1029
 #define	RPMTAG_FILEMODES		1030
@@ -75,7 +86,7 @@ extern const struct headerSprintfExtension rpmHeaderFormats[];
 #define	RPMTAG_FILEMD5S			1035
 #define	RPMTAG_FILELINKTOS		1036
 #define	RPMTAG_FILEFLAGS		1037
-#define	RPMTAG_ROOT			1038
+#define	RPMTAG_ROOT			1038 /* obsolete */
 #define	RPMTAG_FILEUSERNAME		1039
 #define	RPMTAG_FILEGROUPNAME		1040
 #define	RPMTAG_EXCLUDE			1041 /* internal - depricated */
@@ -135,9 +146,9 @@ extern const struct headerSprintfExtension rpmHeaderFormats[];
 #define	RPMTAG_TRIGGERPOSTUN		1102 /* internal */
 #define	RPMTAG_AUTOREQ			1103 /* internal */
 #define	RPMTAG_AUTOPROV			1104 /* internal */
-#define	RPMTAG_CAPABILITY		1105
+#define	RPMTAG_CAPABILITY		1105 /* unused */
 #define	RPMTAG_SOURCEPACKAGE		1106 /* internal */
-#define	RPMTAG_ORIGFILENAMES		1107
+#define	RPMTAG_OLDORIGFILENAMES		1107 /* obsolete */
 #define	RPMTAG_BUILDPREREQ		1108 /* internal */
 #define	RPMTAG_BUILDREQUIRES		1109 /* internal */
 #define	RPMTAG_BUILDCONFLICTS		1110 /* internal */
@@ -146,8 +157,14 @@ extern const struct headerSprintfExtension rpmHeaderFormats[];
 #define	RPMTAG_PROVIDEVERSION		1113
 #define	RPMTAG_OBSOLETEFLAGS		1114
 #define	RPMTAG_OBSOLETEVERSION		1115
+#define	RPMTAG_DIRINDEXES		1116
+#define	RPMTAG_BASENAMES		1117
+#define	RPMTAG_DIRNAMES			1118
+#define	RPMTAG_ORIGDIRINDEXES		1119 /* internal */
+#define	RPMTAG_ORIGBASENAMES		1120 /* internal */
+#define	RPMTAG_ORIGDIRNAMES		1121 /* internal */
 
-#define	RPMTAG_FIRSTFREE_TAG		1116 /* internal */
+#define	RPMTAG_FIRSTFREE_TAG		1122 /* internal */
 #define	RPMTAG_EXTERNAL_TAG		1000000
 
 #define	RPMFILE_STATE_NORMAL 		0
@@ -260,7 +277,7 @@ extern const struct headerSprintfExtension rpmHeaderFormats[];
 
 #define	xfree(_p)	free((void *)_p)
 
-char * rpmGetVar(int var);
+const char * rpmGetVar(int var);
 void rpmSetVar(int var, const char *val);
 
 /** rpmrc.c **/
@@ -273,15 +290,15 @@ void rpmSetVar(int var, const char *val);
 
 int rpmReadConfigFiles(const char * file, const char * target);
 int rpmReadRC(const char * file);
-void rpmGetArchInfo( /*@out@*/ char ** name, /*@out@*/ int * num);
-void rpmGetOsInfo( /*@out@*/ char ** name, /*@out@*/ int * num);
-int rpmMachineScore(int type, char * name);
+void rpmGetArchInfo( /*@out@*/ const char ** name, /*@out@*/ int * num);
+void rpmGetOsInfo( /*@out@*/ const char ** name, /*@out@*/ int * num);
+int rpmMachineScore(int type, const char * name);
 int rpmShowRC(FILE *f);
 void rpmSetTables(int archTable, int osTable);  /* only used by build code */
 /* if either are NULL, they are set to the default value (munged uname())
    pushed through a translation table (if appropriate) */
 void rpmSetMachine(const char * arch, const char * os);
-void rpmGetMachine( /*@out@*/ char **arch, /*@out@*/ char **os);
+void rpmGetMachine( /*@out@*/ const char **arch, /*@out@*/ const char **os);
 void rpmFreeRpmrc(void);
 
 /** **/
@@ -300,6 +317,8 @@ typedef void * (*rpmCallbackFunction)(const Header h,
 				      const unsigned long total,
 				      const void * pkgKey, void * data);
 
+void	urlSetCallback(rpmCallbackFunction notify, void *notifyData, int notifyCount);
+
 int rpmdbOpen (const char * root, /*@out@*/ rpmdb * dbp, int mode, int perms);
     /* 0 on error */
 int rpmdbInit(const char * root, int perms);
@@ -307,7 +326,7 @@ int rpmdbInit(const char * root, int perms);
 void rpmdbClose ( /*@only@*/ rpmdb db);
 /* Databases like this should only have rpmdb*RecNum and rpmdbGetRecord
    used on them. Anything else could fail! */
-int rpmdbOpenForTraversal(const char * prefix, rpmdb * rpmdbp);
+int rpmdbOpenForTraversal(const char * prefix, /*@out@*/ rpmdb * rpmdbp);
 
 int rpmdbFirstRecNum(rpmdb db);
 int rpmdbNextRecNum(rpmdb db, unsigned int lastOffset);
@@ -364,7 +383,7 @@ struct rpmDependencyConflict {
        which isn't needed as well */
     char * needsName, * needsVersion;
     int needsFlags;
-    const void * suggestedPackage;			/* NULL if none */
+    /*@observer@*/ /*@null@*/ const void * suggestedPackage; /* NULL if none */
     enum { RPMDEP_SENSE_REQUIRES, RPMDEP_SENSE_CONFLICTS } sense;
 } ;
 
@@ -380,8 +399,10 @@ struct rpmDependencyConflict {
    returns 0 on success, 1 on I/O error, 2 if the package needs capabilities
    which are not implemented */
 int rpmtransAddPackage(rpmTransactionSet rpmdep, Header h, FD_t fd,
-			const void * key, int update, rpmRelocation * relocs);
-void rpmtransAvailablePackage(rpmTransactionSet rpmdep, Header h, void * key);
+		/*@owned@*/ const void * key, int update,
+		rpmRelocation * relocs);
+void rpmtransAvailablePackage(rpmTransactionSet rpmdep, Header h,
+		/*@owned@*/ const void * key);
 void rpmtransRemovePackage(rpmTransactionSet rpmdep, int dboffset);
 void rpmtransFree( /*@only@*/ rpmTransactionSet rpmdep);
 void rpmtransSetScriptFd(rpmTransactionSet ts, FD_t fd);
@@ -421,10 +442,10 @@ typedef enum rpmProblemType_e { RPMPROB_BADARCH,
 
 typedef /*@abstract@*/ struct rpmProblem_s {
     Header h, altH;
-    const void * key;
+    /*@dependent@*/ const void * key;
     rpmProblemType type;
     int ignoreProblem;
-    char * str1;
+    /*@only@*/ const char * str1;
     unsigned long ulong1;
 } rpmProblem;
 
@@ -645,6 +666,7 @@ void rpmFreeSignature(Header h);
 int rpmVerifySignature(const char *file, int_32 sigTag, void *sig, int count,
 		       char *result);
 
+void freeFilesystems(void);
 int rpmGetFilesystemList( /*@out@*/ const char *** listptr, /*@out@*/int * num);
 int rpmGetFilesystemUsage(const char ** filelist, int_32 * fssizes,
 	int numFiles, /*@out@*/ uint_32 ** usagesPtr, int flags);
@@ -678,8 +700,8 @@ int showMatches(QVA_t *qva, rpmdb db, dbiIndexSet matches, QVF_t showPackage);
 #define QUERY_FOR_CONFIG	(1 << 4)
 #define QUERY_FOR_DUMPFILES     (1 << 8)
 
-const char *tagName(int tag);
-extern char *specedit;
+/*@observer@*/ const char *const tagName(int tag);
+extern int specedit;
 extern struct poptOption rpmQueryPoptTable[];
 
 void rpmDisplayQueryTags(FILE * f);
@@ -719,10 +741,6 @@ int rpmInstallSource(const char * prefix, const char * arg, const char ** specFi
 int rpmErase(const char * rootdir, const char ** argv, int uninstallFlags, 
 		 int interfaceFlags);
 
-void printDepFlags(FILE * fp, const char * version, int flags);
-void printDepProblems(FILE * fp, struct rpmDependencyConflict * conflicts,
-			     int numConflicts);
-
 /* ==================================================================== */
 /* --- checksig/resign */
 
