@@ -16,9 +16,9 @@
 static const char *defrcfiles = LIBRPMRC_FILENAME ":/etc/rpmrc:~/.rpmrc";
 
 struct machCacheEntry {
-    char * name;
+    const char * name;
     int count;
-    char ** equivs;
+    const char ** equivs;
     int visited;
 };
 
@@ -28,7 +28,7 @@ struct machCache {
 };
 
 struct machEquivInfo {
-    char * name;
+    const char * name;
     int score;
 };
 
@@ -38,14 +38,14 @@ struct machEquivTable {
 };
 
 struct rpmvarValue {
-    char * value;
+    const char * value;
     /* eventually, this arch will be replaced with a generic condition */
-    char * arch;
+    const char * arch;
     struct rpmvarValue * next;
 };
 
 struct rpmOption {
-    char * name;
+    const char * name;
     int var;
     int archSpecific, required, macroize, localize;
     struct rpmOptionValue * value;
@@ -66,8 +66,9 @@ struct canonEntry {
 
    for giggles, 'key'_canon, 'key'_compat, and 'key'_canon will also work */
 struct tableType {
-    char * key;
-    int hasCanon, hasTranslate;
+    const char * const key;
+    const int hasCanon;
+    const int hasTranslate;
     struct machEquivTable equiv;
     struct machCache cache;
     struct defaultEntry * defaults;
@@ -102,46 +103,12 @@ static int optionTableSize = sizeof(optionTable) / sizeof(*optionTable);
 static char * current[2];
 static int currTables[2] = { RPM_MACHTABLE_INSTOS, RPM_MACHTABLE_INSTARCH };
 static struct rpmvarValue values[RPMVAR_NUM];
+static int defaultsInitialized = 0;
 
 /* prototypes */
-static void defaultMachine(/*@out@*/char ** arch, /*@out@*/char ** os);
-static int doReadRC(FD_t fd, const char * filename);
-static int optionCompare(const void * a, const void * b);
-static int addCanon(struct canonEntry **table, int *tableLen, char *line,
-			const char *fn, int lineNum);
-static int addDefault(struct defaultEntry **table, int *tableLen, char *line,
-			const char *fn, int lineNum);
-static void freeRpmVar(struct rpmvarValue * orig);
+static int doReadRC(FD_t fd, const char * urlfn);
 static void rpmSetVarArch(int var, const char * val, const char * arch);
-static struct canonEntry *lookupInCanonTable(char *name,
-					   struct canonEntry *table,
-					   int tableLen);
-static const char *lookupInDefaultTable(const char *name,
-				  struct defaultEntry *table,
-				  int tableLen);
-
-static void setVarDefault(int var, const char *macroname, const char *val, const char *body);
-static void setPathDefault(int var, const char *macroname, const char *subdir);
-static void setDefaults(void);
-
-static void rebuildCompatTables(int type, char *name);
-
-/* compatiblity tables */
-static int machCompatCacheAdd(char * name, const char * fn, int linenum,
-				struct machCache * cache);
-static struct machCacheEntry * machCacheFindEntry(struct machCache * cache,
-						  char * key);
-static struct machEquivInfo * machEquivSearch(
-		struct machEquivTable * table, char * name);
-static void machAddEquiv(struct machEquivTable * table, char * name,
-			   int distance);
-static void machCacheEntryVisit(struct machCache * cache,
-				  struct machEquivTable * table,
-				  char * name,
-	  			  int distance);
-static void machFindEquivs(struct machCache * cache,
-			     struct machEquivTable * table,
-			     char * key);
+static void rebuildCompatTables(int type, const char *name);
 
 static int optionCompare(const void * a, const void * b) {
     return strcasecmp(((struct rpmOption *) a)->name,
@@ -151,7 +118,7 @@ static int optionCompare(const void * a, const void * b) {
 static void rpmRebuildTargetVars(const char **target, const char ** canontarget);
 
 static struct machCacheEntry * machCacheFindEntry(struct machCache * cache,
-						  char * key)
+						  const char * key)
 {
     int i;
 
@@ -194,8 +161,9 @@ static int machCompatCacheAdd(char * name, const char * fn, int linenum,
 	entry = machCacheFindEntry(cache, name);
 	if (entry) {
 	    for (i = 0; i < entry->count; i++)
-		free(entry->equivs[i]);
-	    if (entry->count) free(entry->equivs);
+		xfree(entry->equivs[i]);
+	    xfree(entry->equivs);
+	    entry->equivs = NULL;
 	    entry->count = 0;
 	}
     }
@@ -229,7 +197,7 @@ static int machCompatCacheAdd(char * name, const char * fn, int linenum,
 }
 
 static struct machEquivInfo * machEquivSearch(
-		struct machEquivTable * table, char * name)
+		struct machEquivTable * table, const char * name)
 {
     int i;
 
@@ -248,7 +216,7 @@ static struct machEquivInfo * machEquivSearch(
     return NULL;
 }
 
-static void machAddEquiv(struct machEquivTable * table, char * name,
+static void machAddEquiv(struct machEquivTable * table, const char * name,
 			   int distance)
 {
     struct machEquivInfo * equiv;
@@ -259,16 +227,16 @@ static void machAddEquiv(struct machEquivTable * table, char * name,
 	    table->list = xrealloc(table->list, (table->count + 1)
 				    * sizeof(*table->list));
 	else
-	    table->list = xmalloc(sizeof(*table->list)); /* XXX memory leak */
+	    table->list = xmalloc(sizeof(*table->list));
 
-	table->list[table->count].name = xstrdup(name); /* XXX memory leak */
+	table->list[table->count].name = xstrdup(name);
 	table->list[table->count++].score = distance;
     }
 }
 
 static void machCacheEntryVisit(struct machCache * cache,
 				  struct machEquivTable * table,
-				  char * name,
+				  const char * name,
 	  			  int distance)
 {
     struct machCacheEntry * entry;
@@ -290,27 +258,38 @@ static void machCacheEntryVisit(struct machCache * cache,
 
 static void machFindEquivs(struct machCache * cache,
 			     struct machEquivTable * table,
-			     char * key)
+			     const char * key)
 {
     int i;
 
     for (i = 0; i < cache->size; i++)
 	cache->cache[i].visited = 0;
 
+    while (table->count > 0) {
+	xfree(table->list[--table->count].name);
+	table->list[table->count].name = NULL;
+    }
     table->count = 0;
+    if (table->list) xfree(table->list);
+    table->list = NULL;
 
-    /* We have a general graph built using strings instead of pointers.
-       Yuck. We have to start at a point at traverse it, remembering how
-       far away everything is. */
+    /*
+     *	We have a general graph built using strings instead of pointers.
+     *	Yuck. We have to start at a point at traverse it, remembering how
+     *	far away everything is.
+     */
     machAddEquiv(table, key, 1);
     machCacheEntryVisit(cache, table, key, 2);
 }
 
-static int addCanon(struct canonEntry **table, int *tableLen, char *line,
-		    const char *fn, int lineNum)
+static int addCanon(struct canonEntry ** table, int * tableLen, char * line,
+		    const char * fn, int lineNum)
 {
     struct canonEntry *t;
     char *s, *s1;
+    const char * tname;
+    const char * tshort_name;
+    int tnum;
 
     if (! *tableLen) {
 	*tableLen = 2;
@@ -321,10 +300,10 @@ static int addCanon(struct canonEntry **table, int *tableLen, char *line,
     }
     t = & ((*table)[*tableLen - 2]);
 
-    t->name = strtok(line, ": \t");
-    t->short_name = strtok(NULL, " \t");
+    tname = strtok(line, ": \t");
+    tshort_name = strtok(NULL, " \t");
     s = strtok(NULL, " \t");
-    if (! (t->name && t->short_name && s)) {
+    if (! (tname && tshort_name && s)) {
 	rpmError(RPMERR_RPMRC, _("Incomplete data line at %s:%d"), fn, lineNum);
 	return RPMERR_RPMRC;
     }
@@ -334,21 +313,22 @@ static int addCanon(struct canonEntry **table, int *tableLen, char *line,
 	return RPMERR_RPMRC;
     }
 
-    t->num = strtoul(s, &s1, 10);
-    if ((*s1) || (s1 == s) || (t->num == ULONG_MAX)) {
+    tnum = strtoul(s, &s1, 10);
+    if ((*s1) || (s1 == s) || (tnum == ULONG_MAX)) {
 	rpmError(RPMERR_RPMRC, _("Bad arch/os number: %s (%s:%d)"), s,
 	      fn, lineNum);
 	return(RPMERR_RPMRC);
     }
 
-    t->name = xstrdup(t->name);
-    t->short_name = xstrdup(t->short_name);
+    t[0].name = xstrdup(tname);
+    t[0].short_name = xstrdup(tshort_name);
+    t[0].num = tnum;
 
     /* From A B C entry */
     /* Add  B B C entry */
-    t[1].name = xstrdup(t->short_name);
-    t[1].short_name = xstrdup(t->short_name);
-    t[1].num = t->num;
+    t[1].name = xstrdup(tshort_name);
+    t[1].short_name = xstrdup(tshort_name);
+    t[1].num = tnum;
 
     return 0;
 }
@@ -531,11 +511,10 @@ int rpmReadRC(const char * rcfiles)
 {
     char *myrcfiles, *r, *re;
     int rc;
-    static int first = 1;
 
-    if (first) {
+    if (!defaultsInitialized) {
 	setDefaults();
-	first = 0;
+	defaultsInitialized = 1;
     }
 
     if (rcfiles == NULL)
@@ -548,7 +527,11 @@ int rpmReadRC(const char * rcfiles)
 	FD_t fd;
 
 	/* Get pointer to rest of files */
-	if ((re = strchr(r, ':')) != NULL)
+	for (re = r; (re = strchr(re, ':')) != NULL; re++) {
+	    if (!(re[1] == '/' && re[2] == '/'))
+		break;
+	}
+	if (re && *re == ':')
 	    *re++ = '\0';
 	else
 	    re = r + strlen(r);
@@ -558,6 +541,9 @@ int rpmReadRC(const char * rcfiles)
 	if (r[0] == '~' && r[1] == '/') {
 	    char *home = getenv("HOME");
 	    if (home == NULL) {
+	    /* XXX Only /usr/lib/rpm/rpmrc must exist in default rcfiles list */
+		if (rcfiles == defrcfiles && myrcfiles != r)
+		    continue;
 		rpmError(RPMERR_RPMRC, _("Cannot expand %s"), r);
 		rc = 1;
 		break;
@@ -568,18 +554,18 @@ int rpmReadRC(const char * rcfiles)
 	strcat(fn, r);
 
 	/* Read another rcfile */
-	fd = fdOpen(fn, O_RDONLY, 0);
-	if (fdFileno(fd) < 0) {
+	fd = Fopen(fn, "r.fpio");
+	if (fd == NULL || Ferror(fd)) {
 	    /* XXX Only /usr/lib/rpm/rpmrc must exist in default rcfiles list */
 	    if (rcfiles == defrcfiles && myrcfiles != r)
 		continue;
 	    rpmError(RPMERR_RPMRC, _("Unable to open %s for reading: %s."),
-		 fn, strerror(errno));
+		 fn, Fstrerror(fd));
 	    rc = 1;
 	    break;
+	} else {
+	    rc = doReadRC(fd, fn);
 	}
-	rc = doReadRC(fd, fn);
-	fdClose(fd);
  	if (rc) break;
     }
     if (myrcfiles)	free(myrcfiles);
@@ -588,45 +574,65 @@ int rpmReadRC(const char * rcfiles)
 
     rpmSetMachine(NULL, NULL);	/* XXX WTFO? Why bother? */
 
-    if ((r = rpmGetVar(RPMVAR_MACROFILES)) != NULL)
-	initMacros(NULL, r);
+    {	const char *macrofiles;
+	if ((macrofiles = rpmGetVar(RPMVAR_MACROFILES)) != NULL) {
+	    macrofiles = strdup(macrofiles);
+	    rpmInitMacros(NULL, macrofiles);
+	    xfree(macrofiles);
+	}
+    }
 
     return rc;
 }
 
-static int doReadRC(FD_t fd, const char * filename)
+static int doReadRC( /*@killref@*/ FD_t fd, const char * urlfn)
 {
-    char *s, *se, *next;
+    const char *s;
+    char *se, *next;
     int linenum = 0;
     struct rpmOption searchOption, * option;
     int rc;
 
-  { struct stat sb;
-    fstat(fdFileno(fd), &sb);
-    next = alloca(sb.st_size + 2);
-    if (fdRead(fd, next, sb.st_size) != sb.st_size) {
-	rpmError(RPMERR_RPMRC, _("Failed to read %s: %s."), filename,
-		 strerror(errno));
-	return 1;
+    /* XXX really need rc = Slurp(fd, const char * filename, char ** buf) */
+  { off_t size = fdSize(fd);
+    size_t nb = (size >= 0 ? size : (8*BUFSIZ - 2));
+    if (nb == 0) {
+	Fclose(fd);
+	return 0;
     }
-    next[sb.st_size] = '\n';
-    next[sb.st_size + 1] = '\0';
+    next = alloca(nb + 2);
+    next[0] = '\0';
+    rc = Fread(next, sizeof(*next), nb, fd);
+    if (Ferror(fd) || (size > 0 && rc != nb)) {	/* XXX Feof(fd) */
+	rpmError(RPMERR_RPMRC, _("Failed to read %s: %s."), urlfn,
+		 Fstrerror(fd));
+	rc = 1;
+    } else
+	rc = 0;
+    Fclose(fd);
+    if (rc) return rc;
+    next[nb] = '\n';
+    next[nb + 1] = '\0';
   }
 
     while (*next) {
 	linenum++;
 
-	se = s = next;
+	s = se = next;
+
+	/* Find end-of-line. */
 	while (*se && *se != '\n') se++;
 	if (*se) *se++ = '\0';
 	next = se;
 
+	/* Trim leading spaces */
 	while (*s && isspace(*s)) s++;
 
-	/* we used to allow comments to begin anywhere, but not anymore */
+	/* We used to allow comments to begin anywhere, but not anymore. */
 	if (*s == '#' || *s == '\0') continue;
 
-	se = s;
+	/* Find end-of-keyword. */
+	se = (char *)s;
 	while (*se && !isspace(*se) && *se != ':') se++;
 
 	if (isspace(*se)) {
@@ -635,11 +641,11 @@ static int doReadRC(FD_t fd, const char * filename)
 	}
 
 	if (*se != ':') {
-	    rpmError(RPMERR_RPMRC, _("missing ':' at %s:%d"),
-		     filename, linenum);
+	    rpmError(RPMERR_RPMRC, _("missing ':' (found 0x%02x) at %s:%d"),
+		     (0xff & *se), urlfn, linenum);
 	    return 1;
 	}
-	*se++ = '\0';
+	*se++ = '\0';	/* terminate keyword or option, point to value */
 	while (*se && isspace(*se)) se++;
 
 	/* Find keyword in table */
@@ -653,7 +659,7 @@ static int doReadRC(FD_t fd, const char * filename)
 	    arch = val = fn = NULL;
 	    if (*se == '\0') {
 		rpmError(RPMERR_RPMRC, _("missing argument for %s at %s:%d"),
-		      option->name, filename, linenum);
+		      option->name, urlfn, linenum);
 		return 1;
 	    }
 
@@ -670,29 +676,29 @@ static int doReadRC(FD_t fd, const char * filename)
 		fn = rpmGetPath(s, NULL);
 		if (fn == NULL || *fn == '\0') {
 		    rpmError(RPMERR_RPMRC, _("%s expansion failed at %s:%d \"%s\""),
-			option->name, filename, linenum, s);
+			option->name, urlfn, linenum, s);
 		    if (fn) xfree(fn);
 		    return 1;
+		    /*@notreached@*/
 		}
 
-		fdinc = fdOpen(fn, O_RDONLY, 0);
-		if (fdFileno(fdinc) < 0) {
-		    rpmError(RPMERR_RPMRC, _("cannot open %s at %s:%d"),
-			fn, filename, linenum);
+		fdinc = Fopen(fn, "r.fpio");
+		if (fdinc == NULL || Ferror(fdinc)) {
+		    rpmError(RPMERR_RPMRC, _("cannot open %s at %s:%d: %s"),
+			fn, urlfn, linenum, Fstrerror(fdinc));
 		    rc = 1;
 		} else {
 		    rc = doReadRC(fdinc, fn);
-		    fdClose(fdinc);
 		}
 		if (fn) xfree(fn);
 		if (rc) return rc;
 		continue;	/* XXX don't save include value as var/macro */
-	      }	break;
+	      }	/*@notreached@*/ break;
 	    case RPMVAR_MACROFILES:
 		fn = rpmGetPath(se, NULL);
 		if (fn == NULL || *fn == '\0') {
 		    rpmError(RPMERR_RPMRC, _("%s expansion failed at %s:%d \"%s\""),
-			option->name, filename, linenum, fn);
+			option->name, urlfn, linenum, fn);
 		    if (fn) xfree(fn);
 		    return 1;
 		}
@@ -719,7 +725,7 @@ static int doReadRC(FD_t fd, const char * filename)
 		if (*se == '\0') {
 		    rpmError(RPMERR_RPMRC,
 				_("missing architecture for %s at %s:%d"),
-			  	option->name, filename, linenum);
+			  	option->name, urlfn, linenum);
 		    return 1;
 		}
 		*se++ = '\0';
@@ -727,7 +733,7 @@ static int doReadRC(FD_t fd, const char * filename)
 		if (*se == '\0') {
 		    rpmError(RPMERR_RPMRC,
 				_("missing argument for %s at %s:%d"),
-			  	option->name, filename, linenum);
+			  	option->name, urlfn, linenum);
 		    return 1;
 		}
 	    }
@@ -760,11 +766,11 @@ static int doReadRC(FD_t fd, const char * filename)
 	    }
 
 	    if (i < RPM_MACHTABLE_COUNT) {
-		char *rest = s + strlen(tables[i].key);
+		const char *rest = s + strlen(tables[i].key);
 		if (*rest == '_') rest++;
 
 		if (!strcmp(rest, "compat")) {
-		    if (machCompatCacheAdd(se, filename, linenum,
+		    if (machCompatCacheAdd(se, urlfn, linenum,
 						&tables[i].cache))
 			return 1;
 		    gotit = 1;
@@ -772,13 +778,13 @@ static int doReadRC(FD_t fd, const char * filename)
 			   !strcmp(rest, "translate")) {
 		    if (addDefault(&tables[i].defaults,
 				   &tables[i].defaultsLength,
-				   se, filename, linenum))
+				   se, urlfn, linenum))
 			return 1;
 		    gotit = 1;
 		} else if (tables[i].hasCanon &&
 			   !strcmp(rest, "canon")) {
 		    if (addCanon(&tables[i].canons, &tables[i].canonsLength,
-				 se, filename, linenum))
+				 se, urlfn, linenum))
 			return 1;
 		    gotit = 1;
 		}
@@ -786,7 +792,7 @@ static int doReadRC(FD_t fd, const char * filename)
 
 	    if (!gotit) {
 		rpmError(RPMERR_RPMRC, _("bad option '%s' at %s:%d"),
-			    s, filename, linenum);
+			    s, urlfn, linenum);
 	    }
 	}
     }
@@ -794,7 +800,7 @@ static int doReadRC(FD_t fd, const char * filename)
     return 0;
 }
 
-static void defaultMachine(char ** arch, char ** os) {
+static void defaultMachine(const char ** arch, const char ** os) {
     static struct utsname un;
     static int gotDefaults = 0;
     char * chptr;
@@ -805,69 +811,74 @@ static void defaultMachine(char ** arch, char ** os) {
 
 #if !defined(__linux__)
 #ifdef SNI
-        /* USUALLY un.sysname on sinix does start with the word "SINIX"
-         * let's be absolutely sure
-         */
-        sprintf(un.sysname,"SINIX");
+	/* USUALLY un.sysname on sinix does start with the word "SINIX"
+	 * let's be absolutely sure
+	 */
+	sprintf(un.sysname,"SINIX");
 #endif
 	if (!strcmp(un.sysname, "AIX")) {
 	    strcpy(un.machine, __power_pc() ? "ppc" : "rs6000");
-            sprintf(un.sysname,"aix%s.%s",un.version,un.release);
+	    sprintf(un.sysname,"aix%s.%s",un.version,un.release);
 	}
-        else if (!strcmp(un.sysname, "SunOS")) {
-           if (!strncmp(un.release,"4", 1)) /* SunOS 4.x */ {
-	      int fd;
-              for (fd=0;(un.release[fd] != 0 && (fd < sizeof(un.release)));fd++)
-                 if (!isdigit(un.release[fd]) && (un.release[fd] != '.')) {
-                    un.release[fd] = 0;
-                    break;
-                 }
-              sprintf(un.sysname,"sunos%s",un.release);
-           }
+	else if (!strcmp(un.sysname, "SunOS")) {
+	    if (!strncmp(un.release,"4", 1)) /* SunOS 4.x */ {
+		int fd;
+		for (fd = 0;
+		    (un.release[fd] != 0 && (fd < sizeof(un.release)));
+		    fd++) {
+		      if (!isdigit(un.release[fd]) && (un.release[fd] != '.')) {
+			un.release[fd] = 0;
+			break;
+		      }
+		    }
+		    sprintf(un.sysname,"sunos%s",un.release);
+	    }
 
-           else /* Solaris 2.x: n.x.x becomes n-3.x.x */
-              sprintf(un.sysname,"solaris%1d%s",atoi(un.release)-3,un.release+1+(atoi(un.release)/10));
-        }
-        else if (!strcmp(un.sysname, "HP-UX"))
-           /*make un.sysname look like hpux9.05 for example*/
-           sprintf(un.sysname,"hpux%s",strpbrk(un.release,"123456789"));
-        else if (!strcmp(un.sysname, "OSF1"))
-           /*make un.sysname look like osf3.2 for example*/
-           sprintf(un.sysname,"osf%s",strpbrk(un.release,"123456789"));
-        else if (!strncmp(un.sysname, "IP", 2))
-           un.sysname[2] = '\0';
-        else if (!strncmp(un.sysname, "SINIX", 5)) {
-           sprintf(un.sysname, "sinix%s",un.release);
-           if (!strncmp(un.machine, "RM", 2))
-              sprintf(un.machine, "mips");
-        }
-        else if ((!strncmp(un.machine, "34", 2) || \
-                 !strncmp(un.machine, "33", 2)) && \
-                 !strncmp(un.release, "4.0", 3)) {
-           /* we are on ncr-sysv4 */
-	   char *prelid = NULL;
-           FD_t fd = fdOpen("/etc/.relid", O_RDONLY, 0700);
-           if (fdFileno(fd) > 0) {
-              chptr = (char *) xcalloc(1, 256);
-              if (chptr != NULL) {
-                 int irelid = fdRead(fd, (void *)chptr, 256);
-                 fdClose(fd);
-                 /* example: "112393 RELEASE 020200 Version 01 OS" */
-                 if (irelid > 0) {
-                    if ((prelid=strstr(chptr, "RELEASE "))){
-                       prelid += strlen("RELEASE ")+1;
-                       sprintf(un.sysname,"ncr-sysv4.%.*s",1,prelid);
-                    }
-		 }
-                 free (chptr);
-              }
-           }
-           if (!prelid)
-              /* parsing /etc/.relid file failed */
-              strcpy(un.sysname,"ncr-sysv4");
-           /* wrong, just for now, find out how to look for i586 later*/
-           strcpy(un.machine,"i486");
-        }
+	    else /* Solaris 2.x: n.x.x becomes n-3.x.x */
+		sprintf(un.sysname, "solaris%1d%s", atoi(un.release)-3,
+			un.release+1+(atoi(un.release)/10));
+	}
+	else if (!strcmp(un.sysname, "HP-UX"))
+	    /*make un.sysname look like hpux9.05 for example*/
+	    sprintf(un.sysname, "hpux%s", strpbrk(un.release,"123456789"));
+	else if (!strcmp(un.sysname, "OSF1"))
+	    /*make un.sysname look like osf3.2 for example*/
+	    sprintf(un.sysname,"osf%s",strpbrk(un.release,"123456789"));
+	else if (!strncmp(un.sysname, "IP", 2))
+	    un.sysname[2] = '\0';
+	else if (!strncmp(un.sysname, "SINIX", 5)) {
+	    sprintf(un.sysname, "sinix%s",un.release);
+	    if (!strncmp(un.machine, "RM", 2))
+		sprintf(un.machine, "mips");
+	}
+	else if ((!strncmp(un.machine, "34", 2) ||
+		!strncmp(un.machine, "33", 2)) && \
+		!strncmp(un.release, "4.0", 3))
+	{
+	    /* we are on ncr-sysv4 */
+	    char *prelid = NULL;
+	    FD_t fd;
+	    fd = Fopen("/etc/.relid", "r.fdio");
+	    if (!Ferror(fd)) {
+		chptr = (char *) xcalloc(1, 256);
+		if (chptr != NULL) {
+		    int irelid = Fread(chptr, sizeof(*chptr), 256, fd);
+		    Fclose(fd);
+		    /* example: "112393 RELEASE 020200 Version 01 OS" */
+		    if (irelid > 0) {
+			if ((prelid=strstr(chptr, "RELEASE "))){
+			    prelid += strlen("RELEASE ")+1;
+			    sprintf(un.sysname,"ncr-sysv4.%.*s",1,prelid);
+			}
+		    }
+		    free (chptr);
+		}
+	    }
+	    if (prelid == NULL)	/* parsing /etc/.relid file failed? */
+		strcpy(un.sysname,"ncr-sysv4");
+	    /* wrong, just for now, find out how to look for i586 later*/
+	    strcpy(un.machine,"i486");
+	}
 #endif	/* __linux__ */
 
 	/* get rid of the hyphens in the sysname */
@@ -924,7 +935,7 @@ static void defaultMachine(char ** arch, char ** os) {
 	}
 #	endif	/* hpux */
 
-#	if defined(__linux__) && defined(__sparc__)
+#	if HAVE_PERSONALITY && defined(__linux__) && defined(__sparc__)
 	if (!strcmp(un.machine, "sparc")) {
 	    #define PERS_LINUX		0x00000000
 	    #define PERS_LINUX_32BIT	0x00800000
@@ -966,7 +977,7 @@ static void defaultMachine(char ** arch, char ** os) {
     if (os) *os = un.sysname;
 }
 
-static char * rpmGetVarArch(int var, char * arch) {
+static const char * rpmGetVarArch(int var, char * arch) {
     struct rpmvarValue * next;
 
     if (!arch) arch = current[ARCH];
@@ -985,7 +996,7 @@ static char * rpmGetVarArch(int var, char * arch) {
     return next ? next->value : NULL;
 }
 
-char *rpmGetVar(int var)
+const char *rpmGetVar(int var)
 {
     return rpmGetVarArch(var, NULL);
 }
@@ -997,11 +1008,11 @@ static void freeRpmVar(struct rpmvarValue * orig) {
     while (var) {
 	next = var->next;
 	if (var->arch) {
-	    free(var->arch);
+	    xfree(var->arch);
 	    var->arch = NULL;
 	}
 	if (var->value) {
-	    free(var->value);
+	    xfree(var->value);
 	    var->value = NULL;
 	}
 
@@ -1032,21 +1043,23 @@ static void rpmSetVarArch(int var, const char * val, const char * arch) {
 	}
 
 	if (next->arch && arch && !strcmp(next->arch, arch)) {
-	    if (next->value) free(next->value);
-	    if (next->arch) free(next->arch);
+	    if (next->value) xfree(next->value);
+	    if (next->arch) xfree(next->arch);
 	} else if (next->arch || arch) {
 	    next->next = xmalloc(sizeof(*next->next));
 	    next = next->next;
+	    next->value = NULL;
+	    next->arch = NULL;
 	    next->next = NULL;
 	}
     }
 
-    next->value = xstrdup(val);
+    next->value = xstrdup(val);		/* XXX memory leak, hard to plug */
     next->arch = (arch ? xstrdup(arch) : NULL);
 }
 
 void rpmSetTables(int archTable, int osTable) {
-    char * arch, * os;
+    const char * arch, * os;
 
     defaultMachine(&arch, &os);
 
@@ -1061,12 +1074,12 @@ void rpmSetTables(int archTable, int osTable) {
     }
 }
 
-int rpmMachineScore(int type, char * name) {
+int rpmMachineScore(int type, const char * name) {
     struct machEquivInfo * info = machEquivSearch(&tables[type].equiv, name);
     return (info != NULL ? info->score : 0);
 }
 
-void rpmGetMachine(char **arch, char **os)
+void rpmGetMachine(const char **arch, const char **os)
 {
     if (arch)
 	*arch = current[ARCH];
@@ -1076,7 +1089,7 @@ void rpmGetMachine(char **arch, char **os)
 }
 
 void rpmSetMachine(const char * arch, const char * os) {
-    char * host_cpu, * host_os;
+    const char * host_cpu, * host_os;
 
     defaultMachine(&host_cpu, &host_os);
 
@@ -1119,13 +1132,15 @@ void rpmSetMachine(const char * arch, const char * os) {
     }
 }
 
-static void rebuildCompatTables(int type, char * name) {
+static void rebuildCompatTables(int type, const char * name) {
     machFindEquivs(&tables[currTables[type]].cache,
 		   &tables[currTables[type]].equiv,
 		   name);
 }
 
-static void getMachineInfo(int type, /*@out@*/char ** name, /*@out@*/int * num) {
+static void getMachineInfo(int type, /*@only@*/ /*@out@*/ const char ** name,
+			/*@out@*/int * num)
+{
     struct canonEntry * canon;
     int which = currTables[type];
 
@@ -1150,11 +1165,11 @@ static void getMachineInfo(int type, /*@out@*/char ** name, /*@out@*/int * num)
     }
 }
 
-void rpmGetArchInfo(char ** name, int * num) {
+void rpmGetArchInfo(const char ** name, int * num) {
     getMachineInfo(ARCH, name, num);
 }
 
-void rpmGetOsInfo(char ** name, int * num) {
+void rpmGetOsInfo(const char ** name, int * num) {
     getMachineInfo(OS, name, num);
 }
 
@@ -1190,24 +1205,28 @@ void rpmRebuildTargetVars(const char **buildtarget, const char ** canontarget)
 	    if (co != NULL) co = xstrdup(co);
 	}
     } else {
+	const char *a = NULL;
+	const char *o = NULL;
 	/* Set build target from rpm arch and os */
-	rpmGetArchInfo(&ca,NULL);
-	if (ca)	ca = xstrdup(ca);
-	rpmGetOsInfo(&co,NULL);
-	if (co)	co = xstrdup(co);
+	rpmGetArchInfo(&a, NULL);
+	ca = (a) ? xstrdup(a) : NULL;
+	rpmGetOsInfo(&o, NULL);
+	co = (o) ? xstrdup(o) : NULL;
     }
 
     /* If still not set, Set target arch/os from default uname(2) values */
     if (ca == NULL) {
-	defaultMachine(&ca, NULL);
-	ca = xstrdup(ca);
-     }
+	const char *a = NULL;
+	defaultMachine(&a, NULL);
+	ca = (a) ? xstrdup(a) : NULL;
+    }
     for (x = 0; ca[x]; x++)
 	ca[x] = tolower(ca[x]);
 
     if (co == NULL) {
-	defaultMachine(NULL, &co);
-	co = xstrdup(co);
+	const char *o = NULL;
+	defaultMachine(NULL, &o);
+	co = (o) ? xstrdup(o) : NULL;
     }
     for (x = 0; co[x]; x++)
 	co[x] = tolower(co[x]);
@@ -1220,7 +1239,7 @@ void rpmRebuildTargetVars(const char **buildtarget, const char ** canontarget)
 
 /*
  * XXX All this macro pokery/jiggery could be achieved by doing a delayed
- *	initMacros(NULL, PER-PLATFORM-MACRO-FILE-NAMES);
+ *	rpmInitMacros(NULL, PER-PLATFORM-MACRO-FILE-NAMES);
  */
     delMacro(NULL, "_target");
     addMacro(NULL, "_target", NULL, ct, RMIL_RPMRC);
@@ -1231,7 +1250,7 @@ void rpmRebuildTargetVars(const char **buildtarget, const char ** canontarget)
 /*
  * XXX Make sure that per-arch optflags is initialized correctly.
  */
-  { char *optflags = rpmGetVarArch(RPMVAR_OPTFLAGS, ca);
+  { const char *optflags = rpmGetVarArch(RPMVAR_OPTFLAGS, ca);
     if (optflags != NULL) {
 	delMacro(NULL, "optflags");
 	addMacro(NULL, "optflags", NULL, optflags, RMIL_RPMRC);
@@ -1258,6 +1277,8 @@ void rpmFreeRpmrc(void)
 		if (t->equiv.list[j].name)	xfree(t->equiv.list[j].name);
 	    }
 	    xfree(t->equiv.list);
+	    t->equiv.list = NULL;
+	    t->equiv.count = 0;
 	}
 	if (t->cache.cache) {
 	    for (j = 0; j < t->cache.size; j++) {
@@ -1267,12 +1288,14 @@ void rpmFreeRpmrc(void)
 		if (e->name)		xfree(e->name);
 		if (e->equivs) {
 		    for (k = 0; k < e->count; k++) {
-		    if (e->equivs[k])	xfree(e->equivs[k]);
+			if (e->equivs[k])	xfree(e->equivs[k]);
 		    }
 		    xfree(e->equivs);
 		}
 	    }
 	    xfree(t->cache.cache);
+	    t->cache.cache = NULL;
+	    t->cache.size = 0;
 	}
 	if (t->defaults) {
 	    for (j = 0; j < t->defaultsLength; j++) {
@@ -1280,6 +1303,8 @@ void rpmFreeRpmrc(void)
 		if (t->defaults[j].defName)	xfree(t->defaults[j].defName);
 	    }
 	    xfree(t->defaults);
+	    t->defaults = NULL;
+	    t->defaultsLength = 0;
 	}
 	if (t->canons) {
 	    for (j = 0; j < t->canonsLength; j++) {
@@ -1287,6 +1312,8 @@ void rpmFreeRpmrc(void)
 		if (t->canons[j].short_name)	xfree(t->canons[j].short_name);
 	    }
 	    xfree(t->canons);
+	    t->canons = NULL;
+	    t->canonsLength = 0;
 	}
     }
 
@@ -1298,11 +1325,20 @@ void rpmFreeRpmrc(void)
 	    if (this->arch)	xfree(this->arch);
 	    xfree(this);
 	}
-	if (values[i].value)	xfree(values[i].value);
-	if (values[i].arch)	xfree(values[i].arch);
+	if (values[i].value)
+	    xfree(values[i].value);
+	values[i].value = NULL;
+	if (values[i].arch)
+	    xfree(values[i].arch);
+	values[i].arch = NULL;
     }
-    if (current[OS])	xfree(current[OS]);
-    if (current[ARCH])	xfree(current[ARCH]);
+    if (current[OS])
+	xfree(current[OS]);
+    current[OS] = NULL;
+    if (current[ARCH])
+	xfree(current[ARCH]);
+    current[ARCH] = NULL;
+    defaultsInitialized = 0;
     return;
 }
 
@@ -1350,12 +1386,12 @@ int rpmShowRC(FILE *f)
 
     fprintf(f, "\nRPMRC VALUES:\n");
     for (i = 0, opt = optionTable; i < optionTableSize; i++, opt++) {
-	char *s = rpmGetVar(opt->var);
+	const char *s = rpmGetVar(opt->var);
 	if (s != NULL || rpmGetVerbosity() < RPMMESS_NORMAL)
 	    fprintf(f, "%-21s : %s\n", opt->name, s ? s : "(not set)");
     }
 
-    dumpMacroTable(NULL, f);
+    rpmDumpMacroTable(NULL, f);
 
     return 0;
 }
