@@ -74,7 +74,8 @@ const char * rpmDetectPGPVersion(pgpVersion *pgpVer)
 	char *pgpvbin;
 	struct stat statbuf;
 	
-	if (!pgpbin || ! (pgpvbin = (char *)alloca(strlen(pgpbin) + 2))) {
+	if (!(pgpbin && pgpbin[0] != '%') || ! (pgpvbin = (char *)alloca(strlen(pgpbin) + 2))) {
+	  if (pgpbin) xfree(pgpbin);
 	  saved_pgp_version = -1;
 	  return NULL;
 	}
@@ -98,7 +99,7 @@ static int checkSize(FD_t fd, int size, int sigsize)
     int headerArchiveSize;
     struct stat statbuf;
 
-    fstat(fdFileno(fd), &statbuf);
+    fstat(Fileno(fd), &statbuf);
 
     if (S_ISREG(statbuf.st_mode)) {
 	headerArchiveSize = statbuf.st_size - sizeof(struct rpmlead) - sigsize;
@@ -202,7 +203,7 @@ int rpmWriteSignature(FD_t fd, Header header)
 	rpmMessage(RPMMESS_DEBUG, _("Signature size: %d\n"), sigSize);
 	rpmMessage(RPMMESS_DEBUG, _("Signature pad : %d\n"), pad);
 	memset(buf, 0, pad);
-	if (fdWrite(fd, buf, pad) != pad)
+	if (Fwrite(buf, sizeof(buf[0]), pad, fd) != pad)
 	    rc = 1;
     }
     return rc;
@@ -268,8 +269,8 @@ static int makePGPSignature(const char *file, /*@out@*/void **sig, /*@out@*/int_
     }
 
     close(inpipe[0]);
-    write(inpipe[1], passPhrase, strlen(passPhrase));
-    write(inpipe[1], "\n", 1);
+    (void)write(inpipe[1], passPhrase, strlen(passPhrase));
+    (void)write(inpipe[1], "\n", 1);
     close(inpipe[1]);
 
     (void)waitpid(pid, &status, 0);
@@ -291,10 +292,10 @@ static int makePGPSignature(const char *file, /*@out@*/void **sig, /*@out@*/int_
     
     {	FD_t fd;
 	int rc;
-	fd = fdOpen(sigfile, O_RDONLY, 0);
+	fd = Fopen(sigfile, "r.fdio");
 	rc = timedRead(fd, *sig, *size);
 	unlink(sigfile);
-	fdClose(fd);
+	Fclose(fd);
 	if (rc != *size) {
 	    free(*sig);
 	    rpmError(RPMERR_SIGGEN, _("unable to read the signature"));
@@ -368,10 +369,10 @@ static int makeGPGSignature(const char *file, /*@out@*/void **sig, /*@out@*/int_
     
     {	FD_t fd;
 	int rc;
-	fd = fdOpen(sigfile, O_RDONLY, 0);
+	fd = Fopen(sigfile, "r.fdio");
 	rc = timedRead(fd, *sig, *size);
 	unlink(sigfile);
-	fdClose(fd);
+	Fclose(fd);
 	if (rc != *size) {
 	    free(*sig);
 	    rpmError(RPMERR_SIGGEN, _("unable to read the signature"));
@@ -479,7 +480,7 @@ static int verifyPGPSignature(const char *datafile, void *sig,
     int pid, status, outpipe[2];
     FD_t sfd;
     char *sigfile;
-    unsigned char buf[8192];
+    unsigned char buf[BUFSIZ];
     FILE *file;
     int res = RPMSIG_OK;
     const char *path;
@@ -505,9 +506,9 @@ static int verifyPGPSignature(const char *datafile, void *sig,
     sigfile = tempnam(tmppath, "rpmsig");
     xfree(tmppath);
   }
-    sfd = fdOpen(sigfile, O_WRONLY|O_CREAT|O_TRUNC, 0644);
-    (void)fdWrite(sfd, sig, count);
-    fdClose(sfd);
+    sfd = Fopen(sigfile, "w.fdio");
+    (void)Fwrite(sig, sizeof(char), count, sfd);
+    Fclose(sfd);
 
     /* Now run PGP */
     outpipe[0] = outpipe[1] = 0;
@@ -600,9 +601,9 @@ static int verifyGPGSignature(const char *datafile, void *sig,
     sigfile = tempnam(tmppath, "rpmsig");
     xfree(tmppath);
   }
-    sfd = fdOpen(sigfile, O_WRONLY|O_CREAT|O_TRUNC, 0644);
-    (void)fdWrite(sfd, sig, count);
-    fdClose(sfd);
+    sfd = Fopen(sigfile, "w.fdio");
+    (void)Fwrite(sig, sizeof(char), count, sfd);
+    Fclose(sfd);
 
     /* Now run GPG */
     outpipe[0] = outpipe[1] = 0;
@@ -723,8 +724,8 @@ static int checkPassPhrase(const char *passPhrase, const int sigTag)
     }
 
     close(passPhrasePipe[0]);
-    write(passPhrasePipe[1], passPhrase, strlen(passPhrase));
-    write(passPhrasePipe[1], "\n", 1);
+    (void)write(passPhrasePipe[1], passPhrase, strlen(passPhrase));
+    (void)write(passPhrasePipe[1], "\n", 1);
     close(passPhrasePipe[1]);
 
     (void)waitpid(pid, &status, 0);
