@@ -3,7 +3,11 @@
 
 /* signature.h - generate and verify signatures */
 
-#include "header.h"
+#include <header.h>
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /**************************************************/
 /*                                                */
@@ -34,7 +38,7 @@
 Header rpmNewSignature(void);
 
 /* If an old-style signature is found, we emulate a new style one */
-int rpmReadSignature(FD_t fd, Header *header, short sig_type);
+int rpmReadSignature(FD_t fd, /*@out@*/ Header *header, short sig_type);
 int rpmWriteSignature(FD_t fd, Header header);
 
 /* Generate a signature of data in file, insert in header */
@@ -54,7 +58,16 @@ int rpmLookupSignatureType(int action);
 /* Utility to read a pass phrase from the user */
 char *rpmGetPassPhrase(const char *prompt, const int sigTag);
 
+/* >0 is a valid PGP version */
+typedef enum pgpVersion_e {
+	PGP_NOTDETECTED = -1, PGP_UNKNOWN = 0, PGP_2 = 2, PGP_5 = 5
+} pgpVersion;
+
 /* Return path to pgp executable of given type, or NULL when not found */
-const char *rpmDetectPGPVersion(int sigType);
+const char *rpmDetectPGPVersion( /*@out@*/ pgpVersion *pgpVersion);
+
+#ifdef __cplusplus
+}
+#endif
 
 #endif	/* H_SIGNATURE */
