@@ -1,7 +1,7 @@
 #include "system.h"
 
 #include <rpmlib.h>
-#include <rpmmacro.h>	/* XXX for rpmGetPath */
+#include <rpmmacro.h>	/* XXX for rpmExpand */
 
 #include "depends.h"
 #include "fprint.h"
@@ -41,7 +41,9 @@ typedef struct transactionFileInfo {
     enum fileActions * actions;
     fingerPrint * fps;
     uint_32 * fflags, * fsizes;
-    const char ** fl;
+    const char ** bnl;	    /* base names */
+    const char ** dnl;	    /* dir names */
+    const int * dil;	    /* dir index list */
     char ** fmd5s;
     uint_16 * fmodes;
     Header h;
@@ -89,8 +91,12 @@ static void freeFi(TFI_t *fi)
 	if (fi->replaced) {
 	    free(fi->replaced); fi->replaced = NULL;
 	}
-	if (fi->fl) {
-	    free(fi->fl); fi->fl = NULL;
+	if (fi->bnl) {
+	    free(fi->bnl); fi->bnl = NULL;
+	    free(fi->dnl); fi->dnl = NULL;
+#ifdef	DOUBLE_FREE
+	    xfree(fi->dil); fi->dil = NULL;
+#endif
 	}
 	if (fi->flinks) {
 	    free(fi->flinks); fi->flinks = NULL;
@@ -131,7 +137,7 @@ static void freeFl(rpmTransactionSet ts, TFI_t *flList)
 
 void rpmtransSetScriptFd(rpmTransactionSet ts, FD_t fd)
 {
-    ts->scriptFd = fd;
+    ts->scriptFd = (fd ? fdLink(fd, "rpmtransSetScriptFd") : NULL);
 }
 
 static rpmProblemSet psCreate(void)
@@ -146,8 +152,8 @@ static rpmProblemSet psCreate(void)
 }
 
 static void psAppend(rpmProblemSet probs, rpmProblemType type,
-		     const void * key, Header h, const char * str1,
-		     Header altH, unsigned long ulong1)
+		/*@dependent@*/ const void * key, Header h, const char * str1,
+		Header altH, unsigned long ulong1)
 {
     if (probs->numProblems == probs->numProblemsAlloced) {
 	if (probs->numProblemsAlloced)
@@ -175,6 +181,17 @@ static void psAppend(rpmProblemSet probs, rpmProblemType type,
     probs->probs[probs->numProblems++].ignoreProblem = 0;
 }
 
+static void psAppendFile(rpmProblemSet probs, rpmProblemType type,
+		/*@dependent@*/ const void * key, Header h,
+		const char * dirName, const char * baseName,
+		Header altH, unsigned long ulong1)
+{
+    char * str = alloca(strlen(dirName) + strlen(baseName) + 1);
+
+    sprintf(str, "%s%s", dirName, baseName);
+    psAppend(probs, type, key, h, str, altH, ulong1);
+}
+
 static int archOkay(Header h)
 {
     int_8 * pkgArchNum;
@@ -227,7 +244,7 @@ void rpmProblemSetFree(rpmProblemSet probs)
 
     for (i = 0; i < probs->numProblems; i++) {
 	headerFree(probs->probs[i].h);
-	if (probs->probs[i].str1) free(probs->probs[i].str1);
+	if (probs->probs[i].str1) xfree(probs->probs[i].str1);
 	if (probs->probs[i].altH) {
 	    headerFree(probs->probs[i].altH);
 	}
@@ -245,10 +262,18 @@ static Header relocateFileList(struct availablePackage * alp,
     rpmRelocation * rawRelocations = alp->relocs;
     rpmRelocation * relocations = NULL;
     const char ** validRelocations;
-    const char ** names;
-    int_32 fileCount;
+    char ** baseNames, ** dirNames;
+    int_32 * dirIndexes;
+    int_32 * newDirIndexes;
+    int_32 fileCount, dirCount;
+    char * skipDirList;
     Header h;
-    int relocated = 0;
+    int relocated = 0, len;
+    char * str;
+    int fileAlloced = 0;
+    char * filespec = NULL;
+    char * chptr;
+    int haveRelocatedFile = 0;
 
     if (!headerGetEntry(origH, RPMTAG_PREFIXES, NULL,
 			(void **) &validRelocations, &numValid))
@@ -275,19 +300,19 @@ static Header relocateFileList(struct availablePackage * alp,
 	/* FIXME: default relocations (oldPath == NULL) need to be handled
 	   in the UI, not rpmlib */
 
-	{   const char *s = rawRelocations[i].oldPath;
-	    char *t = alloca(strlen(s) + 1);
-	    strcpy(t, s);
-	    stripTrailingSlashes(t);
-	    relocations[i].oldPath = t;
-	}
+	/* FIXME: Trailing /'s will confuse us greatly. Internal ones will 
+	   too, but those are more trouble to fix up. :-( */
+	str = alloca(strlen(rawRelocations[i].oldPath) + 1);
+	strcpy(str, rawRelocations[i].oldPath);
+	stripTrailingSlashes(str);
+	relocations[i].oldPath = str;
 
+	/* An old path w/o a new path is valid, and indicates exclusion */
 	if (rawRelocations[i].newPath) {
-	    const char *s = rawRelocations[i].newPath;
-	    char *t = alloca(strlen(s) + 1);
-	    strcpy(t, s);
-	    stripTrailingSlashes(t);
-	    relocations[i].newPath = t;
+	    str = alloca(strlen(rawRelocations[i].newPath) + 1);
+	    strcpy(str, rawRelocations[i].newPath);
+	    stripTrailingSlashes(str);
+	    relocations[i].newPath = str;
 
 	    /* Verify that the relocation's old path is in the header. */
 	    for (j = 0; j < numValid; j++)
@@ -319,55 +344,172 @@ static Header relocateFileList(struct availablePackage * alp,
     /* Add relocation values to the header */
     if (numValid) {
 	const char ** actualRelocations;
+	int numActual;
+
 	actualRelocations = xmalloc(sizeof(*actualRelocations) * numValid);
+	numActual = 0;
 	for (i = 0; i < numValid; i++) {
 	    for (j = 0; j < numRelocations; j++) {
 		if (strcmp(validRelocations[i], relocations[j].oldPath))
 		    continue;
-		actualRelocations[i] = relocations[j].newPath;
+		/* On install, a relocate to NULL means skip the path. */
+		if (relocations[j].newPath) {
+		    actualRelocations[numActual] = relocations[j].newPath;
+		    numActual++;
+		}
 		break;
 	    }
-	    if (j == numRelocations)
-		actualRelocations[i] = validRelocations[i];
+	    if (j == numRelocations) {
+		actualRelocations[numActual] = validRelocations[i];
+		numActual++;
+	    }
 	}
 
-	headerAddEntry(h, RPMTAG_INSTPREFIXES, RPM_STRING_ARRAY_TYPE,
-		       (void **) actualRelocations, numValid);
+	if (numActual)
+	    headerAddEntry(h, RPMTAG_INSTPREFIXES, RPM_STRING_ARRAY_TYPE,
+		       (void **) actualRelocations, numActual);
 
 	xfree(actualRelocations);
 	xfree(validRelocations);
     }
 
-    headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &names, &fileCount);
-
-    /* Relocate all file paths.
-     * Go through sorted file and relocation lists backwards so that /usr/local
-     * relocations take precedence over /usr ones.
+    /* For all relocations, we go through sorted file and relocation lists 
+     * backwards so that /usr/local relocations take precedence over /usr 
+     * ones.
      */
+
+    headerGetEntry(h, RPMTAG_BASENAMES, NULL, (void **) &baseNames, 
+		   &fileCount);
+    headerGetEntry(h, RPMTAG_DIRINDEXES, NULL, (void **) &dirIndexes, NULL);
+    headerGetEntry(h, RPMTAG_DIRNAMES, NULL, (void **) &dirNames, 
+		   &dirCount);
+    skipDirList = xcalloc(sizeof(*skipDirList), dirCount);
+
+    newDirIndexes = alloca(sizeof(*newDirIndexes) * fileCount);
+    memcpy(newDirIndexes, dirIndexes, sizeof(*newDirIndexes) * fileCount);
+    dirIndexes = newDirIndexes;
+
+    /* Now relocate individual files. */
+
     for (i = fileCount - 1; i >= 0; i--) {
+	/* If we're skipping the directory this file is part of, skip this
+	 * file as well.
+	 */
+	if (skipDirList[dirIndexes[i]]) {
+	    actions[i] = FA_SKIPNSTATE;
+	    rpmMessage(RPMMESS_DEBUG, _("excluding file %s%s\n"), 
+		       dirNames[dirIndexes[i]], baseNames[i]);
+	    continue;
+	} 
+
+	/* See if this file needs relocating (which will only occur if the
+	 * full file path we have exactly matches a path in the relocation
+	 * list. XXX FIXME: Would a bsearch of the (already sorted) 
+	 * relocation list be a good idea?
+	 */
+
+	len = strlen(dirNames[dirIndexes[i]]) + strlen(baseNames[i]) + 1;
+	if (len >= fileAlloced) {
+	    fileAlloced = len * 2;
+	    filespec = xrealloc(filespec, fileAlloced);
+	}
+	strcpy(filespec, dirNames[dirIndexes[i]]);
+	strcat(filespec, baseNames[i]);
+
+	for (j = numRelocations - 1; j >= 0; j--)
+	    if (!strcmp(relocations[j].oldPath, filespec)) break;
+
+	if (j < 0) continue;
+
+	if (actions && relocations[j].newPath == NULL) {
+	    /* On install, a relocate to NULL means skip the path. */
+	    skipDirList[i] = 1;
+	    rpmMessage(RPMMESS_DEBUG, _("excluding directory %s\n"), 
+		       dirNames[i]);
+	    continue;
+	}
+
+	rpmMessage(RPMMESS_DEBUG, _("relocating %s to %s\n"),
+		    filespec, relocations[j].newPath);
+	relocated = 1;
+
+	len = strlen(relocations[j].newPath);
+	if (len >= fileAlloced) {
+	    fileAlloced = len * 2;
+	    filespec = xrealloc(filespec, fileAlloced);
+	}
+	strcpy(filespec, relocations[j].newPath);
+	chptr = strrchr(filespec, '/');
+	*chptr++ = '\0';
+
+	/* filespec is the new path, and chptr is the new basename */
+	if (strcmp(baseNames[i], chptr)) {
+	    baseNames[i] = alloca(strlen(chptr) + 1);
+	    strcpy(baseNames[i], chptr);
+	}
+
+	/* Does this directory already exist in the directory list? */
+	for (j = 0; j < dirCount; j++)
+	    if (!strcmp(filespec, dirNames[j])) break;
+	
+	if (j < dirCount) {
+	    dirIndexes[i] = j;
+	    continue;
+	}
+
+	/* Creating new paths is a pita */
+	if (!haveRelocatedFile) {
+	    char ** newDirList;
+	    int k;
+
+	    haveRelocatedFile = 1;
+	    newDirList = xmalloc(sizeof(*newDirList) * (dirCount + 1));
+	    for (k = 0; k < dirCount; k++) {
+		newDirList[k] = alloca(strlen(dirNames[k]) + 1);
+		strcpy(newDirList[k], dirNames[k]);
+	    }
+	    free(dirNames);
+	    dirNames = newDirList;
+	} else {
+	    dirNames = xrealloc(dirNames, 
+			       sizeof(*dirNames) * (dirCount + 1));
+	}
+
+	dirNames[dirCount] = alloca(strlen(filespec) + 1);
+	strcpy(dirNames[dirCount], filespec);
+	dirIndexes[i] = dirCount;
+	dirCount++;
+    }
+
+    /* Start off by relocating directories. */
+    for (i = dirCount - 1; i >= 0; i--) {
 	for (j = numRelocations - 1; j >= 0; j--) {
-	    int len;
-	    len = strlen(relocations[j].oldPath);
-	    if (strncmp(relocations[j].oldPath, names[i], len))
+	    int oplen;
+
+	    oplen = strlen(relocations[j].oldPath);
+	    if (strncmp(relocations[j].oldPath, dirNames[i], oplen))
 		continue;
 
-	    /* Only directories or complete file paths can be relocated */
-	    if (!(names[i][len] == '/' || names[i][len] == '\0'))
+	    /* Only subdirectories or complete file paths may be relocated. We
+	       don't check for '\0' as our directory names all end in '/'. */
+	    if (!(dirNames[i][oplen] == '/'))
 		continue;
 
 	    if (relocations[j].newPath) { /* Relocate the path */
 		const char *s = relocations[j].newPath;
-		char *t = alloca(strlen(s) + strlen(names[i]) - len + 1);
+		char *t = alloca(strlen(s) + strlen(dirNames[i]) - oplen + 1);
+
 		strcpy(t, s);
-		strcat(t, names[i] + len);
-		rpmMessage(RPMMESS_DEBUG, _("relocating %s to %s\n"),
-			names[i], t);
-		names[i] = t;
+		strcat(t, dirNames[i] + oplen);
+		rpmMessage(RPMMESS_DEBUG, _("relocating directory %s to %s\n"),
+			dirNames[i], t);
+		dirNames[i] = t;
 		relocated = 1;
-	    } else /* On install, a relocate to NULL means skip the file */
-	    if (actions) {
-		actions[i] = FA_SKIPNSTATE;
-		rpmMessage(RPMMESS_DEBUG, _("excluding %s\n"), names[i]);
+	    } else if (actions) {
+		/* On install, a relocate to NULL means skip the file */
+		skipDirList[i] = 1;
+		rpmMessage(RPMMESS_DEBUG, _("excluding directory %s\n"), 
+			   dirNames[i]);
 	    }
 	    break;
 	}
@@ -375,16 +517,36 @@ static Header relocateFileList(struct availablePackage * alp,
 
     /* Save original filenames in header and replace (relocated) filenames. */
     if (relocated) {
-	const char ** origNames;
-	headerGetEntry(h, RPMTAG_FILENAMES, NULL, (void **) &origNames, NULL);
-	headerAddEntry(h, RPMTAG_ORIGFILENAMES, RPM_STRING_ARRAY_TYPE,
-			  origNames, fileCount);
-	xfree(origNames);
-	headerModifyEntry(h, RPMTAG_FILENAMES, RPM_STRING_ARRAY_TYPE,
-			  names, fileCount);
+	int c;
+	void * p;
+	int t;
+
+	p = NULL;
+	headerGetEntry(h, RPMTAG_BASENAMES, &t, &p, &c);
+	headerAddEntry(h, RPMTAG_ORIGBASENAMES, t, p, c);
+	xfree(p);
+
+	p = NULL;
+	headerGetEntry(h, RPMTAG_DIRNAMES, &t, &p, &c);
+	headerAddEntry(h, RPMTAG_ORIGDIRNAMES, t, p, c);
+	xfree(p);
+
+	p = NULL;
+	headerGetEntry(h, RPMTAG_DIRINDEXES, &t, &p, &c);
+	headerAddEntry(h, RPMTAG_ORIGDIRINDEXES, t, p, c);
+
+	headerModifyEntry(h, RPMTAG_BASENAMES, RPM_STRING_ARRAY_TYPE,
+			  baseNames, fileCount);
+	headerModifyEntry(h, RPMTAG_DIRNAMES, RPM_STRING_ARRAY_TYPE,
+			  dirNames, dirCount);
+	headerModifyEntry(h, RPMTAG_DIRINDEXES, RPM_INT32_TYPE,
+			  dirIndexes, fileCount);
     }
 
-    xfree(names);
+    free(baseNames);
+    free(dirNames);
+    if (filespec) free(filespec);
+    free(skipDirList);
 
     return h;
 }
@@ -392,12 +554,13 @@ static Header relocateFileList(struct availablePackage * alp,
 static int psTrim(rpmProblemSet filter, rpmProblemSet target)
 {
     /* As the problem sets are generated in an order solely dependent
-       on the ordering of the packages in the transaction, and that
-       ordering can't be changed, the problem sets must be parallel to
-       one another. Additionally, the filter set must be a subset of the
-       target set, given the operations available on transaction set.
-       This is good, as it lets us perform this trim in linear time, rather
-       then logarithmic or quadratic. */
+     * on the ordering of the packages in the transaction, and that
+     * ordering can't be changed, the problem sets must be parallel to
+     * one another. Additionally, the filter set must be a subset of the
+     * target set, given the operations available on transaction set.
+     * This is good, as it lets us perform this trim in linear time, rather
+     * then logarithmic or quadratic.
+     */
     rpmProblem * f, * t;
     int gotProblems = 0;
 
@@ -467,10 +630,11 @@ static enum fileTypes whatis(short mode)
     return result;
 }
 
-static enum fileActions decideFileFate(const char * filespec, short dbMode,
+static enum fileActions decideFileFate(const char * dirName,
+			const char * baseName, short dbMode,
 			const char * dbMd5, const char * dbLink, short newMode,
 			const char * newMd5, const char * newLink, int newFlags,
-				int brokenMd5, int transFlags)
+			int brokenMd5, int transFlags)
 {
     char buffer[1024];
     const char * dbAttr, * newAttr;
@@ -478,6 +642,9 @@ static enum fileActions decideFileFate(const char * filespec, short dbMode,
     struct stat sb;
     int i, rc;
     int save = (newFlags & RPMFILE_NOREPLACE) ? FA_ALTNAME : FA_SAVE;
+    char * filespec = alloca(strlen(dirName) + strlen(baseName) + 1);
+
+    sprintf(filespec, "%s%s", dirName, baseName);
 
     if (lstat(filespec, &sb)) {
 	/*
@@ -592,7 +759,8 @@ static int handleInstInstalledFiles(TFI_t * fi, rpmdb db,
     uint_16 * otherModes;
     int numReplaced = 0;
 
-    if ((h = rpmdbGetRecord(db, shared->otherPkg)) == NULL)
+    h = rpmdbGetRecord(db, shared->otherPkg);
+    if (h == NULL)
 	return 1;
 
     headerGetEntryMinMemory(h, RPMTAG_FILEMD5S, NULL,
@@ -625,8 +793,9 @@ static int handleInstInstalledFiles(TFI_t * fi, rpmdb db,
 			fi->fmd5s[fileNum],
 			fi->flinks[fileNum])) {
 	    if (reportConflicts)
-		psAppend(probs, RPMPROB_FILE_CONFLICT, fi->ap->key,
-			 fi->ap->h, fi->fl[fileNum], h,0 );
+		psAppendFile(probs, RPMPROB_FILE_CONFLICT, fi->ap->key,
+			 fi->ap->h, fi->dnl[fi->dil[fileNum]], 
+			 fi->bnl[fileNum], h, 0);
 	    if (!(otherFlags[otherFileNum] | fi->fflags[fileNum])
 			& RPMFILE_CONFIG) {
 		if (!shared->isRemoved)
@@ -635,7 +804,9 @@ static int handleInstInstalledFiles(TFI_t * fi, rpmdb db,
 	}
 
 	if ((otherFlags[otherFileNum] | fi->fflags[fileNum]) & RPMFILE_CONFIG) {
-	    fi->actions[fileNum] = decideFileFate(fi->fl[fileNum],
+	    fi->actions[fileNum] = decideFileFate(
+			fi->dnl[fi->dil[fileNum]],
+			fi->bnl[fileNum],
 			otherModes[otherFileNum],
 			otherMd5s[otherFileNum],
 			otherLinks[otherFileNum],
@@ -669,7 +840,8 @@ static int handleRmvdInstalledFiles(TFI_t * fi, rpmdb db,
     const char * otherStates;
     int i;
    
-    if ((h = rpmdbGetRecord(db, shared->otherPkg)) == NULL)
+    h = rpmdbGetRecord(db, shared->otherPkg);
+    if (h == NULL)
 	return 1;
 
     headerGetEntryMinMemory(h, RPMTAG_FILESTATES, NULL,
@@ -697,6 +869,8 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
     int i, j;
     struct diskspaceInfo * ds = NULL;
     uint_32 fixupSize = 0;
+    char * filespec = NULL;
+    int fileSpecAlloced = 0;
   
     for (i = 0; i < fi->fc; i++) {
 	int otherPkgNum, otherFileNum;
@@ -706,9 +880,17 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
 	if (XFA_SKIPPING(fi->actions[i]))
 	    continue;
 
+	j = strlen(fi->dnl[fi->dil[i]]) + strlen(fi->bnl[i]) + 1;
+	if (j > fileSpecAlloced) {
+	    fileSpecAlloced = j * 2;
+	    filespec = xrealloc(filespec, fileSpecAlloced);
+	}
+
+	sprintf(filespec, "%s%s", fi->dnl[fi->dil[i]], fi->bnl[i]);
+
 	if (dsl) {
 	    ds = dsl;
-	    while (ds->block && ds->dev != fi->fps[i].dev) ds++;
+	    while (ds->block && ds->dev != fi->fps[i].entry->dev) ds++;
 	    if (!ds->block) ds = NULL;
 	    fixupSize = 0;
 	}
@@ -770,7 +952,8 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
 		/* XXX is this test still necessary? */
 		if (fi->actions[i] != FA_UNKNOWN)
 		    break;
-		if ((fi->fflags[i] & RPMFILE_CONFIG) && !lstat(fi->fl[i], &sb)) {
+		if ((fi->fflags[i] & RPMFILE_CONFIG) && 
+			!lstat(filespec, &sb)) {
 		    /* Here is a non-overlapped pre-existing config file. */
 		    fi->actions[i] = (fi->fflags[i] & RPMFILE_NOREPLACE)
 			? FA_ALTNAME : FA_BACKUP;
@@ -788,13 +971,13 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
 			fi->fmd5s[i],
 			fi->flinks[i])) {
 		psAppend(probs, RPMPROB_NEW_FILE_CONFLICT, fi->ap->key,
-			 fi->ap->h, fi->fl[i], recs[otherPkgNum]->ap->h, 0);
+			 fi->ap->h, filespec, recs[otherPkgNum]->ap->h, 0);
 	    }
 
 	    /* Try to get the disk accounting correct even if a conflict. */
 	    fixupSize = recs[otherPkgNum]->fsizes[otherFileNum];
 
-	    if ((fi->fflags[i] & RPMFILE_CONFIG) && !lstat(fi->fl[i], &sb)) {
+	    if ((fi->fflags[i] & RPMFILE_CONFIG) && !lstat(filespec, &sb)) {
 		/* Here is an overlapped  pre-existing config file. */
 		fi->actions[i] = (fi->fflags[i] & RPMFILE_NOREPLACE)
 			? FA_ALTNAME : FA_SKIP;
@@ -829,7 +1012,7 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
 		
 	    /* Here is a pre-existing modified config file that needs saving. */
 	    {	char mdsum[50];
-		if (!mdfile(fi->fl[i], mdsum) && strcmp(fi->fmd5s[i], mdsum)) {
+		if (!mdfile(filespec, mdsum) && strcmp(fi->fmd5s[i], mdsum)) {
 		    fi->actions[i] = FA_BACKUP;
 		    break;
 		}
@@ -867,10 +1050,11 @@ static void handleOverlappedFiles(TFI_t * fi, hashTable ht,
 	    ds->needed -= BLOCK_ROUND(fixupSize, ds->block);
 	}
     }
+    if (filespec) free(filespec);
 }
 
 static int ensureOlder(rpmdb db, Header new, int dbOffset, rpmProblemSet probs,
-		       const void * key)
+			/*@dependent@*/ const void * key)
 {
     Header old;
     int result, rc = 0;
@@ -920,9 +1104,11 @@ static void skipFiles(TFI_t * fi, int noDocs)
     if (s) {
 	languages = (const char **) splitString(s, strlen(s), ':');
 	xfree(s);
+#ifdef	DYING
     /* XXX LINGUAS/LANG is used by the installer so leave alone for now */
     } else if ((s = getenv("LINGUAS")) || (s = getenv("LANG")) || (s = "en")) {
 	languages = (const char **) splitString(s, strlen(s), ':');
+#endif
     } else
 	languages = NULL;
 
@@ -940,12 +1126,14 @@ static void skipFiles(TFI_t * fi, int noDocs)
 	 */
 	for (nsp = netsharedPaths; nsp && *nsp; nsp++) {
 	    int len;
+	    const char * dir = fi->dnl[fi->dil[i]];
+
 	    len = strlen(*nsp);
-	    if (strncmp(fi->fl[i], *nsp, len))
+	    if (strncmp(dir, *nsp, len))
 		continue;
 
 	    /* Only directories or complete file paths can be net shared */
-	    if (!(fi->fl[i][len] == '/' || fi->fl[i][len] == '\0'))
+	    if (!(dir[len] == '/' || dir[len] == '\0'))
 		continue;
 	    break;
 	}
@@ -990,7 +1178,7 @@ static void skipFiles(TFI_t * fi, int noDocs)
     if (languages) freeSplitString((char **)languages);
 }
 
-#define	NOTIFY(_x)	if (notify) notify _x
+#define	NOTIFY(_x)	if (notify) (void) notify _x
 
 /* Return -1 on error, > 0 if newProbs is set, 0 if everything happened */
 
@@ -1019,6 +1207,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
     int filesystemCount;
     struct diskspaceInfo * di = NULL;
     int oc;
+    fingerPrintCache fpc;
 
     /* FIXME: what if the same package is included in ts twice? */
 
@@ -1061,7 +1250,6 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 		di[i].avail = sfb.f_blocks - sfb.f_bfree;
 #endif
 
-
 		stat(filesystems[i], &sb);
 		di[i].dev = sb.st_dev;
 	    }
@@ -1070,8 +1258,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	if (di) di[i].block = 0;
     }
 
-    probs = psCreate();
-    *newProbs = probs;
+    probs = *newProbs = psCreate();
     hdrs = alloca(sizeof(*hdrs) * ts->addedPackages.size);
 
     /* ===============================================
@@ -1083,14 +1270,13 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
      * - count files.
      */
     /* The ordering doesn't matter here */
-    for (alp = ts->addedPackages.list; (alp - ts->addedPackages.list) <
-	    ts->addedPackages.size; alp++) {
+    for (alp = ts->addedPackages.list;
+	(alp - ts->addedPackages.list) < ts->addedPackages.size; alp++) {
 	if (!archOkay(alp->h) && !(ignoreSet & RPMPROB_FILTER_IGNOREARCH))
 	    psAppend(probs, RPMPROB_BADARCH, alp->key, alp->h, NULL, NULL, 0);
 
-	if (!osOkay(alp->h) && !(ignoreSet & RPMPROB_FILTER_IGNOREOS)) {
+	if (!osOkay(alp->h) && !(ignoreSet & RPMPROB_FILTER_IGNOREOS))
 	    psAppend(probs, RPMPROB_BADOS, alp->key, alp->h, NULL, NULL, 0);
-	}
 
 	if (!(ignoreSet & RPMPROB_FILTER_OLDPACKAGE)) {
 	    rc = rpmdbFindPackage(ts->db, alp->name, &dbi);
@@ -1115,7 +1301,8 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	    dbiFreeIndexRecord(dbi);
 	}
 
-	if (headerGetEntry(alp->h, RPMTAG_FILENAMES, NULL, NULL, &fileCount))
+	if (headerGetEntry(alp->h, RPMTAG_BASENAMES, NULL, NULL, 
+			   &fileCount))
 	    totalFileCount += fileCount;
     }
 
@@ -1125,7 +1312,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	Header h;
 
 	if ((h = rpmdbGetRecord(ts->db, ts->removedPackages[i]))) {
-	    if (headerGetEntry(h, RPMTAG_FILENAMES, NULL, NULL,
+	    if (headerGetEntry(h, RPMTAG_BASENAMES, NULL, NULL,
 			       &fileCount))
 		totalFileCount += fileCount;
 	    headerFree(h);
@@ -1138,8 +1325,6 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
     flEntries = ts->addedPackages.size + ts->numRemovedPackages;
     flList = alloca(sizeof(*flList) * (flEntries));
 
-    ht = htCreate(totalFileCount * 2, 0, fpHashFunction, fpEqual);
-
     /* FIXME?: we'd be better off assembling one very large file list and
        calling fpLookupList only once. I'm not sure that the speedup is
        worth the trouble though. */
@@ -1151,7 +1336,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	    i = ts->order[oc].u.addedIndex;
 	    alp = ts->addedPackages.list + ts->order[oc].u.addedIndex;
 
-	    if (!headerGetEntryMinMemory(alp->h, RPMTAG_FILENAMES, NULL,
+	    if (!headerGetEntryMinMemory(alp->h, RPMTAG_BASENAMES, NULL,
 					 NULL, &fi->fc)) {
 		fi->h = headerLink(alp->h);
 		hdrs[i] = headerLink(fi->h);
@@ -1176,14 +1361,18 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	    fi->type = TR_REMOVED;
 	    break;
 	}
-
-	if (!headerGetEntry(fi->h, RPMTAG_FILENAMES, NULL,
-				     (void **) &fi->fl, &fi->fc)) {
+	if (!headerGetEntry(fi->h, RPMTAG_BASENAMES, NULL,
+				     (void **) &fi->bnl, &fi->fc)) {
 	    /* This catches removed packages w/ no file lists */
 	    fi->fc = 0;
 	    continue;
 	}
 
+	headerGetEntry(fi->h, RPMTAG_DIRINDEXES, NULL, (void **) &fi->dil, 
+		       NULL);
+	headerGetEntry(fi->h, RPMTAG_DIRNAMES, NULL, (void **) &fi->dnl, 
+		       NULL);
+
 	/* actions is initialized earlier for added packages */
 	if (fi->actions == NULL)
 	    fi->actions = xcalloc(fi->fc, sizeof(*fi->actions));
@@ -1211,6 +1400,8 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 				fi->fmodes, fi->fc * sizeof(*fi->fmodes));
 	    fi->fstates = memcpy(xmalloc(fi->fc * sizeof(*fi->fstates)),
 				fi->fstates, fi->fc * sizeof(*fi->fstates));
+	    fi->dil = memcpy(xmalloc(fi->fc * sizeof(*fi->dil)),
+				fi->dil, fi->fc * sizeof(*fi->dil));
 	    headerFree(fi->h);
 	    fi->h = NULL;
 	    break;
@@ -1240,13 +1431,16 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
     }
 
     chdir("/");
-    chroot(ts->root);
+    /*@-unrecog@*/ chroot(ts->root); /*@=unrecog@*/
+
+    ht = htCreate(totalFileCount * 2, 0, 0, fpHashFunction, fpEqual);
+    fpc = fpCacheCreate(totalFileCount);
 
     /* ===============================================
      * Add fingerprint for each file not skipped.
      */
     for (fi = flList; (fi - flList) < flEntries; fi++) {
-	fpLookupList(fi->fl, fi->fps, fi->fc, 1);
+	fpLookupList(fpc, fi->dnl, fi->bnl, fi->dil, fi->fc, fi->fps);
 	for (i = 0; i < fi->fc; i++) {
 	    if (XFA_SKIPPING(fi->actions[i]))
 		continue;
@@ -1382,7 +1576,11 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
     for (oc = 0, fi = flList; oc < ts->orderCount; oc++, fi++) {
 	if (fi->fc == 0)
 	    continue;
-	free(fi->fl); fi->fl = NULL;
+	free(fi->bnl); fi->bnl = NULL;
+	free(fi->dnl); fi->dnl = NULL;
+#ifdef	DOUBLE_FREE
+	xfree(fi->dil); fi->dil = NULL;
+#endif
 	switch (fi->type) {
 	case TR_ADDED:
 	    free(fi->fmd5s); fi->fmd5s = NULL;
@@ -1395,6 +1593,8 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	}
     }
 
+    fpCacheFree(fpc);
+
     /* ===============================================
      * If unfiltered problems exist, free memory and return.
      */
@@ -1431,7 +1631,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 		    headerFree(hdrs[i]);
 		    rc = rpmReadPackageHeader(fd, &h, NULL, NULL, NULL);
 		    if (rc) {
-			notify(fi->h, RPMCALLBACK_INST_CLOSE_FILE, 0, 0,
+			(void)notify(fi->h, RPMCALLBACK_INST_CLOSE_FILE, 0, 0,
 				    alp->key, notifyData);
 			ourrc++;
 			fd = NULL;
@@ -1459,7 +1659,7 @@ int rpmRunTransactions(rpmTransactionSet ts, rpmCallbackFunction notify,
 	    headerFree(hdrs[i]);
 
 	    if (alp->fd == NULL && fd)
-		notify(fi->h, RPMCALLBACK_INST_CLOSE_FILE, 0, 0, alp->key,
+		(void)notify(fi->h, RPMCALLBACK_INST_CLOSE_FILE, 0, 0, alp->key,
 		   notifyData);
 	    break;
 	case TR_REMOVED:
