@@ -1,6 +1,8 @@
 #include "system.h"
 
 #include <rpmlib.h>
+#include <rpmurl.h>
+#include <rpmmacro.h>	/* XXX for rpmExpand */
 
 #include "depends.h"
 #include "install.h"
@@ -10,6 +12,8 @@
 
 static char * SCRIPT_PATH = "PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/X11R6/bin";
 
+#define	SUFFIX_RPMSAVE	".rpmsave"
+
 static int removeFile(const char * file, unsigned int flags, short mode, 
 		      enum fileActions action)
 {
@@ -19,9 +23,9 @@ static int removeFile(const char * file, unsigned int flags, short mode,
     switch (action) {
 
       case FA_BACKUP:
-	newfile = alloca(strlen(file) + 20);
-	strcpy(newfile, file);
-	strcat(newfile, ".rpmsave");
+	newfile = alloca(strlen(file) + sizeof(SUFFIX_RPMSAVE));
+	stpcpy(stpcpy(newfile, file), SUFFIX_RPMSAVE);
+
 	if (rename(file, newfile)) {
 	    rpmError(RPMERR_RENAME, _("rename of %s to %s failed: %s"),
 			file, newfile, strerror(errno));
@@ -65,6 +69,7 @@ static int removeFile(const char * file, unsigned int flags, short mode,
     return 0;
 }
 
+/** */
 int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset, 
 			int flags, enum fileActions * actions, FD_t scriptFd)
 {
@@ -72,7 +77,7 @@ int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset,
     int i;
     int fileCount;
     const char * name, * version, * release;
-    const char ** fileList;
+    const char ** baseNames;
     int type;
     int scriptArg;
     int rc = 0;
@@ -90,8 +95,10 @@ int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset,
 
     headerNVR(h, &name, &version, &release);
 
-    /* when we run scripts, we pass an argument which is the number of 
-       versions of this package that will be installed when we are finished */
+    /*
+     * When we run scripts, we pass an argument which is the number of 
+     * versions of this package that will be installed when we are finished.
+     */
     {	dbiIndexSet matches;
 	if (rpmdbFindPackage(db, name, &matches)) {
 	    rpmError(RPMERR_DBCORRUPT, _("cannot read packages named %s for uninstall"),
@@ -131,32 +138,42 @@ int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset,
 		flags & RPMTRANS_FLAG_TEST);
 
     if (!(flags & RPMTRANS_FLAG_JUSTDB) &&
-	headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, 
+	headerGetEntry(h, RPMTAG_BASENAMES, NULL, (void **) &baseNames, 
 	               &fileCount)) {
 	const char ** fileMd5List;
 	uint_32 * fileFlagsList;
 	int_16 * fileModesList;
-	char * fnbuffer = NULL;
-	int prefixlen = 0;
+	const char ** dirNames;
+	int_32 * dirIndexes;
+	char * fileName;
+	int fnmaxlen;
+	int prefixlen = (prefix && !(prefix[0] == '/' && prefix[1] == '\0'))
+			? strlen(prefix) : 0;
 
-	/* Get alloca buffer for largest possible prefix + filename. */
-	if (prefix && prefix[0] != '\0') {
-	    int fnbuffersize = 0;
-	    size_t fnlen;
+	headerGetEntry(h, RPMTAG_DIRINDEXES, NULL, (void **) &dirIndexes,
+	               NULL);
+	headerGetEntry(h, RPMTAG_DIRNAMES, NULL, (void **) &dirNames,
+	               NULL);
 
-	    prefixlen = strlen(prefix);
-	    for (i = 0; i < fileCount; i++) {
-		if ((fnlen = strlen(fileList[i])) > fnbuffersize)
-		    fnbuffersize = fnlen;
-	    }
-	    fnbuffersize += strlen(prefix) + sizeof("/");
-	    fnbuffer = alloca(fnbuffersize);
-	    strcpy(fnbuffer, prefix);
-	    if (fnbuffer[prefixlen-1] != '/') {
-		fnbuffer[prefixlen++] = '/';
-		fnbuffer[prefixlen] = '\0';
-	    }
+	/* Get buffer for largest possible prefix + dirname + filename. */
+	fnmaxlen = 0;
+	for (i = 0; i < fileCount; i++) {
+		size_t fnlen;
+		fnlen = strlen(baseNames[i]) + 
+			strlen(dirNames[dirIndexes[i]]);
+		if (fnlen > fnmaxlen)
+		    fnmaxlen = fnlen;
 	}
+	fnmaxlen += prefixlen + sizeof("/");	/* XXX one byte too many */
+
+	fileName = alloca(fnmaxlen);
+
+	if (prefixlen) {
+	    strcpy(fileName, prefix);
+	    rpmCleanPath(fileName);
+	    prefixlen = strlen(fileName);
+	} else
+	    *fileName = '\0';
 
 	headerGetEntry(h, RPMTAG_FILEMD5S, &type, (void **) &fileMd5List, 
 		 &fileCount);
@@ -167,23 +184,20 @@ int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset,
 
 	/* Traverse filelist backwards to help insure that rmdir() will work. */
 	for (i = fileCount - 1; i >= 0; i--) {
-	    const char * fn;
 
-	    fn = fileList[i];
-	    if (prefixlen) {
-		if (*fn == '/') fn++;
-		strcpy(fnbuffer + prefixlen, fn);
-		fn = fnbuffer;
-	    }
+	    /* XXX this assumes that dirNames always starts/ends with '/' */
+	    stpcpy(stpcpy(fileName+prefixlen, dirNames[dirIndexes[i]]), baseNames[i]);
 
 	    rpmMessage(RPMMESS_DEBUG, _("   file: %s action: %s\n"),
-			fn, fileActionString(actions[i]));
+			fileName, fileActionString(actions[i]));
 
 	    if (!(flags & RPMTRANS_FLAG_TEST))
-		removeFile(fn, fileFlagsList[i], fileModesList[i], actions[i]);
+		removeFile(fileName, fileFlagsList[i], fileModesList[i], 
+			   actions[i]);
 	}
 
-	free(fileList);
+	free(baseNames);
+	free(dirNames);
 	free(fileMd5List);
     }
 
@@ -201,9 +215,15 @@ int removeBinaryPackage(const char * prefix, rpmdb db, unsigned int offset,
 	}
     }
 
-    rpmMessage(RPMMESS_DEBUG, _("removing database entry\n"));
-    if (!(flags & RPMTRANS_FLAG_TEST))
+    if (!(flags & RPMTRANS_FLAG_TEST)) {
+	rpmMessage(RPMMESS_DEBUG, _("removing database entry\n"));
+	/*
+	 * XXX rpmdbRemove used to have 2nd arg, tolerant = 0. This generates
+	 * XXX an error message because of a db-1.85 hash access bug that
+	 * XXX cannot be easily fixed. So, we turn off the error message.
+	 */
 	rpmdbRemove(db, offset, 0);
+    }
 
     rc = 0;
 
@@ -272,10 +292,10 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
 
 	if (rpmIsDebug() &&
 	    (!strcmp(argv[0], "/bin/sh") || !strcmp(argv[0], "/bin/bash")))
-	    (void)fdWrite(fd, "set -x\n", 7);
+	    (void)Fwrite("set -x\n", sizeof(char), 7, fd);
 
-	(void)fdWrite(fd, script, strlen(script));
-	fdClose(fd);
+	(void)Fwrite(script, sizeof(script[0]), strlen(script), fd);
+	Fclose(fd);
 
 	argv[argc++] = fn + strlen(root);
 	if (arg1 >= 0) {
@@ -294,14 +314,16 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
 
     if (errfd != NULL) {
 	if (rpmIsVerbose()) {
-	    out = fdDup(fdFileno(errfd));
+	    out = fdDup(Fileno(errfd));
 	} else {
-	    out = fdOpen("/dev/null", O_WRONLY, 0);
-	    if (fdFileno(out) < 0)
-		out = fdDup(fdFileno(errfd));
+	    out = Fopen("/dev/null", "w.fdio");
+	    if (Ferror(out)) {
+		out = fdDup(Fileno(errfd));
+	    }
 	}
     } else {
 	out = fdDup(STDOUT_FILENO);
+	out = fdLink(out, "runScript persist");
     }
     
     if (!(child = fork())) {
@@ -315,18 +337,27 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
 	close(pipes[0]);
 
 	if (errfd != NULL) {
-	    if (fdFileno(errfd) != STDERR_FILENO)
-		dup2(fdFileno(errfd), STDERR_FILENO);
-	    if (fdFileno(out) != STDOUT_FILENO)
-		dup2(fdFileno(out), STDOUT_FILENO);
+	    if (Fileno(errfd) != STDERR_FILENO)
+		dup2(Fileno(errfd), STDERR_FILENO);
+	    if (Fileno(out) != STDOUT_FILENO)
+		dup2(Fileno(out), STDOUT_FILENO);
 	    /* make sure we don't close stdin/stderr/stdout by mistake! */
-	    if (fdFileno(out) > STDERR_FILENO && fdFileno(out) != fdFileno(errfd))
-		fdClose (out);
-	    if (fdFileno(errfd) > STDERR_FILENO)
-		fdClose (errfd);
+	    if (Fileno(out) > STDERR_FILENO && Fileno(out) != Fileno(errfd)) {
+		Fclose (out);
+	    }
+	    if (Fileno(errfd) > STDERR_FILENO) {
+		Fclose (errfd);
+	    }
 	}
 
-	doputenv(SCRIPT_PATH);
+	{   const char *ipath = rpmExpand("PATH=%{_install_script_path}", NULL);
+	    const char *path = SCRIPT_PATH;
+
+	    if (ipath && ipath[5] != '%')
+		path = ipath;
+	    doputenv(path);
+	    if (ipath)	xfree(ipath);
+	}
 
 	for (i = 0; i < numPrefixes; i++) {
 	    sprintf(prefixBuf, "RPM_INSTALL_PREFIX%d=%s", i, prefixes[i]);
@@ -339,13 +370,21 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
 	    }
 	}
 
-	if (strcmp(root, "/")) {
-	    /*@-unrecog@*/ chroot(root); /*@=unrecog@*/
+	switch(urlIsURL(root)) {
+	case URL_IS_PATH:
+	    root += sizeof("file://") - 1;
+	    root = strchr(root, '/');
+	    /*@fallthrough@*/
+	case URL_IS_UNKNOWN:
+	    if (strcmp(root, "/"))
+		/*@-unrecog@*/ chroot(root); /*@=unrecog@*/
+	    chdir("/");
+	    execv(argv[0], (char *const *)argv);
+	    break;
+	default:
+	    break;
 	}
 
-	chdir("/");
-
-	execv(argv[0], (char *const *)argv);
  	_exit(-1);
 	/*@notreached@*/
     }
@@ -354,7 +393,7 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
 
     if (freePrefixes) free(prefixes);
 
-    fdClose(out);	/* XXX dup'd STDOUT_FILENO */
+    Fclose(out);	/* XXX dup'd STDOUT_FILENO */
     
     if (script) {
 	if (!rpmIsDebug()) unlink(fn);
@@ -369,6 +408,7 @@ static int runScript(Header h, const char * root, int progArgc, const char ** pr
     return 0;
 }
 
+/** */
 int runInstScript(const char * root, Header h, int scriptTag, int progTag,
 	          int arg, int norunScripts, FD_t err)
 {
@@ -482,6 +522,7 @@ static int handleOneTrigger(const char * root, rpmdb db, int sense, Header sourc
     return rc;
 }
 
+/** */
 int runTriggers(const char * root, rpmdb db, int sense, Header h,
 		int countCorrection, FD_t scriptFd)
 {
@@ -521,6 +562,7 @@ int runTriggers(const char * root, rpmdb db, int sense, Header h,
     
 }
 
+/** */
 int runImmedTriggers(const char * root, rpmdb db, int sense, Header h,
 		     int countCorrection, FD_t scriptFd)
 {
