@@ -31,6 +31,7 @@ static void verifyArgCallback(/*@unused@*/poptContext con,
 }
 
 static int noFiles = 0;
+/** */
 struct poptOption rpmVerifyPoptTable[] = {
 	{ NULL, '\0', POPT_ARG_CALLBACK | POPT_CBFLAG_INC_DATA, 
 		verifyArgCallback, 0, NULL, NULL },
@@ -41,12 +42,13 @@ struct poptOption rpmVerifyPoptTable[] = {
 
 /* ======================================================================== */
 /* XXX static */
+/** */
 int rpmVerifyFile(const char * prefix, Header h, int filenum, int * result, 
 		  int omitMask)
 {
-    char ** fileList, ** md5List, ** linktoList;
+    char ** baseNames, ** md5List, ** linktoList, ** dirNames;
     int_32 * verifyFlags, flags;
-    int_32 * sizeList, * mtimeList;
+    int_32 * sizeList, * mtimeList, * dirIndexes;
     unsigned short * modeList, * rdevList;
     char * fileStatesList;
     char * filespec;
@@ -90,13 +92,18 @@ int rpmVerifyFile(const char * prefix, Header h, int filenum, int * result,
 	flags = RPMVERIFY_ALL;
     }
 
-    headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, &count);
-    filespec = alloca(strlen(fileList[filenum]) + strlen(prefix) + 5);
-    strcpy(filespec, prefix);
-    strcat(filespec, "/");
-    strcat(filespec, fileList[filenum]);
+    headerGetEntry(h, RPMTAG_BASENAMES, &type, (void **) &baseNames, 
+		   &count);
+    headerGetEntry(h, RPMTAG_DIRINDEXES, &type, (void **) &dirIndexes, 
+		   NULL);
+    headerGetEntry(h, RPMTAG_DIRNAMES, &type, (void **) &dirNames, NULL);
 
-    free(fileList);
+    filespec = alloca(strlen(dirNames[dirIndexes[filenum]]) + 
+		      strlen(baseNames[filenum]) + strlen(prefix) + 5);
+    sprintf(filespec, "%s/%s%s", prefix, dirNames[dirIndexes[filenum]],
+		baseNames[filenum]);
+    free(baseNames);
+    free(dirNames);
     
     *result = 0;
 
@@ -153,7 +160,7 @@ int rpmVerifyFile(const char * prefix, Header h, int filenum, int * result,
     } 
     if (flags & RPMVERIFY_LINKTO) {
 	headerGetEntry(h, RPMTAG_FILELINKTOS, &type, (void **) &linktoList, &count);
-	size = readlink(filespec, linkto, sizeof(linkto));
+	size = readlink(filespec, linkto, sizeof(linkto)-1);
 	if (size == -1)
 	    *result |= (RPMVERIFY_READLINKFAIL|RPMVERIFY_LINKTO);
 	else  {
@@ -233,6 +240,7 @@ int rpmVerifyFile(const char * prefix, Header h, int filenum, int * result,
 }
 
 /* XXX static */
+/** */
 int rpmVerifyScript(const char * root, Header h, FD_t err)
 {
     return runInstScript(root, h, RPMTAG_VERIFYSCRIPT, RPMTAG_VERIFYSCRIPTPROG,
@@ -242,8 +250,8 @@ int rpmVerifyScript(const char * root, Header h, FD_t err)
 /* ======================================================================== */
 static int verifyHeader(QVA_t *qva, Header h)
 {
-    const char ** fileList;
-    int count, type;
+    const char ** fileNames;
+    int count;
     int verifyResult;
     int i, ec, rc;
     int_32 * fileFlagsList;
@@ -253,15 +261,17 @@ static int verifyHeader(QVA_t *qva, Header h)
     if (!(qva->qva_flags & VERIFY_MD5)) omitMask = RPMVERIFY_MD5;
 
     if (headerGetEntry(h, RPMTAG_FILEFLAGS, NULL, (void **) &fileFlagsList, NULL) &&
-       headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, &count)) {
+	headerIsEntry(h, RPMTAG_BASENAMES)) {
+	rpmBuildFileList(h, &fileNames, &count);
+
 	for (i = 0; i < count; i++) {
 	    if ((rc = rpmVerifyFile(qva->qva_prefix, h, i, &verifyResult, omitMask)) != 0) {
-		fprintf(stdout, _("missing    %s\n"), fileList[i]);
+		fprintf(stdout, _("missing    %s\n"), fileNames[i]);
 	    } else {
 		const char * size, * md5, * link, * mtime, * mode;
 		const char * group, * user, * rdev;
-		static const char * aok = ".";
-		static const char * unknown = "?";
+		/*@observer@*/ static const char *const aok = ".";
+		/*@observer@*/ static const char *const unknown = "?";
 
 		if (!verifyResult) continue;
 	    
@@ -292,13 +302,13 @@ static int verifyHeader(QVA_t *qva, Header h)
 		fprintf(stdout, "%s%s%s%s%s%s%s%s %c %s\n",
 		       size, mode, md5, rdev, link, user, group, mtime, 
 		       fileFlagsList[i] & RPMFILE_CONFIG ? 'c' : ' ', 
-		       fileList[i]);
+		       fileNames[i]);
 	    }
 	    if (rc)
 		ec = rc;
 	}
 	
-	free(fileList);
+	free(fileNames);
     }
     return ec;
 }
@@ -335,6 +345,7 @@ static int verifyDependencies(/*@only@*/rpmdb db, Header h) {
     return 0;
 }
 
+/** */
 int showVerifyPackage(QVA_t *qva, rpmdb db, Header h)
 {
     int ec, rc;
@@ -350,10 +361,11 @@ int showVerifyPackage(QVA_t *qva, rpmdb db, Header h)
     if ((qva->qva_flags & VERIFY_SCRIPT) &&
 	(rc = rpmVerifyScript(qva->qva_prefix, h, fdo)) != 0)
 	    ec = rc;
-    fdClose(fdo);
+    Fclose(fdo);
     return ec;
 }
 
+/** */
 int rpmVerify(QVA_t *qva, enum rpmQVSources source, const char *arg)
 {
     rpmdb db = NULL;
