@@ -4,9 +4,12 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <unistd.h>
+#include <glob.h>	/* XXX rpmio.h */
+#include <dirent.h>	/* XXX rpmio.h */
 
 #include "Python.h"
 #include "rpmlib.h"
+#include "misc.h"
 #include "rpmmacro.h"
 #include "upgrade.h"
 
@@ -37,9 +40,13 @@ static PyObject * hdrSubscript(hdrObject * s, PyObject * item);
 static PyObject * hdrKeyList(hdrObject * s, PyObject * args);
 static PyObject * hdrUnload(hdrObject * s, PyObject * args);
 static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args);
+static PyObject * hdrCompressFilelist(hdrObject * s, PyObject * args);
+static PyObject * hdrExpandFilelist(hdrObject * s, PyObject * args);
+static PyObject * hdrFullFilelist(hdrObject * s, PyObject * args);
 
 void initrpm(void);
 static PyObject * doAddMacro(PyObject * self, PyObject * args);
+static PyObject * doDelMacro(PyObject * self, PyObject * args);
 static rpmdbObject * rpmOpenDB(PyObject * self, PyObject * args);
 static PyObject * hdrLoad(PyObject * self, PyObject * args);
 static PyObject * rpmHeaderFromPackage(PyObject * self, PyObject * args);
@@ -60,26 +67,31 @@ static PyObject * rpmtransRun(rpmtransObject * s, PyObject * args);
 static PyObject * rpmtransOrder(rpmtransObject * s, PyObject * args);
 static void rpmtransDealloc(PyObject * o);
 static PyObject * rpmtransGetAttr(rpmtransObject * o, char * name);
-static int rpmtransSetAttr(rpmtransObject * o, char * name, 
+static int rpmtransSetAttr(rpmtransObject * o, char * name,
 			   PyObject * val);
+static PyObject * doFopen(PyObject * self, PyObject * args);
 
 /* Types */
 
 static PyMethodDef rpmModuleMethods[] = {
     { "TransactionSet", (PyCFunction) rpmtransCreate, METH_VARARGS, NULL },
     { "addMacro", (PyCFunction) doAddMacro, METH_VARARGS, NULL },
+    { "delMacro", (PyCFunction) doDelMacro, METH_VARARGS, NULL },
     { "archscore", (PyCFunction) archScore, METH_VARARGS, NULL },
     { "findUpgradeSet", (PyCFunction) findUpgradeSet, METH_VARARGS, NULL },
     { "headerFromPackage", (PyCFunction) rpmHeaderFromPackage, METH_VARARGS, NULL },
     { "headerLoad", (PyCFunction) hdrLoad, METH_VARARGS, NULL },
     { "opendb", (PyCFunction) rpmOpenDB, METH_VARARGS, NULL },
     { "rebuilddb", (PyCFunction) rebuildDB, METH_VARARGS, NULL },
-    { "readHeaderListFromFD", (PyCFunction) rpmHeaderFromFD, METH_VARARGS, NULL },    
+    { "readHeaderListFromFD", (PyCFunction) rpmHeaderFromFD, METH_VARARGS, NULL },
     { "readHeaderListFromFile", (PyCFunction) rpmHeaderFromFile, METH_VARARGS, NULL },
     { "errorSetCallback", (PyCFunction) errorSetCallback, METH_VARARGS, NULL },
     { "errorString", (PyCFunction) errorString, METH_VARARGS, NULL },
     { "versionCompare", (PyCFunction) versionCompare, METH_VARARGS, NULL },
     { "labelCompare", (PyCFunction) labelCompare, METH_VARARGS, NULL },
+    /* don't use me yet
+    { "Fopen", (PyCFunction) doFopen, METH_VARARGS, NULL },
+    */
     { NULL }
 } ;
 
@@ -197,18 +209,28 @@ static struct PyMethodDef hdrMethods[] = {
 	{"keys",	(PyCFunction) hdrKeyList,	1 },
 	{"unload",	(PyCFunction) hdrUnload,	1 },
 	{"verifyFile",	(PyCFunction) hdrVerifyFile,	1 },
+	{"expandFilelist",	(PyCFunction) hdrExpandFilelist,	1 },
+	{"compressFilelist",	(PyCFunction) hdrCompressFilelist,	1 },
+	{"fullFilelist",	(PyCFunction) hdrFullFilelist,	1 },
 	{NULL,		NULL}		/* sentinel */
 };
 
 /* External functions */
 int mdfile(const char *fn, unsigned char *digest);
-    
+
 /* Code */
 
+extern int _rpmio_debug;
+
 void initrpm(void) {
     PyObject * m, * d, * tag, * dict;
     int i;
+    const struct headerSprintfExtension * extensions = rpmHeaderFormats;
+    struct headerSprintfExtension * ext;
 
+/*      i18ndomains = "redhat-dist"; */
+   
+/*      _rpmio_debug = -1;  */
     rpmReadConfigFiles(NULL, NULL);
 
     m = Py_InitModule("rpm", rpmModuleMethods);
@@ -221,29 +243,38 @@ void initrpm(void) {
 
     for (i = 0; i < rpmTagTableSize; i++) {
 	tag = PyInt_FromLong(rpmTagTable[i].val);
-	PyDict_SetItemString(d, rpmTagTable[i].name, tag);
+	PyDict_SetItemString(d, (char *) rpmTagTable[i].name, tag);
 
         PyDict_SetItem(dict, tag, PyString_FromString(rpmTagTable[i].name + 7));
     }
 
+    while (extensions->name) {
+	if (extensions->type == HEADER_EXT_TAG) {
+            (const struct headerSprintfExtension *) ext = extensions;
+            PyDict_SetItemString(d, extensions->name, PyCObject_FromVoidPtr(ext, NULL));
+            PyDict_SetItem(dict, tag, PyString_FromString(ext->name + 7));
+        }
+        extensions++;
+    }
+
     PyDict_SetItemString(d, "tagnames", dict);
 
-    PyDict_SetItemString(d, "RPMFILE_STATE_NORMAL", 
+    PyDict_SetItemString(d, "RPMFILE_STATE_NORMAL",
 			 PyInt_FromLong(RPMFILE_STATE_NORMAL));
-    PyDict_SetItemString(d, "RPMFILE_STATE_REPLACED", 
+    PyDict_SetItemString(d, "RPMFILE_STATE_REPLACED",
 			 PyInt_FromLong(RPMFILE_STATE_REPLACED));
-    PyDict_SetItemString(d, "RPMFILE_STATE_NOTINSTALLED", 
+    PyDict_SetItemString(d, "RPMFILE_STATE_NOTINSTALLED",
 			 PyInt_FromLong(RPMFILE_STATE_NOTINSTALLED));
-    PyDict_SetItemString(d, "RPMFILE_CONFIG", 
+    PyDict_SetItemString(d, "RPMFILE_CONFIG",
 			 PyInt_FromLong(RPMFILE_CONFIG));
     PyDict_SetItemString(d, "RPMFILE_MISSINGOK",
 			 PyInt_FromLong(RPMFILE_MISSINGOK));
-    PyDict_SetItemString(d, "RPMFILE_DOC", 
+    PyDict_SetItemString(d, "RPMFILE_DOC",
 			 PyInt_FromLong(RPMFILE_DOC));
 
-    PyDict_SetItemString(d, "RPMDEP_SENSE_REQUIRES", 
+    PyDict_SetItemString(d, "RPMDEP_SENSE_REQUIRES",
 			 PyInt_FromLong(RPMDEP_SENSE_REQUIRES));
-    PyDict_SetItemString(d, "RPMDEP_SENSE_CONFLICTS", 
+    PyDict_SetItemString(d, "RPMDEP_SENSE_CONFLICTS",
 			 PyInt_FromLong(RPMDEP_SENSE_CONFLICTS));
 
     PyDict_SetItemString(d, "RPMSENSE_SERIAL",
@@ -334,6 +365,28 @@ void initrpm(void) {
 			 PyInt_FromLong(RPMPROB_DISKSPACE));
 }
 
+/* make a header with _all_ the tags we need */
+void mungeFilelist(Header h)
+{
+    const char ** fileNames = NULL;
+    int count = 0;
+
+    if (!headerIsEntry (h, RPMTAG_BASENAMES)
+	|| !headerIsEntry (h, RPMTAG_DIRNAMES)
+	|| !headerIsEntry (h, RPMTAG_DIRINDEXES))
+	compressFilelist(h);
+    
+    rpmBuildFileList(h, &fileNames, &count);
+
+    if (fileNames == NULL || count <= 0)
+	return;
+
+    headerAddEntry(h, RPMTAG_OLDFILENAMES, RPM_STRING_ARRAY_TYPE,
+			fileNames, count);
+
+    xfree(fileNames);
+}
+
 
 static int psGetArchScore(Header h) {
     void * pkgArch;
@@ -421,12 +474,12 @@ static PyObject * findUpgradeSet(PyObject * self, PyObject * args) {
 	list.packages[i]->selected = 0;
 	list.packages[i]->data = hdr;
 
-	headerGetEntry(hdr->h, RPMTAG_NAME, NULL, 
+	headerGetEntry(hdr->h, RPMTAG_NAME, NULL,
 		      (void **) &list.packages[i]->name, NULL);
     }
 
     pkgSort (&list);
-    
+
     if (ugFindUpgradePackages(&list, root)) {
 	PyErr_SetString(pyrpmError, "error during upgrade check");
 	return NULL;
@@ -457,7 +510,7 @@ static rpmdbObject * rpmOpenDB(PyObject * self, PyObject * args) {
 	char * errmsg = "cannot open database in %s";
 	char * errstr = NULL;
 	int errsize;
-	
+
 	Py_DECREF(o);
 	/* PyErr_SetString should take varargs... */
 	errsize = strlen(errmsg) + *root == '\0' ? 15 /* "/var/lib/rpm" */ : strlen(root);
@@ -518,12 +571,12 @@ static PyObject * rpmHeaderFromFD(PyObject * self, PyObject * args) {
     FD_t fd;
     int fileno;
     PyObject * list;
-    
+
     if (!PyArg_ParseTuple(args, "i", &fileno)) return NULL;
     fd = fdDup(fileno);
- 
+
     list = rpmReadHeaders (fd);
-    fdClose(fd);
+    Fclose(fd);
 
     return list;
 }
@@ -548,7 +601,7 @@ static PyObject * hdrLoad(PyObject * self, PyObject * args) {
     h->fileList = h->linkList = h->md5list = NULL;
     h->uids = h->gids = h->mtimes = h->fileSizes = NULL;
     h->modes = h->rdevs = NULL;
-    
+
     return (PyObject *) h;
 }
 
@@ -558,7 +611,7 @@ static PyObject * rpmHeaderFromFile(PyObject * self, PyObject * args) {
     PyObject * list;
 
     if (!PyArg_ParseTuple(args, "s", &filespec)) return NULL;
-    fd = fdOpen(filespec, O_RDONLY, 0);
+    fd = Fopen(filespec, "r.fdio");
 
     if (!fd) {
 	PyErr_SetFromErrno(pyrpmError);
@@ -566,8 +619,8 @@ static PyObject * rpmHeaderFromFile(PyObject * self, PyObject * args) {
     }
 
     list = rpmReadHeaders (fd);
-    fdClose(fd);
-    
+    Fclose(fd);
+
     return list;
 }
 
@@ -577,7 +630,7 @@ static void errorcb (void)
 {
     PyObject * result, * args = NULL;
 
-    if (errorData) 
+    if (errorData)
 	args = Py_BuildValue("(O)", errorData);
 
     result = PyEval_CallObject(errorCB, args);
@@ -602,19 +655,32 @@ static PyObject * errorSetCallback (PyObject * self, PyObject * args) {
     }
 
     if (!PyArg_ParseTuple(args, "O|O", &errorCB, &errorData)) return NULL;
+
+    /* if we're getting a void*, set the error callback to this. */
+    /* also, we can possibly decref any python callbacks we had  */
+    /* and set them to NULL.                                     */
+    if (PyCObject_Check (errorCB)) {
+	rpmErrorSetCallback (PyCObject_AsVoidPtr(errorCB));
+
+	Py_XDECREF (errorCB);
+	Py_XDECREF (errorData);
+
+	errorCB   = NULL;
+	errorData = NULL;
+	
+	Py_INCREF(Py_None);
+	return Py_None;
+    }
     
     if (!PyCallable_Check (errorCB)) {
 	PyErr_SetString(PyExc_TypeError, "parameter must be callable");
 	return NULL;
-    }   
+    }
 
     Py_INCREF (errorCB);
     Py_XINCREF (errorData);
-    
-    rpmErrorSetCallback (errorcb);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    return PyCObject_FromVoidPtr(rpmErrorSetCallback (errorcb), NULL);
 }
 
 static PyObject * errorString (PyObject * self, PyObject * args) {
@@ -623,7 +689,7 @@ static PyObject * errorString (PyObject * self, PyObject * args) {
 
 static PyObject * versionCompare (PyObject * self, PyObject * args) {
     hdrObject * h1, * h2;
-    
+
     if (!PyArg_ParseTuple(args, "O!O!", &hdrType, &h1, &hdrType, &h2)) return NULL;
 
     return Py_BuildValue("i", rpmVersionCompare(h1->h, h2->h));
@@ -639,7 +705,7 @@ static PyObject * labelCompare (PyObject * self, PyObject * args) {
 
     if (e1 && !e2)
 	return Py_BuildValue("i", 1);
-    else if (!e1 && e2) 
+    else if (!e1 && e2)
 	return Py_BuildValue("i", -1);
     else if (e1 && e2) {
 	int ep1, ep2;
@@ -650,7 +716,7 @@ static PyObject * labelCompare (PyObject * self, PyObject * args) {
 	else if (ep1 > ep2)
 	    return Py_BuildValue("i", 1);
     }
-	
+
     rc = rpmvercmp(v1, v2);
     if (rc)
 	return Py_BuildValue("i", rc);
@@ -670,7 +736,7 @@ static PyObject * rpmHeaderFromPackage(PyObject * self, PyObject * args) {
     fd = fdDup(rawFd);
 
     rc = rpmReadPackageHeader(fd, &header, &isSource, NULL, NULL);
-    fdClose(fd);
+    Fclose(fd);
 
     switch (rc) {
       case 0:
@@ -808,7 +874,7 @@ static int rpmdbLength(rpmdbObject * s) {
 static hdrObject * rpmdbSubscript(rpmdbObject * s, PyObject * key) {
     int offset;
     hdrObject * h;
-    
+
     if (!PyInt_Check(key)) {
 	PyErr_SetString(PyExc_TypeError, "integer expected");
 	return NULL;
@@ -827,7 +893,7 @@ static hdrObject * rpmdbSubscript(rpmdbObject * s, PyObject * key) {
 	PyErr_SetString(pyrpmError, "cannot read rpmdb entry");
 	return NULL;
     }
-    
+
     return h;
 }
 
@@ -850,29 +916,49 @@ static PyObject * hdrSubscript(hdrObject * s, PyObject * item) {
     PyObject * o, * metao;
     char ** stringArray;
     int forceArray = 0;
+    int freeData = 0;
     char * str;
+    struct headerSprintfExtension * ext = NULL;
+    const struct headerSprintfExtension * extensions = rpmHeaderFormats;
 
-    if (PyInt_Check(item)) {
+    if (PyCObject_Check (item)) {
+        ext = PyCObject_AsVoidPtr(item);
+    } else if (PyInt_Check(item)) {
 	tag = PyInt_AsLong(item);
     } else if (PyString_Check(item)) {
 	str = PyString_AsString(item);
 	for (i = 0; i < rpmTagTableSize; i++)
 	    if (!strcasecmp(rpmTagTable[i].name + 7, str)) break;
 	if (i < rpmTagTableSize) tag = rpmTagTable[i].val;
+        if (tag == -1) {
+            /* if we still don't have the tag, go looking for the header
+               extensions */
+            while (extensions->name) {
+                if (extensions->type == HEADER_EXT_TAG
+                    && !strcasecmp(extensions->name + 7, str)) {
+                    (const struct headerSprintfExtension *) ext = extensions;
+                }
+                extensions++;
+            }
+        }
     }
 
-    if (tag == -1) {
-	PyErr_SetString(PyExc_KeyError, "unknown header tag");
-	return NULL;
-    }
-
-    if (!headerGetEntry(s->h, tag, &type, &data, &count)) {
-	Py_INCREF(Py_None);
-	return Py_None;
+    if (ext) {
+        ext->u.tagFunction(s->h, &type, (const void **) &data, &count, &freeData);
+    } else {
+        if (tag == -1) {
+            PyErr_SetString(PyExc_KeyError, "unknown header tag");
+            return NULL;
+        }
+        
+        if (!rpmHeaderGetEntry(s->h, tag, &type, &data, &count)) {
+            Py_INCREF(Py_None);
+            return Py_None;
+        }
     }
 
     switch (tag) {
-      case RPMTAG_FILENAMES:
+      case RPMTAG_OLDFILENAMES:
       case RPMTAG_FILESIZES:
       case RPMTAG_FILESTATES:
       case RPMTAG_FILEMODES:
@@ -887,6 +973,14 @@ static PyObject * hdrSubscript(hdrObject * s, PyObject * item) {
       case RPMTAG_FILEUSERNAME:
       case RPMTAG_FILEGROUPNAME:
 	forceArray = 1;
+	break;
+      case RPMTAG_SUMMARY:
+      case RPMTAG_GROUP:
+      case RPMTAG_DESCRIPTION:
+	freeData = 1;
+	break;
+      default:
+        break;
     }
 
     switch (type) {
@@ -952,7 +1046,7 @@ static PyObject * hdrSubscript(hdrObject * s, PyObject * item) {
       case RPM_STRING_TYPE:
 	if (count != 1 || forceArray) {
 	    stringArray = data;
-	    
+
 	    metao = PyList_New(0);
 	    for (i=0; i < count; i++) {
 		o = PyString_FromString(stringArray[i]);
@@ -960,8 +1054,11 @@ static PyObject * hdrSubscript(hdrObject * s, PyObject * item) {
 		Py_DECREF(o);
 	    }
 	    o = metao;
-	} else
-	  o = PyString_FromString(data);
+	} else {
+	    o = PyString_FromString(data);
+	    if (freeData)
+		free (data);
+	}
 	break;
 
       default:
@@ -1009,7 +1106,7 @@ static PyObject * hdrUnload(hdrObject * s, PyObject * args) {
 
     len = headerSizeof(s->h, 0);
     buf = headerUnload(s->h);
-    
+
     rc = PyString_FromStringAndSize(buf, len);
     free(buf);
 
@@ -1050,7 +1147,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
     if (!verifyResult) return list;
 
     if (!s->fileList) {
-	headerGetEntry(s->h, RPMTAG_FILENAMES, &type, (void **) &s->fileList, 
+	headerGetEntry(s->h, RPMTAG_OLDFILENAMES, &type, (void **) &s->fileList,
 		 &count);
     }
 
@@ -1058,14 +1155,14 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_MD5) {
 	if (!s->md5list) {
-	    headerGetEntry(s->h, RPMTAG_FILEMD5S, &type, (void **) &s->md5list, 
+	    headerGetEntry(s->h, RPMTAG_FILEMD5S, &type, (void **) &s->md5list,
 		     &count);
 	}
-	
+
 	if (mdfile(s->fileList[fileNumber], buf)) {
 	    strcpy(buf, "(unknown)");
-	} 
-	
+	}
+
 	tuple = PyTuple_New(3);
 	attrName = PyString_FromString("checksum");
 	PyTuple_SetItem(tuple, 0, attrName);
@@ -1076,7 +1173,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_FILESIZE) {
 	if (!s->fileSizes) {
-	    headerGetEntry(s->h, RPMTAG_FILESIZES, &type, (void **) &s->fileSizes, 
+	    headerGetEntry(s->h, RPMTAG_FILESIZES, &type, (void **) &s->fileSizes,
 		     &count);
 
 	}
@@ -1094,7 +1191,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_LINKTO) {
 	if (!s->linkList) {
-	    headerGetEntry(s->h, RPMTAG_FILELINKTOS, &type, (void **) &s->linkList, 
+	    headerGetEntry(s->h, RPMTAG_FILELINKTOS, &type, (void **) &s->linkList,
 		     &count);
 	}
 
@@ -1114,7 +1211,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_MTIME) {
 	if (!s->mtimes) {
-	    headerGetEntry(s->h, RPMTAG_FILEMTIMES, &type, (void **) &s->mtimes, 
+	    headerGetEntry(s->h, RPMTAG_FILEMTIMES, &type, (void **) &s->mtimes,
 		     &count);
 	}
 
@@ -1138,7 +1235,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_RDEV) {
 	if (!s->rdevs) {
-	    headerGetEntry(s->h, RPMTAG_FILERDEVS, &type, (void **) &s->rdevs, 
+	    headerGetEntry(s->h, RPMTAG_FILERDEVS, &type, (void **) &s->rdevs,
 		     &count);
 	}
 
@@ -1157,7 +1254,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
        doesn't do these correctly either. At least this is consisten */
     if (verifyResult & RPMVERIFY_USER) {
 	if (!s->uids) {
-	    headerGetEntry(s->h, RPMTAG_FILEUIDS, &type, (void **) &s->uids, 
+	    headerGetEntry(s->h, RPMTAG_FILEUIDS, &type, (void **) &s->uids,
 		     &count);
 	}
 
@@ -1173,7 +1270,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_GROUP) {
 	if (!s->gids) {
-	    headerGetEntry(s->h, RPMTAG_FILEGIDS, &type, (void **) &s->gids, 
+	    headerGetEntry(s->h, RPMTAG_FILEGIDS, &type, (void **) &s->gids,
 		     &count);
 	}
 
@@ -1189,7 +1286,7 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
 
     if (verifyResult & RPMVERIFY_MODE) {
 	if (!s->modes) {
-	    headerGetEntry(s->h, RPMTAG_FILEMODES, &type, (void **) &s->modes, 
+	    headerGetEntry(s->h, RPMTAG_FILEMODES, &type, (void **) &s->modes,
 		     &count);
 	}
 
@@ -1206,6 +1303,27 @@ static PyObject * hdrVerifyFile(hdrObject * s, PyObject * args) {
     return list;
 }
 
+static PyObject * hdrCompressFilelist(hdrObject * s, PyObject * args) {
+    compressFilelist (s->h);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * hdrExpandFilelist(hdrObject * s, PyObject * args) {
+    expandFilelist (s->h);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject * hdrFullFilelist(hdrObject * s, PyObject * args) {
+    mungeFilelist (s->h);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 static PyObject * rpmtransCreate(PyObject * self, PyObject * args) {
     rpmtransObject * o;
     rpmdbObject * db = NULL;
@@ -1235,7 +1353,7 @@ static void rpmtransDealloc(PyObject * o) {
     if (trans->dbo) {
 	Py_DECREF(trans->dbo);
     }
-    if (trans->scriptFd) fdClose(trans->scriptFd);
+    if (trans->scriptFd) Fclose(trans->scriptFd);
     /* this will free the keyList, and decrement the ref count of all
        the items on the list as well :-) */
     Py_DECREF(trans->keyList);
@@ -1318,10 +1436,10 @@ static PyObject * rpmtransDepCheck(rpmtransObject * s, PyObject * args) {
 	list = PyList_New(0);
 
 	for (i = 0; i < numConflicts; i++) {
-	    cf = Py_BuildValue("((sss)(ss)iOi)", conflicts[i].byName, 
+	    cf = Py_BuildValue("((sss)(ss)iOi)", conflicts[i].byName,
 			       conflicts[i].byVersion, conflicts[i].byRelease,
 
-			       conflicts[i].needsName, 
+			       conflicts[i].needsName,
 			       conflicts[i].needsVersion,
 
 			       conflicts[i].needsFlags,
@@ -1347,7 +1465,7 @@ struct tsCallbackType {
     int pythonError;
 };
 
-static void * tsCallback(const Header h, const rpmCallbackType what, 
+static void * tsCallback(const Header h, const rpmCallbackType what,
 		         const unsigned long amount, const unsigned long total,
 	                 const void * pkgKey, void * data) {
     struct tsCallbackType * cbInfo = data;
@@ -1374,12 +1492,13 @@ static void * tsCallback(const Header h, const rpmCallbackType what,
 	    return NULL;
 	}
 	fdt = fdDup(fd);
+	
 	Py_DECREF(result);
 	return fdt;
     }
 
     if (what == RPMCALLBACK_INST_CLOSE_FILE) {
-	fdClose (fdt);
+	Fclose (fdt);
     }
 
     Py_DECREF(result);
@@ -1395,16 +1514,16 @@ static PyObject * rpmtransRun(rpmtransObject * s, PyObject * args) {
     struct tsCallbackType cbInfo;
 
     if (!PyArg_ParseTuple(args, "iiOO", &flags, &ignoreSet, &cbInfo.cb,
-			  &cbInfo.data)) 
+			  &cbInfo.data))
 	return NULL;
 
     cbInfo.pythonError = 0;
 
-    rc = rpmRunTransactions(s->ts, tsCallback, &cbInfo, NULL, &probs, flags, 
+    rc = rpmRunTransactions(s->ts, tsCallback, &cbInfo, NULL, &probs, flags,
 			    ignoreSet);
 
     if (cbInfo.pythonError) {
-	if (rc > 0) 
+	if (rc > 0)
 	    rpmProblemSetFree(probs);
 	return NULL;
     }
@@ -1455,3 +1574,97 @@ static PyObject * doAddMacro(PyObject * self, PyObject * args) {
     Py_INCREF(Py_None);
     return Py_None;
 }
+
+static PyObject * doDelMacro(PyObject * self, PyObject * args) {
+    char * name;
+
+    if (!PyArg_ParseTuple(args, "s", &name))
+	return NULL;
+
+    delMacro(NULL, name);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+typedef struct FDlist_t FDlist;
+
+struct FDlist_t {
+    FILE *f;
+    FD_t fd;
+    FDlist *next;
+} ;
+
+static FDlist *fdhead = NULL;
+static FDlist *fdtail = NULL;
+
+static int closeCallback(FILE * f) {
+    FDlist *node, *last;
+
+    node = fdhead;
+    last = NULL;
+    while (node) {
+        if (node->f == f)
+            break;
+        last = node;
+        node = node->next;
+    }
+    if (node) {
+        if (last)
+            last->next = node->next;
+        else
+            fdhead = node->next;
+        printf ("closing\n");
+        node->fd = fdLink(node->fd, "closeCallback");
+        Fclose (node->fd);
+        while (node->fd)
+            node->fd = fdFree(node->fd, "closeCallback");
+        free (node);
+    }
+    return 0; 
+}
+
+static PyObject * doFopen(PyObject * self, PyObject * args) {
+    char * path, * mode;
+    FDlist *node;
+    
+    if (!PyArg_ParseTuple(args, "ss", &path, &mode))
+	return NULL;
+    
+    node = malloc (sizeof(FDlist));
+    
+    node->fd = Fopen(path, mode);
+    node->fd = fdLink(node->fd, "doFopen");
+    
+    if (!node->fd) {
+	PyErr_SetFromErrno(pyrpmError);
+        free (node);
+	return NULL;
+    }
+    
+    if (Ferror(node->fd)) {
+	const char *err = Fstrerror(node->fd);
+        free(node);
+	if (err) {
+	    PyErr_SetString(pyrpmError, err);
+	    return NULL;
+	}
+    }
+    node->f = fdGetFp(node->fd);
+    if (!node->f) {
+	PyErr_SetString(pyrpmError, "FD_t has no FILE*");
+        free(node);
+	return NULL;
+    }
+
+    node->next = NULL;
+    if (fdtail) {
+        fdtail->next = node;
+    } else {
+        fdhead = node;
+    }
+    fdtail = node;
+    
+    return PyFile_FromFile (node->f, path, mode, closeCallback);
+}
+
