@@ -1,6 +1,7 @@
 #include "system.h"
 
-#include "rpmbuild.h"
+#include <rpmbuild.h>
+#include <rpmurl.h>
 
 #include "build.h"
 #include "install.h"
@@ -20,10 +21,34 @@
 #define	GETOPT_DEFINEMACRO	1020
 #define	GETOPT_EVALMACRO	1021
 
-enum modes { MODE_QUERY, MODE_INSTALL, MODE_UNINSTALL, MODE_VERIFY,
-	     MODE_BUILD, MODE_REBUILD, MODE_CHECKSIG, MODE_RESIGN,
-	     MODE_RECOMPILE, MODE_QUERYTAGS, MODE_INITDB, MODE_TARBUILD,
-	     MODE_REBUILDDB, MODE_UNKNOWN };
+enum modes {
+    MODE_UNKNOWN	= 0,
+    MODE_QUERY		= (1 <<  0),
+    MODE_INSTALL	= (1 <<  1),
+    MODE_UNINSTALL	= (1 <<  2),
+    MODE_VERIFY		= (1 <<  3),
+    MODE_BUILD		= (1 <<  4),
+    MODE_REBUILD	= (1 <<  5),
+    MODE_CHECKSIG	= (1 <<  6),
+    MODE_RESIGN		= (1 <<  7),
+    MODE_RECOMPILE	= (1 <<  8),
+    MODE_QUERYTAGS	= (1 <<  9),
+    MODE_INITDB		= (1 << 10),
+    MODE_TARBUILD	= (1 << 11),
+    MODE_REBUILDDB	= (1 << 12)
+};
+
+#define	MODES_QV (MODE_QUERY | MODE_VERIFY)
+#define	MODES_BT (MODE_BUILD | MODE_TARBUILD | MODE_REBUILD | MODE_RECOMPILE)
+#define	MODES_IE (MODE_INSTALL | MODE_UNINSTALL)
+#define	MODES_DB (MODE_INITDB | MODE_REBUILDDB)
+#define	MODES_K	 (MODE_CHECKSIG | MODES_RESIGN)
+
+#define	MODES_FOR_DBPATH	(MODES_BT | MODES_IE | MODES_QV | MODES_DB)
+#define	MODES_FOR_TIMECHECK	(MODES_BT)
+#define	MODES_FOR_NODEPS	(MODES_BT | MODES_IE | MODE_VERIFY)
+#define	MODES_FOR_TEST		(MODES_BT | MODES_IE)
+#define	MODES_FOR_ROOT		(MODES_BT | MODES_IE | MODES_QV | MODES_DB)
 
 /* the flags for the various options */
 static int allFiles;
@@ -31,6 +56,7 @@ static int allMatches;
 static int badReloc;
 static int excldocs;
 static int force;
+extern int _ftp_debug;
 static char * ftpPort;
 static char * ftpProxy;
 static char * httpPort;
@@ -45,6 +71,7 @@ static int initdb;
 static int justdb;
 static int noDeps;
 static int noGpg;
+extern int noLibio;
 static int noMd5;
 static int noOrder;
 static int noPgp;
@@ -52,7 +79,6 @@ static int noScripts;
 static int noTriggers;
 static int noUsageMsg;
 static int oldPackage;
-static int showPercents;
 static char * pipeOutput;
 static char * prefix;
 static int queryTags;
@@ -61,17 +87,23 @@ static char * rcfile;
 static int replaceFiles;
 static int replacePackages;
 static char * rootdir;
+extern int _rpmio_debug;
+static int showPercents;
 static int showrc;
 static int signIt;
 static int test;
+extern int _url_debug;
+extern int _noDirTokens;
 
 static int showVersion;
 extern const char * rpmNAME;
 extern const char * rpmEVR;
 extern int rpmFLAGS;
 
+extern MacroContext rpmCLIMacroContext;
+
 static struct rpmQVArguments rpmQVArgs;
-static struct rpmBuildArguments buildArgs;
+static struct rpmBuildArguments rpmBArgs;
 
 /* the structure describing the options we take and the defaults */
 static struct poptOption optionsTable[] = {
@@ -85,11 +117,13 @@ static struct poptOption optionsTable[] = {
  { "checksig", 'K', 0, 0, 'K',			NULL, NULL},
  { "dbpath", '\0', POPT_ARG_STRING, 0, GETOPT_DBPATH,		NULL, NULL},
  { "define", '\0', POPT_ARG_STRING, &noUsageMsg, GETOPT_DEFINEMACRO,NULL, NULL},
+ { "dirtokens", '\0', POPT_ARG_VAL, &_noDirTokens, 0,	NULL, NULL},
  { "erase", 'e', 0, 0, 'e',			NULL, NULL},
  { "eval", '\0', POPT_ARG_STRING, &noUsageMsg, GETOPT_EVALMACRO, NULL, NULL},
  { "excludedocs", '\0', 0, &excldocs, 0,	NULL, NULL},
  { "excludepath", '\0', POPT_ARG_STRING, 0, GETOPT_EXCLUDEPATH,	NULL, NULL},
  { "force", '\0', 0, &force, 0,			NULL, NULL},
+ { "ftpdebug", '\0', POPT_ARG_VAL, &_ftp_debug, -1,		NULL, NULL},
  { "ftpport", '\0', POPT_ARG_STRING, &ftpPort, 0,	NULL, NULL},
  { "ftpproxy", '\0', POPT_ARG_STRING, &ftpProxy, 0,	NULL, NULL},
  { "hash", 'h', 0, &showHash, 0,		NULL, NULL},
@@ -106,7 +140,11 @@ static struct poptOption optionsTable[] = {
  { "install", '\0', 0, 0, GETOPT_INSTALL,	NULL, NULL},
  { "justdb", '\0', 0, &justdb, 0,		NULL, NULL},
  { "nodeps", '\0', 0, &noDeps, 0,		NULL, NULL},
+ { "nodirtokens", '\0', POPT_ARG_VAL, &_noDirTokens, 1,	NULL, NULL},
  { "nogpg", '\0', 0, &noGpg, 0,			NULL, NULL},
+#if HAVE_LIBIO_H && defined(_IO_BAD_SEEN)
+ { "nolibio", '\0', POPT_ARG_VAL, &noLibio, 1,		NULL, NULL},
+#endif
  { "nomd5", '\0', 0, &noMd5, 0,			NULL, NULL},
  { "noorder", '\0', 0, &noOrder, 0,		NULL, NULL},
  { "nopgp", '\0', 0, &noPgp, 0,			NULL, NULL},
@@ -128,12 +166,14 @@ static struct poptOption optionsTable[] = {
  { "replacepkgs", '\0', 0, &replacePackages, 0,	NULL, NULL},
  { "resign", '\0', 0, 0, GETOPT_RESIGN,		NULL, NULL},
  { "root", 'r', POPT_ARG_STRING, &rootdir, 0,	NULL, NULL},
+ { "rpmiodebug", '\0', POPT_ARG_VAL, &_rpmio_debug, -1,		NULL, NULL},
  { "showrc", '\0', 0, &showrc, GETOPT_SHOWRC,	NULL, NULL},
  { "sign", '\0', 0, &signIt, 0,			NULL, NULL},
- { "tarball", 't', POPT_ARG_STRING, 0, 't',	NULL, NULL},
+ { "tarbuild", 't', POPT_ARG_STRING, 0, 't',	NULL, NULL},
  { "test", '\0', 0, &test, 0,			NULL, NULL},
  { "timecheck", '\0', POPT_ARG_STRING, 0, GETOPT_TIMECHECK,	NULL, NULL},
  { "upgrade", 'U', 0, 0, 'U',			NULL, NULL},
+ { "urldebug", '\0', POPT_ARG_VAL, &_url_debug, -1,		NULL, NULL},
  { "uninstall", 'u', 0, 0, 'u',			NULL, NULL},
  { "verbose", 'v', 0, 0, 'v',			NULL, NULL},
  { "verify", 'V', 0, 0, 'V',			NULL, NULL},
@@ -146,7 +186,7 @@ static struct poptOption optionsTable[] = {
  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, 
 		rpmVerifyPoptTable, 0,		(void *) &rpmQVArgs, NULL },
  { NULL, '\0', POPT_ARG_INCLUDE_TABLE, 
-		rpmBuildPoptTable, 0,		(void *) &buildArgs, NULL },
+		rpmBuildPoptTable, 0,		(void *) &rpmBArgs, NULL },
  { 0, 0, 0, 0, 0,	NULL, NULL }
 };
 
@@ -222,7 +262,7 @@ static void printUsage(void) {
     puts(_("       rpm {-b|t}[plciba] [-v] [--short-circuit] [--clean] [--rcfile  <file>]"));
     puts(_("                        [--sign] [--nobuild] [--timecheck <s>] ]"));
     puts(_("                        [--target=platform1[,platform2...]]"));
-    puts(_("                        [--rmsource] specfile"));
+    puts(_("                        [--rmsource] [--rmspec] specfile"));
     puts(_("       rpm {--rmsource} [--rcfile <file>] [-v] specfile"));
     puts(_("       rpm {--rebuild} [--rcfile <file>] [-v] source1.rpm ... sourceN.rpm"));
     puts(_("       rpm {--recompile} [--rcfile <file>] [-v] source1.rpm ... sourceN.rpm"));
@@ -499,10 +539,11 @@ static void printHelp(void) {
 		  _("use <dir> as the top level directory"));
 }
 
-int main(int argc, char ** argv)
+int main(int argc, const char ** argv)
 {
     enum modes bigMode = MODE_UNKNOWN;
     QVA_t *qva = &rpmQVArgs;
+    struct rpmBuildArguments *ba = &rpmBArgs;
     enum rpmQVSources QVSource = RPMQV_PACKAGE;
     int arg;
     int installFlags = 0, uninstallFlags = 0, interfaceFlags = 0;
@@ -514,7 +555,7 @@ int main(int argc, char ** argv)
     char * tce;
     char * passPhrase = "";
     char * cookie = NULL;
-    char * optArg;
+    const char * optArg;
     pid_t pipeChild = 0;
     const char * pkg;
     char * errString = NULL;
@@ -541,6 +582,7 @@ int main(int argc, char ** argv)
     badReloc = 0;
     excldocs = 0;
     force = 0;
+    _ftp_debug = 0;
     ftpProxy = NULL;
     ftpPort = NULL;
     httpProxy = NULL;
@@ -555,6 +597,11 @@ int main(int argc, char ** argv)
     justdb = 0;
     noDeps = 0;
     noGpg = 0;
+#if HAVE_LIBIO_H && defined(_IO_BAD_SEEN)
+    noLibio = 0;
+#else
+    noLibio = 1;
+#endif
     noMd5 = 0;
     noOrder = 0;
     noPgp = 0;
@@ -567,18 +614,20 @@ int main(int argc, char ** argv)
     prefix = NULL;
     queryTags = 0;
     quiet = 0;
+    _rpmio_debug = 0;
     replaceFiles = 0;
     replacePackages = 0;
     rootdir = "/";
     showrc = 0;
     signIt = 0;
-    test = 0;
     showVersion = 0;
+    specedit = 0;
+    test = 0;
+    _url_debug = 0;
 
     /* XXX Eliminate query linkage loop */
     parseSpecVec = parseSpec;
     freeSpecVec = freeSpec;
-    specedit = NULL;
 
     /* set up the correct locale */
     setlocale(LC_ALL, "" );
@@ -625,10 +674,10 @@ int main(int argc, char ** argv)
 
     if (qva->qva_queryFormat) xfree(qva->qva_queryFormat);
     memset(qva, 0, sizeof(*qva));
-    if (buildArgs.buildRootOverride) xfree(buildArgs.buildRootOverride);
-    if (buildArgs.targets) free(buildArgs.targets);
-    memset(&buildArgs, 0, sizeof(buildArgs));
-    buildArgs.buildChar = ' ';
+    if (ba->buildRootOverride) xfree(ba->buildRootOverride);
+    if (ba->targets) free(ba->targets);
+    memset(ba, 0, sizeof(*ba));
+    ba->buildChar = ' ';
 
     while ((arg = poptGetNextOpt(optCon)) > 0) {
 	optArg = poptGetOptArg(optCon);
@@ -686,8 +735,8 @@ int main(int argc, char ** argv)
 	    if (strlen(optArg) > 1) 
 		argerror(errString);
 
-	    buildArgs.buildChar = optArg[0];
-	    switch (buildArgs.buildChar) {
+	    ba->buildChar = optArg[0];
+	    switch (ba->buildChar) {
 	      case 'a':
 	      case 'b':
 	      case 'i':
@@ -785,14 +834,22 @@ int main(int argc, char ** argv)
 	    break;
 
 	  case GETOPT_DBPATH:
-            if (optArg[0] != '/')
-                argerror(_("arguments to --dbpath must begin with a /"));
-	    addMacro(NULL,"_dbpath", NULL, optArg, RMIL_CMDLINE);
+	    switch (urlIsURL(optArg)) {
+	    case URL_IS_UNKNOWN:
+		if (optArg[0] != '/')
+		    argerror(_("arguments to --dbpath must begin with a /"));
+		break;
+	    default:
+		break;
+	    }
+	    addMacro(NULL, "_dbpath", NULL, optArg, RMIL_CMDLINE);
+	    addMacro(&rpmCLIMacroContext, "_dbpath", NULL, optArg, RMIL_CMDLINE);
 	    gotDbpath = 1;
 	    break;
 
 	  case GETOPT_DEFINEMACRO:
 	    rpmDefineMacro(NULL, optArg, RMIL_CMDLINE);
+	    rpmDefineMacro(&rpmCLIMacroContext, optArg, RMIL_CMDLINE);
 	    break;
 
 	  case GETOPT_EVALMACRO:
@@ -808,6 +865,7 @@ int main(int argc, char ** argv)
 		argerror("Argument to --timecheck must be integer");
 	    }
 	    addMacro(NULL, "_timecheck", NULL, optArg, RMIL_CMDLINE);
+	    addMacro(&rpmCLIMacroContext, "_timecheck", NULL, optArg, RMIL_CMDLINE);
 	    timeCheck = 1;
 	    break;
 
@@ -860,10 +918,10 @@ int main(int argc, char ** argv)
 	exit(EXIT_FAILURE);
     }
 
-    if ((buildArgs.buildAmount & RPMBUILD_RMSOURCE) && bigMode == MODE_UNKNOWN)
+    if ((ba->buildAmount & RPMBUILD_RMSOURCE) && bigMode == MODE_UNKNOWN)
 	bigMode = MODE_BUILD;
 
-    if ((buildArgs.buildAmount & RPMBUILD_RMSPEC) && bigMode == MODE_UNKNOWN)
+    if ((ba->buildAmount & RPMBUILD_RMSPEC) && bigMode == MODE_UNKNOWN)
 	bigMode = MODE_BUILD;
     
     if (initdb) {
@@ -887,34 +945,30 @@ int main(int argc, char ** argv)
 	QVSource = qva->qva_source;
     }
 
-    if (buildArgs.buildRootOverride && bigMode != MODE_BUILD &&
+    if (ba->buildRootOverride && bigMode != MODE_BUILD &&
 	bigMode != MODE_REBUILD && bigMode != MODE_TARBUILD) {
 	argerror("--buildroot may only be used during package builds");
     }
 
-    if (bigMode != MODE_QUERY && bigMode != MODE_INSTALL && 
-	bigMode != MODE_UNINSTALL && bigMode != MODE_VERIFY &&
-	bigMode != MODE_INITDB && bigMode != MODE_REBUILDDB && gotDbpath)
+    if (gotDbpath && (bigMode & ~MODES_FOR_DBPATH))
 	argerror(_("--dbpath given for operation that does not use a "
 			"database"));
 
-    if (timeCheck && bigMode != MODE_BUILD && bigMode != MODE_REBUILD &&
-	bigMode != MODE_RECOMPILE && bigMode != MODE_TARBUILD) 
+    if (timeCheck && (bigMode & ~MODES_FOR_TIMECHECK))
 	argerror(_("--timecheck may only be used during package builds"));
     
-    if (!(bigMode == MODE_QUERY || bigMode == MODE_VERIFY) && qva->qva_flags) 
+    if (qva->qva_flags && (bigMode & ~MODES_QV)) 
 	argerror(_("unexpected query flags"));
 
-    if (bigMode != MODE_QUERY && qva->qva_queryFormat) 
+    if (qva->qva_queryFormat && (bigMode & ~MODES_QV)) 
 	argerror(_("unexpected query format"));
 
-    if (bigMode != MODE_QUERY && bigMode != MODE_VERIFY &&
-	QVSource != RPMQV_PACKAGE) 
+    if (QVSource != RPMQV_PACKAGE && (bigMode & ~MODES_QV)) 
 	argerror(_("unexpected query source"));
 
     if (!(bigMode == MODE_INSTALL ||
-	 (bigMode==MODE_BUILD && (buildArgs.buildAmount & RPMBUILD_RMSOURCE))||
-	 (bigMode==MODE_BUILD && (buildArgs.buildAmount & RPMBUILD_RMSPEC))) 
+	 (bigMode==MODE_BUILD && (ba->buildAmount & RPMBUILD_RMSOURCE))||
+	 (bigMode==MODE_BUILD && (ba->buildAmount & RPMBUILD_RMSPEC))) 
 	&& force)
 	argerror(_("only installation, upgrading, rmsource and rmspec may be forced"));
 
@@ -994,25 +1048,32 @@ int main(int argc, char ** argv)
 	argerror(_("--notriggers may only be specified during package "
 		   "installation, erasure, and verification"));
 
-    if (bigMode != MODE_INSTALL && bigMode != MODE_UNINSTALL && 
-	bigMode != MODE_BUILD && bigMode != MODE_TARBUILD &&
-	bigMode != MODE_VERIFY && noDeps)
+    if (noDeps & (bigMode & ~MODES_FOR_NODEPS))
 	argerror(_("--nodeps may only be specified during package "
-		   "building, installation, erasure, and verification"));
+		   "building, rebuilding, recompilation, installation,"
+		   "erasure, and verification"));
 
-    if (bigMode != MODE_INSTALL && bigMode != MODE_UNINSTALL &&
-	bigMode != MODE_BUILD && bigMode != MODE_TARBUILD && test )
+    if (test && (bigMode & ~MODES_FOR_TEST))
 	argerror(_("--test may only be specified during package installation, "
 		 "erasure, and building"));
 
-    if (bigMode != MODE_INSTALL && bigMode != MODE_UNINSTALL && 
-	bigMode != MODE_QUERY   && bigMode != MODE_VERIFY    && 			bigMode != MODE_REBUILDDB && bigMode != MODE_INITDB && rootdir[1])
+    if (rootdir[1] && (bigMode & ~MODES_FOR_ROOT))
 	argerror(_("--root (-r) may only be specified during "
 		 "installation, erasure, querying, and "
 		 "database rebuilds"));
 
-    if (rootdir && rootdir[0] != '/')
-	argerror(_("arguments to --root (-r) must begin with a /"));
+    if (rootdir) {
+	switch (urlIsURL(rootdir)) {
+	default:
+	    if (bigMode & MODES_FOR_ROOT)
+		break;
+	    /*@fallthrough@*/
+	case URL_IS_UNKNOWN:
+	    if (rootdir[0] != '/')
+		argerror(_("arguments to --root (-r) must begin with a /"));
+	    break;
+	}
+    }
 
     if (oldPackage && !upgrade)
 	argerror(_("--oldpackage may only be used during upgrades"));
@@ -1041,15 +1102,19 @@ int main(int argc, char ** argv)
 
     if (ftpProxy) {
 	addMacro(NULL, "_ftpproxy", NULL, ftpProxy, RMIL_CMDLINE);
+	addMacro(&rpmCLIMacroContext, "_ftpproxy", NULL, ftpProxy, RMIL_CMDLINE);
     }
     if (ftpPort) {
 	addMacro(NULL, "_ftpport", NULL, ftpPort, RMIL_CMDLINE);
+	addMacro(&rpmCLIMacroContext, "_ftpport", NULL, ftpPort, RMIL_CMDLINE);
     }
     if (httpProxy) {
 	addMacro(NULL, "_httpproxy", NULL, httpProxy, RMIL_CMDLINE);
+	addMacro(&rpmCLIMacroContext, "_httpproxy", NULL, httpProxy, RMIL_CMDLINE);
     }
     if (httpPort) {
 	addMacro(NULL, "_httpport", NULL, httpPort, RMIL_CMDLINE);
+	addMacro(&rpmCLIMacroContext, "_httpport", NULL, httpPort, RMIL_CMDLINE);
     }
 
     if (signIt) {
@@ -1082,7 +1147,7 @@ int main(int argc, char ** argv)
 		        !rpmDetectPGPVersion(NULL)) {
 		        fprintf(stderr, _("pgp not found: "));
 			exit(EXIT_FAILURE);
-		    }	/* fall through */
+		    }	/*@fallthrough@*/
 		  case RPMSIGTAG_GPG:
 		    passPhrase = rpmGetPassPhrase(_("Enter pass phrase: "), sigTag);
 		    if (passPhrase == NULL) {
@@ -1096,7 +1161,7 @@ int main(int argc, char ** argv)
 		    fprintf(stderr,
 		            _("Invalid %%_signature spec in macro file.\n"));
 		    exit(EXIT_FAILURE);
-		    break;
+		    /*@notreached@*/ break;
 		}
 	    }
 	} else {
@@ -1152,7 +1217,6 @@ int main(int argc, char ** argv)
 	ec = rpmCheckSig(checksigFlags, (const char **)poptGetArgs(optCon));
 	/* XXX don't overflow single byte exit status */
 	if (ec > 255) ec = 255;
-	exit(ec);
 	break;
 
       case MODE_RESIGN:
@@ -1161,7 +1225,6 @@ int main(int argc, char ** argv)
 	ec = rpmReSign(addSign, passPhrase, (const char **)poptGetArgs(optCon));
 	/* XXX don't overflow single byte exit status */
 	if (ec > 255) ec = 255;
-	exit(ec);
 	break;
 	
       case MODE_REBUILD:
@@ -1172,22 +1235,24 @@ int main(int argc, char ** argv)
 	if (!poptPeekArg(optCon))
 	    argerror(_("no packages files given for rebuild"));
 
-	buildArgs.buildAmount = RPMBUILD_PREP | RPMBUILD_BUILD | RPMBUILD_INSTALL;
+	ba->buildAmount = RPMBUILD_PREP | RPMBUILD_BUILD | RPMBUILD_INSTALL;
 	if (bigMode == MODE_REBUILD) {
-	    buildArgs.buildAmount |= RPMBUILD_PACKAGEBINARY;
-	    buildArgs.buildAmount |= RPMBUILD_RMSOURCE;
-	    buildArgs.buildAmount |= RPMBUILD_RMSPEC;
-	    buildArgs.buildAmount |= RPMBUILD_CLEAN;
-	    buildArgs.buildAmount |= RPMBUILD_RMBUILD;
+	    ba->buildAmount |= RPMBUILD_PACKAGEBINARY;
+	    ba->buildAmount |= RPMBUILD_RMSOURCE;
+	    ba->buildAmount |= RPMBUILD_RMSPEC;
+	    ba->buildAmount |= RPMBUILD_CLEAN;
+	    ba->buildAmount |= RPMBUILD_RMBUILD;
 	}
 
 	while ((pkg = poptGetArg(optCon))) {
-	    if (rpmInstallSource("/", pkg, &specFile, &cookie))
-		exit(EXIT_FAILURE);
+	    ec = rpmInstallSource("", pkg, &specFile, &cookie);
+	    if (ec)
+		break;
 
-	    if (build(specFile, &buildArgs, passPhrase, 0, cookie, rcfile, force, noDeps)) {
-		exit(EXIT_FAILURE);
-	    }
+	    ba->rootdir = rootdir;
+	    ec = build(specFile, ba, passPhrase, 0, cookie, rcfile, force, noDeps);
+	    if (ec)
+		break;
 	    free(cookie);
 	    xfree(specFile);
 	}
@@ -1198,30 +1263,34 @@ int main(int argc, char ** argv)
         if (rpmGetVerbosity() == RPMMESS_NORMAL)
 	    rpmSetVerbosity(RPMMESS_VERBOSE);
        
-	switch (buildArgs.buildChar) {
+	switch (ba->buildChar) {
 	  /* these fallthroughs are intentional */
 	  case 'a':
-	    buildArgs.buildAmount |= RPMBUILD_PACKAGESOURCE;
+	    ba->buildAmount |= RPMBUILD_PACKAGESOURCE;
+	    /*@fallthrough@*/
 	  case 'b':
-	    buildArgs.buildAmount |= RPMBUILD_PACKAGEBINARY;
-	    buildArgs.buildAmount |= RPMBUILD_CLEAN;
+	    ba->buildAmount |= RPMBUILD_PACKAGEBINARY;
+	    ba->buildAmount |= RPMBUILD_CLEAN;
+	    /*@fallthrough@*/
 	  case 'i':
-	    buildArgs.buildAmount |= RPMBUILD_INSTALL;
-	    if ((buildArgs.buildChar == 'i') && buildArgs.shortCircuit)
+	    ba->buildAmount |= RPMBUILD_INSTALL;
+	    if ((ba->buildChar == 'i') && ba->shortCircuit)
 		break;
+	    /*@fallthrough@*/
 	  case 'c':
-	    buildArgs.buildAmount |= RPMBUILD_BUILD;
-	    if ((buildArgs.buildChar == 'c') && buildArgs.shortCircuit)
+	    ba->buildAmount |= RPMBUILD_BUILD;
+	    if ((ba->buildChar == 'c') && ba->shortCircuit)
 		break;
+	    /*@fallthrough@*/
 	  case 'p':
-	    buildArgs.buildAmount |= RPMBUILD_PREP;
+	    ba->buildAmount |= RPMBUILD_PREP;
 	    break;
 	    
 	  case 'l':
-	    buildArgs.buildAmount |= RPMBUILD_FILECHECK;
+	    ba->buildAmount |= RPMBUILD_FILECHECK;
 	    break;
 	  case 's':
-	    buildArgs.buildAmount |= RPMBUILD_PACKAGESOURCE;
+	    ba->buildAmount |= RPMBUILD_PACKAGESOURCE;
 	    break;
 	}
 
@@ -1232,11 +1301,15 @@ int main(int argc, char ** argv)
 		argerror(_("no tar files given for build"));
 	}
 
-	while ((pkg = poptGetArg(optCon)))
-	    if (build(pkg, &buildArgs, passPhrase, bigMode == MODE_TARBUILD,
-			NULL, rcfile, force, noDeps)) {
-		exit(EXIT_FAILURE);
-	    }
+	while ((pkg = poptGetArg(optCon))) {
+	    ba->rootdir = rootdir;
+	    ec = build(pkg, ba, passPhrase, bigMode == MODE_TARBUILD,
+			NULL, rcfile, force, noDeps);
+	    if (ec)
+		break;
+	    rpmFreeMacros(NULL);
+	    rpmReadConfigFiles(rcfile, NULL);
+	}
 	break;
 
       case MODE_UNINSTALL:
@@ -1321,7 +1394,7 @@ int main(int argc, char ** argv)
 	    if (!poptPeekArg(optCon))
 		argerror(_("no arguments given for query"));
 	    while ((pkg = poptGetArg(optCon)))
-		ec = rpmQuery(qva, QVSource, pkg);
+		ec += rpmQuery(qva, QVSource, pkg);
 	}
 	break;
 
@@ -1342,13 +1415,14 @@ int main(int argc, char ** argv)
 	    if (!poptPeekArg(optCon))
 		argerror(_("no arguments given for verify"));
 	    while ((pkg = poptGetArg(optCon)))
-		ec = rpmVerify(qva, QVSource, pkg);
+		ec += rpmVerify(qva, QVSource, pkg);
 	}
 	break;
     }
 
     poptFreeContext(optCon);
-    freeMacros(NULL);
+    rpmFreeMacros(NULL);
+    rpmFreeMacros(&rpmCLIMacroContext);
     rpmFreeRpmrc();
 
     if (pipeChild) {
@@ -1357,7 +1431,15 @@ int main(int argc, char ** argv)
     }
 
     /* keeps memory leak checkers quiet */
+    freeNames();
+    freeFilesystems();
+    urlFreeCache();
     if (qva->qva_queryFormat) xfree(qva->qva_queryFormat);
+    if (ba->buildRootOverride) xfree(ba->buildRootOverride);
+    if (ba->targets) free(ba->targets);
 
+#if HAVE_MCHECK_H && HAVE_MTRACE
+    muntrace();   /* Trace malloc only if MALLOC_TRACE=mtrace-output-file. */
+#endif
     return ec;
 }
