@@ -7,26 +7,24 @@
 int main(int argc, char **argv)
 {
     FD_t fdi, fdo;
-    Header hd;
+    Header h;
     int rc, isSource;
-    char buffer[1024];
-    int ct;
     FD_t gzdi;
     
     setprogname(argv[0]);	/* Retrofit glibc __progname */
     if (argc == 1) {
 	fdi = fdDup(STDIN_FILENO);
     } else {
-	fdi = fdOpen(argv[1], O_RDONLY, 0644);
+	fdi = Fopen(argv[1], "r.ufdio");
     }
 
-    if (fdFileno(fdi) < 0) {
-	perror("cannot open package");
+    if (fdi == NULL || Ferror(fdi)) {
+	fprintf(stderr, _("cannot open package: %s\n"), Fstrerror(fdi));
 	exit(EXIT_FAILURE);
     }
     fdo = fdDup(STDOUT_FILENO);
 
-    rc = rpmReadPackageHeader(fdi, &hd, &isSource, NULL, NULL);
+    rc = rpmReadPackageHeader(fdi, &h, &isSource, NULL, NULL);
     switch (rc) {
     case 0:
 	break;
@@ -40,20 +38,33 @@ int main(int argc, char **argv)
 	break;
     }
 
-    gzdi = gzdFdopen(fdi, "r");	/* XXX gzdi == fdi */
+    gzdi = Fdopen(fdi, "r.gzdio");	/* XXX gzdi == fdi */
+    if (gzdi == NULL || Ferror(gzdi)) {
+	fprintf(stderr, _("cannot re-open payload: %s\n"), Fstrerror(gzdi));
+	exit(EXIT_FAILURE);
+    }
 
-    while ((ct = gzdRead(gzdi, &buffer, sizeof(buffer))) > 0) {
-	fdWrite(fdo, &buffer, ct);
+#ifdef	DYING
+  { char buffer[BUFSIZ];
+    int ct;
+    while ((ct = Fread(buffer, sizeof(buffer[0]), sizeof(buffer), gzdi)) > 0) {
+	Fwrite(buffer, sizeof(buffer[0]), ct, fdo);
     }
 
     if (ct < 0) {
-        fprintf (stderr, "rpm2cpio: zlib: %s\n", gzdStrerror(gzdi));
+        fprintf (stderr, "rpm2cpio: zlib: %s\n", Fstrerror(gzdi));
 	rc = EXIT_FAILURE;
     } else {
 	rc = EXIT_SUCCESS;
     }
+  }
+#else
+    rc = ufdCopy(gzdi, fdo);
+    rc = (rc <= 0) ? EXIT_FAILURE : EXIT_SUCCESS;
+    Fclose(fdo);
+#endif
 
-    gzdClose(gzdi);	/* XXX gzdi == fdi */
+    Fclose(gzdi);	/* XXX gzdi == fdi */
 
     return rc;
 }
