@@ -16,9 +16,13 @@ int main(int argc, char **argv)
     
     setprogname(argv[0]);	/* Retrofit glibc __progname */
     if (argc == 1) {
-	fdi = fdDup(STDIN_FILENO);
+	fdi = Fopen("-", "r.ufdio");
     } else {
-	fdi = fdOpen(argv[1], O_RDONLY, 0644);
+	fdi = Fopen(argv[1], "r.ufdio");
+    }
+    if (fdi == NULL || Ferror(fdi)) {
+	perror("input");
+	exit(EXIT_FAILURE);
     }
 
     readLead(fdi, &lead);
@@ -26,9 +30,9 @@ int main(int argc, char **argv)
     hd = headerRead(fdi, (lead.major >= 3) ?
 		    HEADER_MAGIC_YES : HEADER_MAGIC_NO);
 
-    fdo = fdDup(STDOUT_FILENO);
-    while ((ct = fdRead(fdi, &buffer, 1024))) {
-	fdWrite(fdo, &buffer, ct);
+    fdo = Fopen("-", "w.ufdio");
+    while ((ct = Fread(buffer, sizeof(buffer), 1, fdi))) {
+	Fwrite(buffer, ct, 1, fdo);
     }
     
     return 0;
