@@ -650,7 +650,7 @@ DPRINTF(100, ("\n"));
 						break;
 				}
 				if (l == nlangs)
-					onlylang[nlangs++] = strdup(lang);
+					onlylang[nlangs++] = xstrdup(lang);
 			}
 			lang = NULL;
 		}
@@ -761,7 +761,7 @@ rewriteBinaryRPM(char *fni, char *fno, message_list_ty *mlp)
     DPRINTF(99, ("rewriteBinaryRPM(\"%s\",\"%s\",%p)\n", fni, fno, mlp));
 
     csa->cpioArchiveSize = 0;
-    csa->cpioFdIn = fdNew();
+    csa->cpioFdIn = fdNew("init (rewriteBinary)");
     csa->cpioList = NULL;
     csa->cpioCount = 0;
     csa->lead = &lead;		/* XXX FIXME: exorcize lead/arch/os */
@@ -784,7 +784,7 @@ rewriteBinaryRPM(char *fni, char *fno, message_list_ty *mlp)
     }
 
 exit:
-    fdClose(csa->cpioFdIn);
+    Fclose(csa->cpioFdIn);
     return rc;
 
 }
@@ -836,7 +836,9 @@ rpmgettext(FD_t fd, const char *file, FILE *ofp)
 	    }
 	}
 
-	if (fdFileno(fd = fdOpen(fni, O_RDONLY, 0644)) < 0) {
+	fd = Fopen(fni, "r.fdio");
+	if (Ferror(fd)) {
+	    /* XXX Fstrerror */
 	    fprintf(stderr, _("rpmgettext: open %s: %s\n"), fni, strerror(errno));
 	    return 2;
 	}
@@ -847,7 +849,7 @@ rpmgettext(FD_t fd, const char *file, FILE *ofp)
 
 	if (ofp != stdout)
 	    fclose(ofp);
-	fdClose(fd);
+	Fclose(fd);
 
 	return 0;
 }
@@ -1005,7 +1007,7 @@ main(int argc, char **argv)
     while((c = getopt(argc, argv, "defgEMl:C:I:O:Tv")) != EOF)
     switch (c) {
     case 'C':
-	mastercatalogue = strdup(optarg);
+	mastercatalogue = xstrdup(optarg);
 	break;
     case 'd':
 	debug++;
@@ -1021,7 +1023,7 @@ main(int argc, char **argv)
 	break;
     case 'l':
 	gottalang = 1;
-	onlylang[nlangs++] = strdup(optarg);
+	onlylang[nlangs++] = xstrdup(optarg);
 	break;
     case 'I':
 	inputdir = optarg;
