@@ -12,17 +12,17 @@ int main(int argc, char **argv)
     
     setprogname(argv[0]);	/* Retrofit glibc __progname */
     if (argc == 1) {
-	fdi = fdDup(STDIN_FILENO);
+	fdi = Fopen("-", "r.ufdio");
     } else {
-	fdi = fdOpen(argv[1], O_RDONLY, 0644);
-	if (fdFileno(fdi) < 0) {
-	    perror(argv[1]);
-	    exit(1);
-	}
+	fdi = Fopen(argv[1], "r.ufdio");
+    }
+    if (fdi == NULL || Ferror(fdi)) {
+	perror(argv[1]);
+	exit(EXIT_FAILURE);
     }
 
     readLead(fdi, &lead);
-    fdo = fdDup(STDOUT_FILENO);
+    fdo = Fopen("-", "w.ufdio");
     writeLead(fdo, &lead);
     
     return 0;
